/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the S4 iOS Libraries.
 *
 * The Initial Developer of the Original Code is
 * Michael Papp dba SeaStones Software Company.
 * All software created by the Initial Developer are Copyright (C) 2008-2011
 * the Initial Developer. All Rights Reserved.
 *
 * Original Author:
 *			Michael Papp, San Francisco, CA, USA
 *
 * ***** END LICENSE BLOCK ***** */

/* ***** FILE BLOCK ******
 *
 * Name:		S4BookCache.m
 * Module:		UI
 * Library:		S4 iOS_UI Libraries
 *
 * ***** FILE BLOCK *****/


// ================================== Includes =========================================

#import "S4BookCache.h"
#import "S4CommonDefines.h"


// ================================== Defines ==========================================



// ================================== Typedefs ==========================================



// ================================== Globals ==========================================

static const NSUInteger						kMaxPagesInCache = 8;


// ============================= Forward Declarations ==================================



// ================================== Inlines ==========================================



// =========================== Begin Class S4BookCache () ==============================

@interface S4BookCache ()

@property (nonatomic, strong, readwrite) NSMutableArray				*pageRingBuffer;

@end



// ===================== Begin Class S4BookCache (PrivateImpl) =========================

@interface S4BookCache (PrivateImpl)


@end



// ============================ Begin Class S4BookCache ================================

@implementation S4BookCache


//============================================================================
//	S4BookCache synthesize properties
//============================================================================
@dynamic maxCacheCapacity;
@synthesize pageRingBuffer = m_pageRingBuffer;


//============================================================================
//	S4BookCache :: init
//============================================================================
- (id)init
{
	NSUInteger					idx;
	NSNull						*nullObject;

	self = [super init];
	if (nil != self)
	{
		m_cachedPageCGSize = CGSizeZero;
		m_maxCacheCapacity = kMaxPagesInCache;

		// set up the cache ring buffer
		m_pageRingBuffer = [[NSMutableArray alloc] initWithCapacity: kMaxPagesInCache];
		nullObject = [NSNull null];
		for (idx = 0; idx < m_maxCacheCapacity; idx++)
		{
			[m_pageRingBuffer addObject: nullObject];
		}
		m_cacheIndex = 0;
	}
	return self;
}


//============================================================================
//	S4BookCache :: dealloc
//============================================================================
- (void)dealloc
{
	NS_SAFE_RELEASE(m_pageRingBuffer);

	[super dealloc];
}


//============================================================================
//	S4BookCache :: maxCacheCapacity
//============================================================================
- (NSUInteger)maxCacheCapacity
{
	return (m_maxCacheCapacity);
}


//============================================================================
//	S4BookCache :: setMaxCacheCapacity:
//============================================================================
- (void)setMaxCacheCapacity: (NSUInteger)newCapacity
{
	NSUInteger				delta;
	NSNull					*nullObject;

	if (newCapacity > m_maxCacheCapacity)
	{
		nullObject = [NSNull null];
		delta = newCapacity - m_maxCacheCapacity;
		while (0 < delta)
		{
			[m_pageRingBuffer addObject: nullObject];
			delta--;
		}
	}
	m_maxCacheCapacity = newCapacity;
}


//============================================================================
//	S4BookCache :: getImageForPageNumber:withPageSize:
//============================================================================
- (S4BookCachePage *)bookCachePageForIndex: (NSUInteger)pageIndex withPageSize: (CGSize)newSize
{
	__block S4BookCachePage			*curBookPage = nil;

	if (true == CGSizeEqualToSize(m_cachedPageCGSize,  newSize))
	{
		if (NSNotFound == [self.pageRingBuffer indexOfObjectPassingTest: ^BOOL(id obj, NSUInteger idx, BOOL *stop)
						   {
							   curBookPage = (S4BookCachePage *)obj;
							   if ((IS_NOT_NULL(curBookPage)) && (curBookPage.pageIndex == pageIndex))
							   {
								   *stop = YES;
							   }
							   else
							   {
								   *stop = NO;
							   }
							   return (*stop);
						   }])
		{
			curBookPage = nil;
		}
	}
	else
	{
		m_cachedPageCGSize = newSize;
		[self flushBookCache];
	}
	return (curBookPage);
}


//============================================================================
//	S4BookCache :: getImageForPageNumber:withPageSize:
//============================================================================
- (CGImageRef)imageForIndex: (NSUInteger)pageIndex withPageSize: (CGSize)newSize
{
	S4BookCachePage					*curBookPage;

	curBookPage = [self bookCachePageForIndex: pageIndex withPageSize: newSize];
	if (nil != curBookPage)
	{
		return (curBookPage.pageImage);
	}
	return (NULL);
}


//============================================================================
//	S4BookCache :: setImage:forPageNumber:withPageSize:
//============================================================================
- (void)setBookCachePage: (S4BookCachePage *)bookCachePage withPageSize: (CGSize)newSize
{
	@synchronized(self)
	{
			if (false == CGSizeEqualToSize(m_cachedPageCGSize,  newSize))
			{
				m_cachedPageCGSize = newSize;
				[self flushBookCache];
			}

			// add it to the ring buffer and advance the cacheIndex (or roll it over)
			[m_pageRingBuffer replaceObjectAtIndex: m_cacheIndex withObject: bookCachePage];
			if ((m_cacheIndex + 1) == m_maxCacheCapacity)
			{
				m_cacheIndex = 0;
			}
			else
			{
				m_cacheIndex++;
			}
		}
}


//============================================================================
//	S4BookCache :: flushBookCache
//============================================================================
- (void)flushBookCache
{
	NSUInteger					idx;
	NSNull						*nullObject;

	// reset the ring buffer cache
	nullObject = [NSNull null];
	for (idx = 0; idx < m_maxCacheCapacity; idx++)
	{
		[m_pageRingBuffer replaceObjectAtIndex: idx withObject: nullObject];
	}
	m_cacheIndex = 0;
}

@end
