/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the S4 iOS Libraries.
 *
 * The Initial Developer of the Original Code is
 * Michael Papp dba SeaStones Software Company.
 * All software created by the Initial Developer are Copyright (C) 2008-2011
 * the Initial Developer. All Rights Reserved.
 *
 * Original Author:
 *			Michael Papp, San Francisco, CA, USA
 *
 * ***** END LICENSE BLOCK ***** */

/* ***** FILE BLOCK ******
 *
 * Name:		S4BookView.h
 * Module:		UI
 * Library:		S4 iOS_UI Libraries
 *
 * ***** FILE BLOCK *****/


// ================================== Includes =========================================

#import <UIKit/UIKit.h>
#import <QuartzCore/QuartzCore.h>
#import "S4BookCache.h"


// =================================== Defines =========================================



// ================================== Typedefs =========================================

typedef enum
{
    OnePageViewMode		= 0,
    TwoPageViewMode		= 1
} S4BookViewMode;


typedef enum
{
    LeftPageImage		= 0,
    LeftPageRevImage	= 1,
	RightPageImage		= 2,
	RightPageRevImage	= 3
} PageImageType;


typedef enum
{
    TouchModeInactive	= 0,
    TouchModePageTurn	= 1,
	TouchModeTapInPage	= 2
} TouchMode;


// ================================== Globals =========================================



// ============================= Forward Declarations =================================

@class S4BookView;


// ================================== Protocols ========================================

@protocol S4BookViewDelegate <NSObject>

@optional
// called when the user touches up on the left or right side of the page, or finishes dragging the page
- (void)bookview: (S4BookView *)theView willTurnToPageAtIndex: (NSUInteger)pageIndex;

// called when the page-turn animation (following a touch-up or drag) completes 
- (void)bookview: (S4BookView *)theView didTurnToPageAtIndex: (NSUInteger)pageIndex;
- (void)bookview: (S4BookView *)theView zoomingToScale: (CGFloat)scale;
- (void)bookview: (S4BookView *)theView doubleTapCurrentView: (CGFloat)fooVariable;

@end


// =============================== S4BookView Class ====================================

@interface S4BookView : UIView <UIGestureRecognizerDelegate>
{
@protected
	CALayer											*m_topPageLayer;
	CALayer											*m_topPageOverLayer;
	CAGradientLayer									*m_topPageShadowLayer;
	CALayer											*m_topPageRevLayer;
	CALayer											*m_topPageRevImgLayer;
	CALayer											*m_topPageRevOverLayer;
	CAGradientLayer									*m_topPageRevShadeLayer;
	CALayer											*m_bottomPageLayer;
	CAGradientLayer									*m_bottomPageShadowLayer;
    CALayer											*m_leftPageLayer;
	CALayer											*m_leftPageOverLayer;
    S4BookViewMode									mode;
	CGFloat											m_leafEdge;
	NSUInteger										m_currentPageIndex;
    NSUInteger										numberOfVisiblePages;
	id <S4BookViewDelegate>							m_delegate;
	CGSize											m_pageSize;
	S4BookCache										*m_bookCache;
	CGFloat											preferredTargetWidth;
	CGPoint											m_touchBeganPoint;
	TouchMode										m_touchMode;
	CGRect											m_nextPageTouchRect;
	CGRect											m_prevPageTouchRect;
	BOOL											interactionLocked;
	BOOL											zoomActive;
	BOOL											panActive;
}

// Properties
@property (nonatomic, assign) id <S4BookViewDelegate>					delegate;
@property (nonatomic, assign) S4BookViewMode							mode;
// the automatically determined width of the interactive areas on either side of the page
@property (assign, readonly) CGFloat									targetWidth;
// set this to a nonzero value to get a targetWidth other than the default
@property (nonatomic, assign) CGFloat									preferredTargetWidth;
// the zero-based index of the page currently being displayed.
@property (nonatomic, assign) NSUInteger								currentPageIndex;
@property (nonatomic, assign, readonly) BOOL							zoomActive;
@property (readonly, assign, atomic) TouchMode							touchMode;
@property (readwrite, assign, nonatomic) UIViewController				*myViewController;

// Class methods

// Instance methods
- (id)initWithFrame: (CGRect)frame;
- (void)reloadData;
- (void)reloadDataWithPageIndex: (NSUInteger)newPageIndex;
- (void)turnToNextPageWithDuration: (NSTimeInterval)duration;
- (void)turnToPrevPageWithDuration: (NSTimeInterval)duration;
- (void)adjustAnchorPointForGestureRecognizer: (UIPinchGestureRecognizer *)gestureRecognizer;
- (void)doubleTap: (UIGestureRecognizer *)gestureRecognizer;

// instance method subclasses MUST override to provide page count
- (NSUInteger)documentPageCount;
// instance method subclasses MUST override to provide image source
- (CGImageRef)getImageForPageIndex: (NSUInteger)pageIndex forPageType: (PageImageType)pageType;
// instance method subclasses MAY override to handle taps within pages
- (void)handleTouchPointInPage: (CGPoint)touchPoint dragDistance: (CGFloat)delta;

@end
