/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the S4 iOS Libraries.
 *
 * The Initial Developer of the Original Code is
 * Michael Papp dba SeaStones Software Company.
 * All software created by the Initial Developer are Copyright (C) 2008-2011
 * the Initial Developer. All Rights Reserved.
 *
 * Original Author:
 *			Michael Papp, San Francisco, CA, USA
 *
 * ***** END LICENSE BLOCK ***** */

/* ***** FILE BLOCK ******
 *
 * Name:		S4BookView.m
 * Module:		UI
 * Library:		S4 iOS_UI Libraries
 *
 * ***** FILE BLOCK *****/


// ================================== Includes =========================================

#import "S4BookView.h"
#import "S4BookTileLayer.h"
#import "S4CommonDefines.h"


// ================================== Defines ==========================================



// ================================== Typedefs ==========================================



// ================================== Globals ==========================================



// ============================= Forward Declarations ==================================

CGFloat distance(CGPoint a, CGPoint b);


// ================================== Inlines ==========================================

CGFloat distance(CGPoint a, CGPoint b)
{
	return sqrtf(powf(a.x-b.x, 2) + powf(a.y-b.y, 2));
}



// ============================ Begin Class S4BookView () ==============================

@interface S4BookView () 

@property (readwrite, assign, atomic) TouchMode							touchMode;

@end



// ====================== Begin Class S4BookView (PrivateImpl) =========================

@interface S4BookView (PrivateImpl)

- (void)initS4BookView;
- (void)callWillTurnToPageAtIndex: (NSUInteger)index;
- (void)callDidTurnToPageAtIndex: (NSUInteger)index;
- (void)callZoomingToScale: (CGFloat)zoomLevel;
- (void)callDoubleTapCurrentView: (CGFloat)zoomLevel;
- (void)setLeafEdge: (CGFloat)newLeafEdge;
- (void)setLayerFrames;
- (BOOL)hasNextPage;
- (BOOL)hasPrevPage;
- (BOOL)touchedNextPage;
- (BOOL)touchedPrevPage;
- (CGFloat)dragThreshold;
- (void)getImages;
- (void)updateTargetRects;
- (void)panMove: (UIPanGestureRecognizer *)gestureRecognizer;

@end



@implementation S4BookView (PrivateImpl)

//============================================================================
//	S4BookView (PrivateImpl) :: initS4BookView
//============================================================================
- (void)initS4BookView
{
	CGFloat									scaleValue;
	UIPinchGestureRecognizer				*pinchGesture;
	UITapGestureRecognizer					*tapGesture;

	scaleValue = [[UIScreen mainScreen] scale];
	self.clipsToBounds = YES;
	
	m_topPageLayer = [[CALayer alloc] init];
	m_topPageLayer.masksToBounds = YES;
	m_topPageLayer.contentsGravity = kCAGravityLeft;
	m_topPageLayer.backgroundColor = [[UIColor whiteColor] CGColor];
	m_topPageLayer.contentsScale = scaleValue;
	
	m_topPageOverLayer = [[CALayer alloc] init];
	m_topPageOverLayer.backgroundColor = [[[UIColor blackColor] colorWithAlphaComponent:0.2] CGColor];
	m_topPageOverLayer.contentsScale = scaleValue;
	
	m_topPageShadowLayer = [[CAGradientLayer alloc] init];
	m_topPageShadowLayer.colors = [NSArray arrayWithObjects: (id)[[[UIColor blackColor] colorWithAlphaComponent: 0.6] CGColor],
								   (id)[[UIColor clearColor] CGColor],
								   nil];
	m_topPageShadowLayer.startPoint = CGPointMake(1,0.5);
	m_topPageShadowLayer.endPoint = CGPointMake(0,0.5);
	m_topPageShadowLayer.contentsScale = scaleValue;
	
	m_topPageRevLayer = [[CALayer alloc] init];
	m_topPageRevLayer.backgroundColor = [[UIColor whiteColor] CGColor];
	m_topPageRevLayer.masksToBounds = YES;
	m_topPageRevLayer.contentsScale = scaleValue;
	
	m_topPageRevImgLayer = [[CALayer alloc] init];
	m_topPageRevImgLayer.masksToBounds = YES;
	m_topPageRevImgLayer.contentsScale = scaleValue;
	
	m_topPageRevOverLayer = [[CALayer alloc] init];
	m_topPageRevOverLayer.backgroundColor = [[[UIColor whiteColor] colorWithAlphaComponent:0.8] CGColor];
	m_topPageRevOverLayer.contentsScale = scaleValue;
	
	m_topPageRevShadeLayer = [[CAGradientLayer alloc] init];
	m_topPageRevShadeLayer.colors = [NSArray arrayWithObjects:
									(id)[[[UIColor blackColor] colorWithAlphaComponent:0.6] CGColor],
									(id)[[UIColor clearColor] CGColor],
									nil];
	m_topPageRevShadeLayer.startPoint = CGPointMake(1,0.5);
	m_topPageRevShadeLayer.endPoint = CGPointMake(0,0.5);
	m_topPageRevShadeLayer.contentsScale = scaleValue;
	
	m_bottomPageLayer = [[CALayer alloc] init];
	m_bottomPageLayer.backgroundColor = [[UIColor whiteColor] CGColor];
	m_bottomPageLayer.masksToBounds = YES;
	m_bottomPageLayer.contentsScale = scaleValue;
	
	m_bottomPageShadowLayer = [[CAGradientLayer alloc] init];
	m_bottomPageShadowLayer.colors = [NSArray arrayWithObjects:
							   (id)[[[UIColor blackColor] colorWithAlphaComponent:0.6] CGColor],
							   (id)[[UIColor clearColor] CGColor],
							   nil];
	m_bottomPageShadowLayer.startPoint = CGPointMake(0,0.5);
	m_bottomPageShadowLayer.endPoint = CGPointMake(1,0.5);
	m_bottomPageShadowLayer.contentsScale = scaleValue;
	
	[m_topPageLayer addSublayer: m_topPageOverLayer];
	[m_topPageRevLayer addSublayer: m_topPageRevImgLayer];
	[m_topPageRevLayer addSublayer: m_topPageRevOverLayer];
	[m_topPageRevLayer addSublayer: m_topPageRevShadeLayer];
	[m_bottomPageLayer addSublayer: m_bottomPageShadowLayer];
	
    // Setup for the left page in two-page mode
    m_leftPageLayer = [[CALayer alloc] init];
	m_leftPageLayer.masksToBounds = YES;
	m_leftPageLayer.contentsGravity = kCAGravityLeft;
	m_leftPageLayer.backgroundColor = [[UIColor whiteColor] CGColor];
	m_leftPageLayer.contentsScale = scaleValue;
	
	m_leftPageOverLayer = [[CALayer alloc] init];
	m_leftPageOverLayer.backgroundColor = [[[UIColor blackColor] colorWithAlphaComponent:0.2] CGColor];
	m_leftPageOverLayer.contentsScale = scaleValue;
	[m_leftPageLayer addSublayer: m_leftPageOverLayer];
	
	[self.layer addSublayer: m_leftPageLayer];
	[self.layer addSublayer: m_bottomPageLayer];
	[self.layer addSublayer: m_topPageLayer];
	[self.layer addSublayer: m_topPageRevLayer];
    [self.layer addSublayer: m_topPageShadowLayer];
    self.layer.contentsScale = scaleValue;
    
	m_topPageRevImgLayer.contentsGravity = kCAGravityRight;
	m_topPageRevOverLayer.backgroundColor = [[[UIColor whiteColor] colorWithAlphaComponent: 0.8] CGColor];
	m_topPageRevImgLayer.transform = CATransform3DMakeScale(-1, 1, 1);

	[self setLeafEdge: (CGFloat)1.0];
	
	pinchGesture = [[UIPinchGestureRecognizer alloc] initWithTarget: self action: @selector(pinchZoom:)];
    [pinchGesture setDelegate: self];
    [self addGestureRecognizer: pinchGesture];
    [pinchGesture release];

	tapGesture = [[UITapGestureRecognizer alloc] initWithTarget: self action: @selector(doubleTap:)];
    [tapGesture setNumberOfTapsRequired: 2];
	[tapGesture setDelegate: self];
    [self addGestureRecognizer: tapGesture];
    [tapGesture release];

	mode = OnePageViewMode;
	m_currentPageIndex = 0;
    numberOfVisiblePages = 1;
	m_touchMode = TouchModeInactive;
	zoomActive = NO;
	panActive = NO;
	m_delegate = nil;
    m_pageSize = self.bounds.size;
	m_bookCache = [[S4BookCache alloc] init];
}


//============================================================================
//	S4BookView (PrivateImpl) :: callWillTurnToPageAtIndex:
//============================================================================
- (void)callWillTurnToPageAtIndex: (NSUInteger)index
{
	if ((nil != m_delegate) && (YES == [m_delegate respondsToSelector: @selector(bookview:willTurnToPageAtIndex:)]))
	{
		[m_delegate bookview: self willTurnToPageAtIndex: index];
	}
}


//============================================================================
//	S4BookView (PrivateImpl) :: callDidTurnToPageAtIndex:
//============================================================================
- (void)callDidTurnToPageAtIndex: (NSUInteger)index
{
	if ((nil != m_delegate) && (YES == [m_delegate respondsToSelector: @selector(bookview:didTurnToPageAtIndex:)]))
	{
		[m_delegate bookview: self didTurnToPageAtIndex: index];
	}
}


//============================================================================
//	S4BookView (PrivateImpl) :: callZoomingToScale:
//============================================================================
- (void)callZoomingToScale: (CGFloat)scale
{
	if ((nil != m_delegate) && (YES == [m_delegate respondsToSelector: @selector(bookview:zoomingToScale:)]))
	{
		[m_delegate bookview: self zoomingToScale: scale];
	}
}


//============================================================================
//	S4BookView (PrivateImpl) :: callDoubleTapCurrentView:
//============================================================================
- (void)callDoubleTapCurrentView: (CGFloat)zoomLevel
{
	if ((nil != m_delegate) && (YES == [m_delegate respondsToSelector: @selector(bookview:doubleTapCurrentView:)]))
	{
		[m_delegate bookview: self doubleTapCurrentView: zoomLevel];
	}
}


//============================================================================
//	S4BookView (PrivateImpl) :: setLeafEdge:
//============================================================================
- (void)setLeafEdge: (CGFloat)newLeafEdge
{
	m_leafEdge = newLeafEdge;
	
    CGFloat pageOpacity = MIN(1.0, (4 * (1 - m_leafEdge)));
    
    m_topPageShadowLayer.opacity = pageOpacity;
	m_bottomPageShadowLayer.opacity = pageOpacity;
	m_topPageOverLayer.opacity = pageOpacity;
	m_leftPageOverLayer.opacity = pageOpacity;
	
	[self setLayerFrames];
}


//============================================================================
//	S4BookView (PrivateImpl) :: setLayerFrames
//============================================================================
- (void)setLayerFrames
{
	CGRect						rightPageBoundsRect;
	CGRect						leftHalf, rightHalf;

	rightPageBoundsRect = self.layer.bounds;
    CGRectDivide(rightPageBoundsRect, &leftHalf, &rightHalf, CGRectGetWidth(rightPageBoundsRect) / 2.0f, CGRectMinXEdge);
    if (self.mode == TwoPageViewMode)
	{
        rightPageBoundsRect = rightHalf;
    }

	m_topPageLayer.frame = CGRectMake(rightPageBoundsRect.origin.x, 
									  rightPageBoundsRect.origin.y, 
									  m_leafEdge * rightPageBoundsRect.size.width, 
									  rightPageBoundsRect.size.height);
	m_topPageRevLayer.frame = CGRectMake(rightPageBoundsRect.origin.x + (2*m_leafEdge-1) * rightPageBoundsRect.size.width, 
											 rightPageBoundsRect.origin.y, 
											 (1-m_leafEdge) * rightPageBoundsRect.size.width, 
											 rightPageBoundsRect.size.height);
	m_bottomPageLayer.frame = rightPageBoundsRect;
	m_topPageShadowLayer.frame = CGRectMake(m_topPageRevLayer.frame.origin.x - 40, 0, 40, m_bottomPageLayer.bounds.size.height);
	m_topPageRevImgLayer.frame = m_topPageRevLayer.bounds;
//	m_topPageRevImgLayer.transform = CATransform3DMakeScale(-1, 1, 1);
	m_topPageRevOverLayer.frame = m_topPageRevLayer.bounds;
	m_topPageRevShadeLayer.frame = CGRectMake(m_topPageRevLayer.bounds.size.width - 50, 
											 0, 
											 50 + 1, 
											 m_topPageRevLayer.bounds.size.height);
	m_bottomPageShadowLayer.frame = CGRectMake(m_leafEdge * rightPageBoundsRect.size.width, 
										0, 
										40, 
										m_bottomPageLayer.bounds.size.height);
	m_topPageOverLayer.frame = m_topPageLayer.bounds;

    if (self.mode == OnePageViewMode)
	{
        m_leftPageLayer.hidden = YES;
        m_leftPageOverLayer.hidden = m_leftPageLayer.hidden;
    }
	else
	{
        m_leftPageLayer.hidden = NO;
        m_leftPageOverLayer.hidden = m_leftPageLayer.hidden;
        m_leftPageLayer.frame = CGRectMake(leftHalf.origin.x, leftHalf.origin.y, leftHalf.size.width, leftHalf.size.height);
        m_leftPageOverLayer.frame = m_leftPageLayer.bounds;
    }
}


//============================================================================
//	S4BookView (PrivateImpl) :: hasNextPage
//============================================================================
- (BOOL)hasNextPage
{
	NSUInteger					numberOfPages;
	
	numberOfPages = [self documentPageCount];
	if (self.mode == OnePageViewMode)
	{
		return self.currentPageIndex < numberOfPages - 1;
	}
	else
	{
		return  ((self.currentPageIndex % 2 == 0) && (self.currentPageIndex < numberOfPages - 1)) ||
		((self.currentPageIndex % 2 != 0) && (self.currentPageIndex < numberOfPages - 2));
	}
}


//============================================================================
//	S4BookView (PrivateImpl) :: hasPrevPage
//============================================================================
- (BOOL)hasPrevPage
{
    return self.currentPageIndex > (numberOfVisiblePages - 1);
}


//============================================================================
//	S4BookView (PrivateImpl) :: touchedNextPage
//============================================================================
- (BOOL)touchedNextPage
{
	return CGRectContainsPoint(m_nextPageTouchRect, m_touchBeganPoint);
}


//============================================================================
//	S4BookView (PrivateImpl) :: touchedPrevPage
//============================================================================
- (BOOL)touchedPrevPage
{
	return CGRectContainsPoint(m_prevPageTouchRect, m_touchBeganPoint);
}


//============================================================================
//	S4BookView (PrivateImpl) :: dragThreshold
//============================================================================
- (CGFloat)dragThreshold
{
	return (10);				// Magic empirical number
}


//============================================================================
//	S4BookView (PrivateImpl) :: getImages
//============================================================================
- (void)getImages
{
	NSUInteger					numberOfPages;
	CGImageRef					curImageRef;

	numberOfPages = [self documentPageCount];
    if (OnePageViewMode == self.mode) 
	{
		if (self.currentPageIndex < numberOfPages)
		{
			curImageRef = [self getImageForPageIndex: self.currentPageIndex forPageType: RightPageImage];
			m_topPageLayer.contents = (id)curImageRef;
            m_leftPageLayer.contents = (id)curImageRef;
			if (self.currentPageIndex < numberOfPages - 1)
			{
				m_topPageRevImgLayer.contents = (id)curImageRef;
				m_bottomPageLayer.contents = (id)[self getImageForPageIndex: (self.currentPageIndex + 1) forPageType: RightPageRevImage];
			}
		}
		else
		{
			m_topPageLayer.contents = nil;
			m_topPageRevImgLayer.contents = nil;
			m_bottomPageLayer.contents = nil;
            m_leftPageLayer.contents = nil;
		}
	}
	else
	{
		if (self.currentPageIndex <= numberOfPages)
		{
			m_topPageLayer.contents = (id)[self getImageForPageIndex: self.currentPageIndex forPageType: RightPageImage];
			if (self.currentPageIndex > 0)
			{
				m_leftPageLayer.contents = (id)[self getImageForPageIndex: (self.currentPageIndex - 1) forPageType: LeftPageImage];
			}
			else
			{
				m_leftPageLayer.contents = nil;
			}
			
			if (self.currentPageIndex < numberOfPages - 1)
			{
				m_topPageRevImgLayer.contents = (id)[self getImageForPageIndex: (self.currentPageIndex + 1) forPageType: RightPageRevImage];
				m_bottomPageLayer.contents = (id)[self getImageForPageIndex: (self.currentPageIndex + 2) forPageType: LeftPageRevImage];
			}
		}
		else
		{
			m_topPageLayer.contents = nil;
			m_topPageRevImgLayer.contents = nil;
			m_bottomPageLayer.contents = nil;
			m_leftPageLayer.contents = nil;
		}
	}
}


//============================================================================
//	S4BookView (PrivateImpl) :: updateTargetRects
//============================================================================
- (void)updateTargetRects
{
	CGFloat targetWidth = [self targetWidth];
	m_nextPageTouchRect = CGRectMake(self.bounds.size.width - targetWidth, 0, targetWidth, self.bounds.size.height);
	m_prevPageTouchRect = CGRectMake(0, 0, targetWidth, self.bounds.size.height);
}


//============================================================================
//	S4BookView :: panMove:
//		This method handles pan and move gestures
//============================================================================
- (void)panMove: (UIPanGestureRecognizer *)gestureRecognizer
{
    if ([gestureRecognizer state] == UIGestureRecognizerStateBegan || [gestureRecognizer state] == UIGestureRecognizerStateChanged)
	{
        CGPoint translation = [gestureRecognizer translationInView:[[gestureRecognizer view] superview]];
        [[gestureRecognizer view] setCenter: CGPointMake([[gestureRecognizer view] center].x + translation.x, [[gestureRecognizer view] center].y + translation.y)];
        [gestureRecognizer setTranslation: CGPointZero inView: [[gestureRecognizer view] superview]];
	}
}

@end



// =========================== Begin Class S4BookView ===============================

@implementation S4BookView


//============================================================================
//	S4BookView synthesize properties
//============================================================================
@synthesize delegate = m_delegate;
@dynamic mode;
@dynamic currentPageIndex;
@synthesize preferredTargetWidth;
@synthesize zoomActive;
@synthesize touchMode = m_touchMode;
@synthesize myViewController = m_myViewController;


//============================================================================
//	S4BookView :: layerClass
//============================================================================
+ (Class)layerClass
{
	return [S4BookTileLayer class];
}


//============================================================================
//	S4BookView :: init
//============================================================================
- (id)init
{
    return ([self initWithFrame: CGRectZero]);
}


//============================================================================
//	S4BookView :: initWithFrame:
//		Designated initializer
//============================================================================
- (id)initWithFrame: (CGRect)frame
{
	self = [super initWithFrame: frame];
    if (nil != self)
	{
		[self initS4BookView];
    }
    return self;
}


//============================================================================
//	S4BookView :: awakeFromNib
//============================================================================
- (void)awakeFromNib
{
	[super awakeFromNib];
	[self initS4BookView];
}


//============================================================================
//	S4BookView :: dealloc
//============================================================================
- (void)dealloc
{
	NS_SAFE_RELEASE(m_topPageLayer);
	NS_SAFE_RELEASE(m_topPageShadowLayer);
	NS_SAFE_RELEASE(m_topPageOverLayer);
	NS_SAFE_RELEASE(m_topPageRevLayer);
	NS_SAFE_RELEASE(m_topPageRevImgLayer);
	NS_SAFE_RELEASE(m_topPageRevOverLayer);
	NS_SAFE_RELEASE(m_topPageRevShadeLayer);
	NS_SAFE_RELEASE(m_bottomPageLayer);
	NS_SAFE_RELEASE(m_bottomPageShadowLayer);
	NS_SAFE_RELEASE(m_leftPageLayer);
	NS_SAFE_RELEASE(m_leftPageOverLayer);
	NS_SAFE_RELEASE(m_bookCache);

    [super dealloc];
}



#pragma mark UIView methods

//============================================================================
//	S4BookView :: layoutSubviews
//============================================================================
- (void)layoutSubviews
{
	CGSize					desiredPageSize;
	CGFloat					touchRectsWidth;
	
	[super layoutSubviews];
	
	desiredPageSize = self.bounds.size;
    if (self.mode == TwoPageViewMode)
	{
        desiredPageSize = CGSizeMake(self.bounds.size.width/2.0f, self.bounds.size.height);
    }
	
	if (false == CGSizeEqualToSize(m_pageSize, desiredPageSize))
	{
		m_pageSize = desiredPageSize;
		
		[CATransaction begin];
		[CATransaction setValue: (id)kCFBooleanTrue forKey: kCATransactionDisableActions];
		[self setLayerFrames];
		[CATransaction commit];
		[self getImages];
		
		//		self.currentPageIndex = 0;
		//		[self updateTargetRects];
		
		touchRectsWidth = self.bounds.size.width / 7;
		m_nextPageTouchRect = CGRectMake(self.bounds.size.width - touchRectsWidth,
										 0,
										 touchRectsWidth,
										 self.bounds.size.height);
		m_prevPageTouchRect = CGRectMake(0,
										 0,
										 touchRectsWidth,
										 self.bounds.size.height);
	}
}


//============================================================================
//	S4BookView :: reloadData
//============================================================================
- (void)reloadData
{
	[self reloadDataWithPageIndex: 0];
}


//============================================================================
//	S4BookView :: reloadDataWithPageIndex:
//============================================================================
- (void)reloadDataWithPageIndex: (NSUInteger)newPageIndex
{
	[m_bookCache flushBookCache];
	self.currentPageIndex = newPageIndex;
}


//============================================================================
//	S4BookView :: turnToNextPageWithDuration:
//============================================================================
- (void)turnToNextPageWithDuration: (NSTimeInterval)duration
{
	NSUInteger					pageIndex;
	
	pageIndex = self.currentPageIndex + 1;
	if ([self hasNextPage])
	{
		[CATransaction begin];
		[CATransaction setValue: [NSNumber numberWithFloat: duration] forKey: kCATransactionAnimationDuration];
		[self callWillTurnToPageAtIndex: pageIndex];
		[self setLeafEdge: (CGFloat)0.0];
		interactionLocked = YES;
		[self performSelector: @selector(didTurnPageForward) withObject: nil afterDelay: duration + 0.25];
		[CATransaction commit];
	}
}


//============================================================================
//	S4BookView :: turnToPrevPageWithDuration:
//============================================================================
- (void)turnToPrevPageWithDuration: (NSTimeInterval)duration
{
	NSUInteger						pageIndex;
	
	pageIndex = self.currentPageIndex - 1;
	if ([self hasPrevPage])
	{
		[CATransaction begin];
		[CATransaction setValue: (id)kCFBooleanTrue forKey: kCATransactionDisableActions];
		self.currentPageIndex = pageIndex;
		[self setLeafEdge: (CGFloat)0.0];
		[CATransaction commit];
		[self performSelector: @selector(completeBackwardsPageTurn:) withObject: [NSNumber numberWithFloat: duration] afterDelay: 0.01];
	}
}


//============================================================================
//	S4BookView :: documentPageCount
//============================================================================
- (NSUInteger)documentPageCount
{
	return (0);
}


//============================================================================
//	S4BookView :: getImageForPageIndex:
//============================================================================
- (CGImageRef)getImageForPageIndex: (NSUInteger)pageIndex forPageType: (PageImageType)pageType
{
	return (NULL);
}



#pragma mark dynamic Setters and Getters

//============================================================================
//	S4BookView :: mode
//============================================================================
- (S4BookViewMode)mode
{
	return (mode);
}


//============================================================================
//	S4BookView :: setMode:
//============================================================================
- (void)setMode: (S4BookViewMode)newMode
{
	NSUInteger					numberOfPages;
	
	numberOfPages = [self documentPageCount];
    mode = newMode;
    if (OnePageViewMode == mode)
	{
        numberOfVisiblePages = 1;
        if (self.currentPageIndex > numberOfPages - 1)
		{
            self.currentPageIndex = numberOfPages - 1;
        }
		m_topPageRevImgLayer.contentsGravity = kCAGravityRight;
        m_topPageRevOverLayer.backgroundColor = [[[UIColor whiteColor] colorWithAlphaComponent: 0.8] CGColor];
        m_topPageRevImgLayer.transform = CATransform3DMakeScale(-1, 1, 1);
    }
	else
	{
        numberOfVisiblePages = 2;
        if (self.currentPageIndex % 2 != 0)
		{
            self.currentPageIndex = self.currentPageIndex + 1;
        }
		m_topPageRevImgLayer.contentsGravity = kCAGravityLeft;
        m_topPageRevOverLayer.backgroundColor = [[[UIColor whiteColor] colorWithAlphaComponent: 0.0] CGColor];
        m_topPageRevImgLayer.transform = CATransform3DMakeScale(1, 1, 1);
    }
	[self setNeedsLayout];
}


//============================================================================
//	S4BookView :: targetWidth
//============================================================================
- (CGFloat)targetWidth
{
	// Magic empirical formula
	if (preferredTargetWidth > 0 && preferredTargetWidth < self.bounds.size.width / 2)
	{
		return preferredTargetWidth;
	}
	return MAX(28, self.bounds.size.width / 5);
}


//============================================================================
//	S4BookView :: preferredTargetWidth
//============================================================================
- (CGFloat)preferredTargetWidth
{
	return (preferredTargetWidth);
}


//============================================================================
//	S4BookView :: setPreferredTargetWidth:
//============================================================================
- (void)setPreferredTargetWidth: (CGFloat)value
{
	preferredTargetWidth = value;
	[self updateTargetRects];
}


//============================================================================
//	S4BookView :: currentPageIndex
//============================================================================
- (NSUInteger)currentPageIndex
{
	return (m_currentPageIndex);
}


//============================================================================
//	S4BookView :: setCurrentPageIndex:
//============================================================================
- (void)setCurrentPageIndex: (NSUInteger)newCurrentPageIndex
{
	// Remember that currentPageIndex start from 0!
	if (newCurrentPageIndex <= [self documentPageCount])
	{
		if ((self.mode == TwoPageViewMode) && (newCurrentPageIndex % 2 != 0))
		{
			m_currentPageIndex = newCurrentPageIndex + 1;
		}
		else
		{
			m_currentPageIndex = newCurrentPageIndex;
		}
		
		[CATransaction begin];
		[CATransaction setValue: (id)kCFBooleanTrue forKey: kCATransactionDisableActions];
		[self getImages];
		[self setLeafEdge: (CGFloat)1.0];
		[CATransaction commit];
	}
}


//============================================================================
//	S4BookView :: completeBackwardsPageTurn:
//============================================================================
- (void)completeBackwardsPageTurn: (NSNumber *)duration
{
	[CATransaction begin];
	[CATransaction setValue: duration forKey: kCATransactionAnimationDuration];
	[self callWillTurnToPageAtIndex: self.currentPageIndex];
	[self setLeafEdge: (CGFloat)1.0];
	interactionLocked = YES;
	[self performSelector:@selector(didTurnPageBackward)
			   withObject:nil 
			   afterDelay:[duration floatValue] + 0.25];
	
	[CATransaction commit];
}


//============================================================================
//	S4BookView :: didTurnPageBackward
//============================================================================
- (void)didTurnPageBackward
{
	interactionLocked = NO;
	[self callDidTurnToPageAtIndex: self.currentPageIndex];
}


//============================================================================
//	S4BookView :: didTurnPageForward
//============================================================================
- (void)didTurnPageForward
{
	interactionLocked = NO;
	if ([self documentPageCount] <= (self.currentPageIndex + numberOfVisiblePages))
	{
		self.currentPageIndex = [self documentPageCount] - 1;
	}
	else
	{
		self.currentPageIndex = self.currentPageIndex + numberOfVisiblePages;
	}
	[self callDidTurnToPageAtIndex: self.currentPageIndex];
}



#pragma mark UIResponder methods

//============================================================================
//	S4BookView :: touchesBegan:
//============================================================================
- (void)touchesBegan: (NSSet *)touches withEvent: (UIEvent *)event
{
	UITouch				*touch;

	if ((NO == interactionLocked) && (NO == zoomActive))
	{
		touch = [event.allTouches anyObject];
		m_touchBeganPoint = [touch locationInView: self];

		if ([self touchedPrevPage] && [self hasPrevPage])
		{		
			[CATransaction begin];
			[CATransaction setValue: (id)kCFBooleanTrue forKey: kCATransactionDisableActions];
			self.currentPageIndex = self.currentPageIndex - numberOfVisiblePages;
			[self setLeafEdge: (CGFloat)0.0];
			[CATransaction commit];
			self.touchMode = TouchModePageTurn;
		}
		else if ([self touchedNextPage] && [self hasNextPage])
		{
			self.touchMode = TouchModePageTurn;
		}
		else
		{
			self.touchMode = TouchModeTapInPage;
		}
	}
}


//============================================================================
//	S4BookView :: touchesMoved:
//============================================================================
- (void)touchesMoved: (NSSet *)touches withEvent: (UIEvent *)event
{
	UITouch					*touch;
	CGPoint					touchPoint;

	if (TouchModePageTurn == self.touchMode)
	{
		touch = [event.allTouches anyObject];
		touchPoint = [touch locationInView: self];

		[CATransaction begin];
		[CATransaction setValue: [NSNumber numberWithFloat: 0.07] forKey: kCATransactionAnimationDuration];
		[self setLeafEdge: (CGFloat)(touchPoint.x / self.bounds.size.width)];
		[CATransaction commit];
	}
}


//============================================================================
//	S4BookView :: touchesEnded:
//============================================================================
- (void)touchesEnded: (NSSet *)touches withEvent: (UIEvent *)event
{
	UITouch						*touch;
	CGPoint						touchPoint;
	CGFloat						dragged;
	float						duration;

	if (TouchModeInactive != self.touchMode)
	{
		touch = [event.allTouches anyObject];
		touchPoint = [touch locationInView: self];
		dragged = distance(touchPoint, m_touchBeganPoint) > [self dragThreshold];

		if (TouchModePageTurn == self.touchMode)
		{
			[CATransaction begin];
			if (((0 < dragged) && (m_leafEdge < 0.5)) || ((0 == dragged) && [self touchedNextPage]))
			{
				[self callWillTurnToPageAtIndex: self.currentPageIndex + numberOfVisiblePages];
				[self setLeafEdge: (CGFloat)0.0];
				duration = m_leafEdge;
				interactionLocked = YES;
				[self performSelector: @selector(didTurnPageForward) withObject: nil afterDelay: duration + 0.25];
			}
			else
			{
				[self callWillTurnToPageAtIndex: self.currentPageIndex];
				[self setLeafEdge: (CGFloat)1.0];
				duration = 1 - m_leafEdge;
				interactionLocked = YES;
				[self performSelector: @selector(didTurnPageBackward) withObject: nil  afterDelay: (duration + 0.25)];
			}
			[CATransaction setValue: [NSNumber numberWithFloat: duration] forKey: kCATransactionAnimationDuration];
			[CATransaction commit];
		}
		else if (TouchModeTapInPage == self.touchMode)
		{
			[self handleTouchPointInPage: touchPoint dragDistance: dragged];
		}
		self.touchMode = TouchModeInactive;
	}
}


//============================================================================
//	S4BookView :: handleTouchPointInPage:
//============================================================================
- (void)handleTouchPointInPage: (CGPoint)touchPoint dragDistance: (CGFloat)delta
{
}



#pragma mark UIGestureRecognizer methods

//============================================================================
//	S4BookView :: adjustAnchorPointForGestureRecognizer:
//		this method will adjust the anchor point of the gesture recognizer
//		in order for it to zoom towards the direction chosen by the user
//============================================================================
- (void)adjustAnchorPointForGestureRecognizer: (UIPinchGestureRecognizer *)gestureRecognizer
{
	UIView						*gestureView;
	CGPoint						ptInView;
	CGPoint						ptInSuperview;

	if (gestureRecognizer.state == UIGestureRecognizerStateBegan)
	{
		gestureView = gestureRecognizer.view;
		ptInView = [gestureRecognizer locationInView: gestureView];
		ptInSuperview = [gestureRecognizer locationInView: gestureView.superview];

		gestureView.layer.anchorPoint = CGPointMake(ptInView.x / gestureView.bounds.size.width, ptInView.y / gestureView.bounds.size.height);
		gestureView.center = ptInSuperview;
    }
}


//============================================================================
//	S4BookView :: pinchZoom:
//		This method handles pinch and zoom gestures
//============================================================================
- (void)pinchZoom: (UIPinchGestureRecognizer *)gestureRecognizer
{
	UIPanGestureRecognizer					*panGesture;
	CGFloat									scale;

	if (NO == interactionLocked)
	{
		[self adjustAnchorPointForGestureRecognizer: gestureRecognizer];		// directing the zoom in the right direction

		if (([gestureRecognizer state] == UIGestureRecognizerStateBegan) || ([gestureRecognizer state] == UIGestureRecognizerStateChanged))
		{
			if (YES == zoomActive)
			{
				panGesture = [[UIPanGestureRecognizer alloc] initWithTarget: self action: @selector(panMove:)];
				[panGesture setMaximumNumberOfTouches: 2];
				[panGesture setDelegate: self];
				[self addGestureRecognizer: panGesture];
				[panGesture release];
			}

			if (gestureRecognizer.state == UIGestureRecognizerStateBegan)
			{
				gestureRecognizer.scale = gestureRecognizer.view.transform.a;
			}
			else
			{
				scale = MAX(gestureRecognizer.scale, 1.0f);// minimum zoom scale is 1:1
				gestureRecognizer.view.transform = CGAffineTransformMakeScale(scale, scale);
				if (1.0 == scale)
				{
					zoomActive = NO;
					[self doubleTap: gestureRecognizer];
				}
				else
				{
					zoomActive = YES;
				}
			}
			// call the delegate
			[self callZoomingToScale: [gestureRecognizer scale]];
		}
	}
}


//============================================================================
//	S4BookView :: doubleTap:
//		This method handles the double tap gesture,
//		and will reposition the view at 1:1 scale
//============================================================================
- (void)doubleTap: (UIGestureRecognizer *)gestureRecognizer
{
	UIView					*gestureView;
	NSArray					*registeredGestures;

	//restore center anchorpoint
	self.layer.anchorPoint = CGPointMake(0.5, 0.5);

	gestureView = [gestureRecognizer view];
	gestureView.transform = CGAffineTransformIdentity;
	[gestureView setCenter: CGPointMake(gestureView.frame.size.width / 2, gestureView.frame.size.height / 2)];

	zoomActive = NO;
	panActive = NO;

	registeredGestures = self.gestureRecognizers;
	for (UIGestureRecognizer *gesture in registeredGestures)
	{
		if ([gesture isKindOfClass: [UIPanGestureRecognizer class]])
		{
			// Let remove the PAN / MOVE gesture recognizer
			[self removeGestureRecognizer: gesture];
		}
	}
	// call the delegate
	[self callDoubleTapCurrentView: 0.0f];
}

@end
