/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the S4 iOS Libraries.
 *
 * The Initial Developer of the Original Code is
 * Michael Papp dba SeaStones Software Company.
 * All software created by the Initial Developer are Copyright (C) 2008-2011
 * the Initial Developer. All Rights Reserved.
 *
 * Original Author:
 *			Michael Papp, San Francisco, CA, USA
 *
 * ***** END LICENSE BLOCK ***** */

/* ***** FILE BLOCK ******
 *
 * Name:		S4PdfBookView.m
 * Module:		UI
 * Library:		S4 iOS_UI Libraries
 *
 * ***** FILE BLOCK *****/


// ================================== Includes =========================================

#import "S4PdfBookView.h"
#import "S4PdfBaseViewController.h"
#import "S4BookViewUtilities.h"
#import "S4PdfDocument.h"
#import "S4PdfPage.h"
#import "S4CommonDefines.h"


// ================================== Defines ==========================================



// ================================== Typedefs ==========================================



// ================================== Globals ==========================================



// ============================= Forward Declarations ==================================



// ================================== Inlines ==========================================



// ========================== Begin Class S4PdfBookView () =============================

@interface S4PdfBookView () 

@property (nonatomic, assign, readwrite) S4PdfPage							*pdfPageRight;
@property (nonatomic, assign, readwrite) S4PdfPage							*pdfPageLeft;

@end



@interface S4PdfBookView (PrivateImpl) 


@end



@implementation S4PdfBookView (PrivateImpl)


@end



#pragma mark -
#pragma mark Implementation

@implementation S4PdfBookView

@synthesize pdfDocument = m_pdfDocument;
@synthesize pdfPageRight = m_pdfPageRight;
@synthesize pdfPageLeft = m_pdfPageLeft;


//============================================================================
//	S4PdfBookView :: init
//============================================================================
- (id)init
{
    return ([self initWithFrame: CGRectZero]);
}


//============================================================================
//	S4PdfBookView :: initWithFrame:
//		Designated initializer
//============================================================================
- (id)initWithFrame: (CGRect)frame
{
	self = [super initWithFrame: frame];
    if (nil != self)
	{
		m_pdfDocument = nil;
		m_pdfPageRight = nil;
		m_pdfPageLeft = nil;
    }
    return (self);
}


//============================================================================
//	S4PdfBookView :: documentPageCount
//============================================================================
- (NSUInteger)documentPageCount
{
	return ([self.pdfDocument pageCount]);
}


//============================================================================
//	S4PdfBookView :: getImageForPageIndex:
//============================================================================
- (CGImageRef)getImageForPageIndex: (NSUInteger)pageIndex forPageType: (PageImageType)pageType
{
	NSUInteger							pdfPageIndex;
	S4PdfPage							*curPdfCachePage;
	CGColorSpaceRef						colorSpace;
	CGContextRef						context;
	CGPDFPageRef						pdfPageRef;
	CGAffineTransform					transform;
	CGImageRef							imgResult = NULL;

	pdfPageIndex = pageIndex + 1;
	curPdfCachePage = (S4PdfPage *)[m_bookCache bookCachePageForIndex: pdfPageIndex withPageSize: m_pageSize];
	if ((IS_NOT_NULL(curPdfCachePage)) && (NULL != curPdfCachePage.pageImage))
	{
		imgResult = curPdfCachePage.pageImage;
	}
	else if ([self documentPageCount] >= pdfPageIndex)
	{
		colorSpace = CGColorSpaceCreateDeviceRGB();
		context = CGBitmapContextCreate(NULL, 
										m_pageSize.width, 
										m_pageSize.height, 
										8,						// bits per component
										m_pageSize.width * 4, 	// bytes per row
										colorSpace, 
										(kCGImageAlphaPremultipliedLast | kCGBitmapByteOrder32Big));
		CGColorSpaceRelease(colorSpace);
		CGContextClipToRect(context, CGRectMake(0, 0, m_pageSize.width, m_pageSize.height));

		CGContextSetInterpolationQuality(context, kCGInterpolationHigh);
		CGContextSetRenderingIntent(context, kCGRenderingIntentDefault);

		// now get the CoreGraphics PDF page representation
		pdfPageRef = [m_pdfDocument getPageRefForPageIndex: pdfPageIndex];
		if (NULL != pdfPageRef)
		{
			CGPDFPageRetain(pdfPageRef);
			transform = aspectFit(CGPDFPageGetBoxRect(pdfPageRef, kCGPDFMediaBox), CGContextGetClipBoundingBox(context));
			CGContextConcatCTM(context, transform);
			CGContextDrawPDFPage(context, pdfPageRef);
			CGPDFPageRelease(pdfPageRef);

			// finally, create an image from the Bitmap instantiated above
			imgResult = CGBitmapContextCreateImage(context);
			if (NULL != imgResult)
			{
				curPdfCachePage = [[S4PdfPage alloc] init];
				curPdfCachePage.dataSource = m_pdfDocument;
				if (YES == [curPdfCachePage setImage: imgResult forPage: pdfPageIndex])
				{
					[m_bookCache setBookCachePage: curPdfCachePage withPageSize: m_pageSize];
				}
				[curPdfCachePage release];
			}
		}
		CGContextRelease(context);
	}

	if (RightPageImage == pageType)
	{
		self.pdfPageRight = curPdfCachePage;
	}
	else if (LeftPageImage == pageType)
	{
		self.pdfPageLeft = curPdfCachePage;
	}
	return (imgResult);
}


//============================================================================
//	S4PdfBookView :: handleTouchPointInPage:
//============================================================================
- (void)handleTouchPointInPage: (CGPoint)touchPoint dragDistance: (CGFloat)delta
{
	NSObject					*linkTarget = nil;
	LinkTargetType				linkType;
	NSInteger					newPageIndex;

	if (0.0 == delta)
	{
		if (nil != self.pdfPageRight)
		{
			linkTarget = [self.pdfPageRight findLinkForViewPoint: touchPoint inViewRect: m_topPageLayer.frame linkTypePtr: &linkType];
		}

		if ((nil == linkTarget) && (nil != self.pdfPageLeft) && (self.mode == TwoPageViewMode))
		{
			linkTarget = [self.pdfPageLeft findLinkForViewPoint: touchPoint inViewRect: m_leftPageLayer.frame linkTypePtr: &linkType];
		}

		if (IS_NOT_NULL(linkTarget))
		{
			if (LinkIsPageNumber == linkType)
			{
				newPageIndex = ([(NSNumber *)linkTarget integerValue] - 1);
				if ((0 <= newPageIndex) && ([self documentPageCount] > newPageIndex))
				{
					self.currentPageIndex = newPageIndex;
				}
			}
			else if (LinkIsURL == linkType)
			{
				[(S4PdfBaseViewController *)self.myViewController handleWebURL: (NSURL *)linkTarget];
			}
		}
	}
}

@end
