/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the S4 iOS Libraries.
 *
 * The Initial Developer of the Original Code is
 * Michael Papp dba SeaStones Software Company.
 * All software created by the Initial Developer are Copyright (C) 2008-2011
 * the Initial Developer. All Rights Reserved.
 *
 * Original Author:
 *			Michael Papp, San Francisco, CA, USA
 *
 * ***** END LICENSE BLOCK ***** */

/* ***** FILE BLOCK ******
 *
 * Name:		S4PdfDocument.m
 * Module:		Core
 * Library:		S4 iOS Libraries
 *
 * ***** FILE BLOCK *****/


// ================================== Includes =========================================

#import "S4PdfDocument.h"
#import "S4FileUtilities.h"
#import "S4CommonDefines.h"


// ================================== Defines ==========================================



// ================================== Typedefs ==========================================



// ================================== Globals ==========================================

S4_INTERN_CONSTANT_NSSTR				kPageCountNumberKey = @"PageCountNumber";
S4_INTERN_CONSTANT_NSSTR				kIsValidDocumentKey = @"IsValidDocument";
S4_INTERN_CONSTANT_NSSTR				kFullFilePathKey = @"FullFilePath";
S4_INTERN_CONSTANT_NSSTR				kPasswordStringKey = @"PasswordString";
S4_INTERN_CONSTANT_NSSTR				kFileDateKey = @"FileDate";
S4_INTERN_CONSTANT_NSSTR				kFileSizeKey = @"FileSize";
S4_INTERN_CONSTANT_NSSTR				kCurrentPageNumberKey = @"CurrentPageNumber";


// ============================= Forward Declarations ==================================



// ================================== Inlines ==========================================



// =========================== Begin Class S4PdfDocument () ============================

@interface S4PdfDocument ()

@property (nonatomic, assign, readwrite) CGPDFDocumentRef				pdfDocumentRef;
@property (nonatomic, strong, readwrite) NSNumber						*pageCountNumber;
@property (nonatomic, assign, readwrite) BOOL							isValidDocument;
@property (nonatomic, copy, readwrite) NSString							*fullFilePath;
@property (nonatomic, strong, readwrite) NSDate							*fileDate;
@property (nonatomic, strong, readwrite) NSNumber						*fileSize;

@end



// ===================== Begin Class S4PdfDocument (PrivateImpl) =======================

@interface S4PdfDocument (PrivateImpl) 

- (BOOL)openPdfDocument;

- (void)drawPageRef: (CGPDFPageRef)pdfPageRef
		  inContext: (CGContextRef)context
			atPoint: (CGPoint)point
		  withScale: (float)scale;

@end


@implementation S4PdfDocument (PrivateImpl)

//============================================================================
//	S4PdfDocument (PrivateImpl) :: openPdfDocument
//		NOTE: m_fullFilePath (self.fullFilePath) and m_passwordStr
//			  (self.password) MUST be set before calling this method.
//============================================================================
- (BOOL)openPdfDocument
{
	NSURL						*pdfURL;
	char						pwdBuffer[128];
	BOOL						bResult = NO;

	pdfURL = self.pdfURL;
	if (IS_NOT_NULL(pdfURL))
	{
		self.pdfDocumentRef = CGPDFDocumentCreateWithURL((CFURLRef)pdfURL);
		if (NULL != self.pdfDocumentRef)
		{
			if (CGPDFDocumentIsEncrypted(self.pdfDocumentRef) == TRUE)				// Encrypted
			{
				// Try a blank password first, per Apple's Quartz PDF example
				if (CGPDFDocumentUnlockWithPassword(self.pdfDocumentRef, "") == FALSE)
				{
					// Nope, now let's try the provided password to unlock the PDF
					if (STR_NOT_EMPTY(self.password))
					{
						[self.password getCString: pwdBuffer maxLength: 126 encoding: NSUTF8StringEncoding];
						CGPDFDocumentUnlockWithPassword(self.pdfDocumentRef, pwdBuffer);
					}
				}
			}

			// if the PDF is not locked, or we unlocked it in the code above, is should be available
			if (CGPDFDocumentIsUnlocked(self.pdfDocumentRef) == TRUE)
			{
				bResult = YES;
			}
			else
			{
				CGPDFDocumentRelease(self.pdfDocumentRef);
				self.pdfDocumentRef = NULL;
			}
		}
	}
	return (bResult);
}


//============================================================================
//	S4PdfDocument (PrivateImpl) :: drawPageRef:inContext:atPoint:withScale:
//============================================================================
- (void)drawPageRef: (CGPDFPageRef)pdfPageRef
		  inContext: (CGContextRef)context
			atPoint: (CGPoint)point
		  withScale: (float)scale
{
	CGRect						cropBox;
	int							pageRotation;
	CGRect						clipRect;
	
    if (NULL != pdfPageRef)
    {			
		cropBox = CGPDFPageGetBoxRect(pdfPageRef, kCGPDFCropBox);
		pageRotation = CGPDFPageGetRotationAngle(pdfPageRef);
		
		CGContextSaveGState(context);
		// Setup the coordinate system.
		// Top left corner of the displayed page must be located at the point specified by the 'point' parameter.
		CGContextTranslateCTM(context, point.x, point.y);
		
		// Scale the page to desired level.
		CGContextScaleCTM(context, (scale/100), (scale/100));
		
		// The coordinate system must be set to match the PDF coordinate system.
		switch (pageRotation)
		{
			case 0:
				CGContextTranslateCTM(context, 0, cropBox.size.height);
				CGContextScaleCTM(context, 1, -1);
				break;
			case 90:
				CGContextScaleCTM(context, 1, -1);
				CGContextRotateCTM(context, -M_PI / 2);
				break;
			case 180:
			case -180:
				CGContextScaleCTM(context, 1, -1);
				CGContextTranslateCTM(context, cropBox.size.width, 0);
				CGContextRotateCTM(context, M_PI);
				break;
			case 270:
			case -90:
				CGContextTranslateCTM(context, cropBox.size.height, cropBox.size.width);
				CGContextRotateCTM(context, M_PI / 2);
				CGContextScaleCTM(context, -1, 1);
				break;
			default:
				break;
		}
		
		// The CropBox defines the page visible area, clip everything outside it.
		clipRect = CGRectMake(0, 0, cropBox.size.width, cropBox.size.height);
		CGContextAddRect(context, clipRect);
		CGContextClip(context);
		CGContextSetRGBFillColor(context, 1.0, 1.0, 1.0, 1.0);
		CGContextFillRect(context, clipRect);
		CGContextTranslateCTM(context, -cropBox.origin.x, -cropBox.origin.y);
		CGContextDrawPDFPage(context, pdfPageRef);
		CGContextRestoreGState(context);
	}
}

@end



// =========================== Begin Class S4PdfDocument ===============================

@implementation S4PdfDocument


//============================================================================
//	S4PdfDocument synthesize properties
//============================================================================
@synthesize pdfDocumentRef = m_pdfDocumentRef;
@synthesize pageCountNumber = m_pageCountNumber;
@synthesize isValidDocument = m_bIsValidDocument;
@synthesize fullFilePath = m_fullFilePath;
@synthesize password = m_passwordStr;
@synthesize fileDate = m_fileDate;
@synthesize fileSize = m_fileSize;
@synthesize pageNumber = m_pageNumber;
@dynamic pageCount;
@dynamic fileName;
@dynamic pdfURL;


//============================================================================
//	S4PdfDocument :: needPasswordForDocumentAtPath:
//============================================================================
+ (BOOL)needPasswordForDocumentAtPath: (NSString *)pdfPathAndFilename
{
	NSURL							*pdfURL;
	CGPDFDocumentRef				pdfDocRef;
	BOOL							bResult = NO;
	
	if (STR_NOT_EMPTY(pdfPathAndFilename))
	{
		pdfURL = [NSURL fileURLWithPath: pdfPathAndFilename];
		if (IS_NOT_NULL(pdfURL))
		{
			pdfDocRef = CGPDFDocumentCreateWithURL((CFURLRef)pdfURL);
			if (NULL != pdfDocRef)
			{
				if (CGPDFDocumentIsEncrypted(pdfDocRef) == TRUE)
				{
					// Try a blank password first, per Apple's Quartz PDF example
					if (CGPDFDocumentUnlockWithPassword(pdfDocRef, "") == FALSE)
					{
						bResult = YES;
					}
				}
				CGPDFDocumentRelease(pdfDocRef);
			}
		}
	}
	return (bResult);
}


//============================================================================
//	S4PdfDocument :: init
//============================================================================
- (id)init
{
	self = [super init];
	if (nil != self)
	{
		// private member vars
		m_pdfDocumentRef = NULL;
		m_pageCountNumber = nil;
		m_bIsValidDocument = NO;
		m_fullFilePath = nil;
		m_passwordStr = nil;
		m_fileDate = nil;
		m_fileSize = nil;
		m_pageNumber = nil;
	}
	return (self);
}


//============================================================================
//	S4PdfDocument :: dealloc
//============================================================================
- (void)dealloc
{
	m_bIsValidDocument = NO;

	if (NULL != m_pdfDocumentRef)
	{
		CGPDFDocumentRelease(m_pdfDocumentRef);
		m_pdfDocumentRef = NULL;
	}

	NS_SAFE_RELEASE(m_pageCountNumber);
	NS_SAFE_RELEASE(m_fullFilePath);
	NS_SAFE_RELEASE(m_passwordStr);
	NS_SAFE_RELEASE(m_fileDate);
	NS_SAFE_RELEASE(m_fileSize);
	NS_SAFE_RELEASE(m_pageNumber);

	[super dealloc];
}


//============================================================================
//	S4PdfDocument :: openDocumentWithFilePath:
//============================================================================
- (BOOL)openDocumentWithFilePath: (NSString *)fullFilePathStr usingPassword: (NSString *)passwordStr
{
	size_t						numDocPages;
	NSDictionary				*fileAttributes;
	BOOL						bResult = NO;

	if ((NO == self.isValidDocument) && (STR_NOT_EMPTY(fullFilePathStr)))
	{
		self.fullFilePath = fullFilePathStr;
		self.password = passwordStr;
		if (YES == [self openPdfDocument])
		{
			numDocPages = CGPDFDocumentGetNumberOfPages(m_pdfDocumentRef);
			if (0 < numDocPages)
			{
				self.pageCountNumber = [NSNumber numberWithUnsignedInteger: (NSUInteger)numDocPages];
				self.isValidDocument = YES;
				fileAttributes = [S4FileUtilities attributesOfItemAtPath: self.fullFilePath];
				if (IS_NOT_NULL(fileAttributes))
				{
					self.fileDate = [fileAttributes objectForKey: NSFileModificationDate];
					self.fileSize = [fileAttributes objectForKey: NSFileSize];
				}
				self.pageNumber = [NSNumber numberWithInteger: 1];
				bResult = YES;
			}
			else
			{
				CGPDFDocumentRelease(m_pdfDocumentRef);
				m_pdfDocumentRef = NULL;
				self.fullFilePath = nil;
				self.password = nil;
			}
		}
		else
		{
			self.fullFilePath = nil;
			self.password = nil;
		}
	}
	return (bResult);
}


//============================================================================
//	S4PdfDocument :: getSizeOfPageAtIndex:
//============================================================================
- (CGSize)getSizeOfPageAtIndex: (NSUInteger)pageIndex
{
	NSUInteger					actualPage;
	CGPDFPageRef				pdfPageRef;
	CGRect						cropBoxRect;
	CGRect						mediaBoxRect;
	CGRect						effectiveRect;
	NSInteger					degrees;
	CGSize						pageSizeResult = CGSizeZero;

	if (YES == self.isValidDocument)
	{
		if (pageIndex < 1)
		{
			actualPage = 1;
		}
		else if (pageIndex > self.pageCount)
		{
			actualPage = self.pageCount;
		}
		else
		{
			actualPage = pageIndex;
		}
		
		pdfPageRef = CGPDFDocumentGetPage(m_pdfDocumentRef, actualPage);
		if (pdfPageRef != NULL)
		{
			cropBoxRect = CGPDFPageGetBoxRect(pdfPageRef, kCGPDFCropBox);
			mediaBoxRect = CGPDFPageGetBoxRect(pdfPageRef, kCGPDFMediaBox);
			effectiveRect = CGRectIntersection(cropBoxRect, mediaBoxRect);

			// Check for page rotation
			degrees = CGPDFPageGetRotationAngle(pdfPageRef);
			if (degrees == 0)
			{
				pageSizeResult = effectiveRect.size;
			}
			else				// Rotate the effective rect so many degrees
			{
				CGFloat radians = (degrees * M_PI / 180.0);
				CGAffineTransform rotation = CGAffineTransformMakeRotation(radians);
				CGRect rotatedRect = CGRectApplyAffineTransform(effectiveRect, rotation);
				pageSizeResult = rotatedRect.size;
			}
		}
	}
	return (pageSizeResult);
}


//============================================================================
//	S4PdfDocument :: pageCount
//============================================================================
- (NSUInteger)pageCount
{
	if (YES == self.isValidDocument)
    {
		return ([self.pageCountNumber unsignedIntegerValue]);
	}
	return (0);
}


//============================================================================
//	S4PdfDocument :: fileName
//============================================================================
- (NSString *)fileName
{
	if (YES == self.isValidDocument)
    {
		return ([self.fullFilePath lastPathComponent]);
	}
	return (nil);
}


//============================================================================
//	S4PdfDocument :: pdfURL
//============================================================================
- (NSURL *)pdfURL
{
	if ((STR_NOT_EMPTY(self.fullFilePath)) && (YES == [S4FileUtilities fileExists: self.fullFilePath]))
	{
		return ([NSURL fileURLWithPath: self.fullFilePath]);
	}
	return (nil);
}


//============================================================================
//	S4PdfDocument :: drawPageAtIndex:inContext:
//============================================================================
- (void)drawPageAtIndex: (NSUInteger)pageIndex inContext: (CGContextRef)context
{
	[self drawPageAtIndex: pageIndex inContext: context atPoint: CGPointMake(0, 0) withScale: 100];
}


//============================================================================
//	S4PdfDocument :: drawPageAtIndex:inContext:atPoint:
//============================================================================
- (void)drawPageAtIndex: (NSUInteger)pageIndex inContext: (CGContextRef)context atPoint: (CGPoint)point
{
	[self drawPageAtIndex: pageIndex inContext: context atPoint: point withScale: 100];
}


//============================================================================
//	S4PdfDocument :: drawPageAtIndex:inContext:atPoint:withScale:
//============================================================================
- (void)drawPageAtIndex: (NSUInteger)pageIndex
	   inContext: (CGContextRef)context
		 atPoint: (CGPoint)point
	   withScale: (float)scale
{
	CGPDFPageRef				pdfPageRef;

	if (YES == self.isValidDocument)
    {
		pdfPageRef = CGPDFDocumentGetPage(m_pdfDocumentRef, pageIndex);
		if (pdfPageRef != NULL)
		{
			CGPDFPageRetain(pdfPageRef);
			[self drawPageRef: pdfPageRef inContext: context atPoint: point withScale: scale];
			CGPDFPageRelease(pdfPageRef);
		}
	}
}


//============================================================================
//	S4PdfDocument :: drawPageAtIndex:inContext:inRect:
//============================================================================
- (void)drawPageAtIndex: (NSUInteger)pageIndex inContext: (CGContextRef)context inRect: (CGRect)displayRect
{
	CGPDFPageRef				pdfPageRef;
	CGRect						cropBox;
	int							pageRotation;
	CGSize						pageVisibleSize;
	CGPoint						topLeftPage;
	
    if (YES == self.isValidDocument)
    {
		if ((displayRect.size.width != 0) && (displayRect.size.height != 0))
		{
			pdfPageRef = CGPDFDocumentGetPage(m_pdfDocumentRef, pageIndex);
			if (pdfPageRef != NULL)
			{
				CGPDFPageRetain(pdfPageRef);
				cropBox = CGPDFPageGetBoxRect(pdfPageRef, kCGPDFCropBox);
				pageRotation = CGPDFPageGetRotationAngle(pdfPageRef);

				pageVisibleSize = CGSizeMake(cropBox.size.width, cropBox.size.height);
				if ((pageRotation == 90) || (pageRotation == 270) ||(pageRotation == -90))
				{
					pageVisibleSize = CGSizeMake(cropBox.size.height, cropBox.size.width);
				}

				float scaleX = displayRect.size.width / pageVisibleSize.width;
				float scaleY = displayRect.size.height / pageVisibleSize.height;
				float scale = scaleX < scaleY ? scaleX : scaleY;

				// Offset relative to top left corner of rectangle where the page will be displayed
				float offsetX = 0;
				float offsetY = 0;

				float rectangleAspectRatio = displayRect.size.width / displayRect.size.height;
				float pageAspectRatio = pageVisibleSize.width / pageVisibleSize.height;

				if (pageAspectRatio < rectangleAspectRatio)
				{
					// The page is narrower than the rectangle, we place it at center on the horizontal
					offsetX = (displayRect.size.width - pageVisibleSize.width * scale) / 2;
				}
				else
				{ 
					// The page is wider than the rectangle, we place it at center on the vertical
					offsetY = (displayRect.size.height - pageVisibleSize.height * scale) / 2;
				}

				topLeftPage = CGPointMake(displayRect.origin.x + offsetX, displayRect.origin.y + offsetY);
				[self drawPageRef: pdfPageRef inContext: context atPoint: topLeftPage withScale: (scale * 100)];
				CGPDFPageRelease(pdfPageRef);
			}
		}
    }
}


//============================================================================
//	S4PdfDocument :: drawPageFastAtIndex:inContext:inRect:
//		'rect' would be "self.bounds" for a UIView
//============================================================================
- (void)drawPageFastAtIndex: (NSUInteger)pageIndex inContext: (CGContextRef)context inRect: (CGRect)displayRect
{
	CGPDFPageRef				cgPdfPage;
	CGAffineTransform			pdfTransform;
	
    if (YES == self.isValidDocument)
    {
        // the pageIndex parameter is 1 based
		cgPdfPage = CGPDFDocumentGetPage(m_pdfDocumentRef, pageIndex);
		
        CGContextSaveGState(context);
		
		// PDF page drawing expects a lower-Left coordinate system, flip the coordinate system before drawing
        CGContextTranslateCTM(context, 0.0, displayRect.size.height);
        CGContextScaleCTM(context, 1.0, -1.0);
		
		// CGPDFPageGetDrawingTransform provides an easy way to get the transform for a PDF page
		pdfTransform = CGPDFPageGetDrawingTransform(cgPdfPage, kCGPDFCropBox, displayRect, 0, true);
		
		// Apply the transform
        CGContextConcatCTM(context, pdfTransform);
		
		// Finally, we draw the page and restore the graphics state
        CGContextDrawPDFPage(context, cgPdfPage);    
        CGContextRestoreGState(context);
    }
}


//============================================================================
//	S4PdfDocument :: getPageRefForPageIndex:
//============================================================================
- (CGPDFPageRef)getPageRefForPageIndex: (NSUInteger)pageIndex
{
	NSUInteger					actualPage;

	if (YES == self.isValidDocument)
	{
		if (1 > pageIndex)
		{
			actualPage = 1;
		}
		else if (self.pageCount < pageIndex)
		{
			actualPage = self.pageCount;
		}
		else
		{
			actualPage = pageIndex;
		}
		return (CGPDFDocumentGetPage(m_pdfDocumentRef, actualPage));
	}
	return (NULL);
}


//============================================================================
//	S4PdfDocument :: reloadData
//============================================================================
- (BOOL)reloadData
{
	BOOL						bResult = NO;

	if (YES == self.isValidDocument)
    {
		@synchronized(self)
		{
			self.isValidDocument = NO;
			if (NULL != m_pdfDocumentRef)
			{
				CGPDFDocumentRelease(m_pdfDocumentRef);
				m_pdfDocumentRef = NULL;
			}

			bResult = [self openPdfDocument];
			if (YES == bResult)
			{
				self.isValidDocument = YES;
				self.pageNumber = [NSNumber numberWithInt: 1];
			}
		}
	}
	return (bResult);
}


//============================================================================
//	S4PdfDocument :: preparePrintDialog:
//============================================================================
- (UIPrintInteractionController *)getPrintDialogWithPageRange: (BOOL)pageRangeVisible
{
	NSURL										*fileURL;
	UIPrintInfo									*curPrintInfo;
	UIPrintInteractionController				*picResult = NULL;

	if (YES == self.isValidDocument)
    {
		fileURL = self.pdfURL;
		if (IS_NOT_NULL(fileURL))
		{
			if ((YES == [UIPrintInteractionController isPrintingAvailable]) &&
				(YES == [UIPrintInteractionController canPrintURL: fileURL]))
			{
				// get the singleton
				picResult = [UIPrintInteractionController sharedPrintController];
				if (IS_NOT_NULL(picResult))
				{
					// optional - set up a UIPrintInfo
					curPrintInfo = [UIPrintInfo printInfo];
					if (IS_NOT_NULL(curPrintInfo))
					{
						curPrintInfo.duplex = UIPrintInfoDuplexLongEdge;
						curPrintInfo.outputType = UIPrintInfoOutputGeneral;
						curPrintInfo.jobName = self.fileName;
						picResult.printInfo = curPrintInfo;
					}

					// this assignment sets the document to be printed
					picResult.printingItem = fileURL;
					picResult.showsPageRange = pageRangeVisible;
				}
			}
		}
	}
	return (picResult);
}


//============================================================================
//	S4PdfDocument :: encodeWithCoder
//============================================================================
- (void)encodeWithCoder: (NSCoder *)encoder
{
	[encoder encodeObject: m_pageCountNumber  forKey: kPageCountNumberKey];
	[encoder encodeBool:   m_bIsValidDocument forKey: kIsValidDocumentKey];
	[encoder encodeObject: m_fullFilePath     forKey: kFullFilePathKey];
	[encoder encodeObject: m_passwordStr      forKey: kPasswordStringKey];
	[encoder encodeObject: m_fileDate         forKey: kFileDateKey];
	[encoder encodeObject: m_fileSize         forKey: kFileSizeKey];
	[encoder encodeObject: m_pageNumber       forKey: kCurrentPageNumberKey];
}


//============================================================================
//	S4PdfDocument :: initWithCoder
//============================================================================
- (id)initWithCoder: (NSCoder *)decoder
{
	self = [super init];
	if (nil != self)
	{
		m_pageCountNumber =  [decoder decodeObjectForKey: kPageCountNumberKey];;
		m_bIsValidDocument = [decoder decodeBoolForKey:   kIsValidDocumentKey];;
		m_fullFilePath =     [decoder decodeObjectForKey: kFullFilePathKey];;
		m_passwordStr =      [decoder decodeObjectForKey: kPasswordStringKey];;
		m_fileDate =         [decoder decodeObjectForKey: kFileDateKey];;
		m_fileSize =         [decoder decodeObjectForKey: kFileSizeKey];;
		m_pageNumber =       [decoder decodeObjectForKey: kCurrentPageNumberKey];;

		if (YES == m_bIsValidDocument)
		{
			[self reloadData];
		}
	}
	return (self);
}

@end
