/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the S4 iOS Libraries.
 *
 * The Initial Developer of the Original Code is
 * Michael Papp dba SeaStones Software Company.
 * All software created by the Initial Developer are Copyright (C) 2008-2011
 * the Initial Developer. All Rights Reserved.
 *
 * Original Author:
 *			Michael Papp, San Francisco, CA, USA
 *
 * ***** END LICENSE BLOCK ***** */

/* ***** FILE BLOCK ******
 *
 * Name:		S4PdfLink.m
 * Module:		Core
 * Library:		S4 iOS Libraries
 *
 * ***** FILE BLOCK *****/


// ================================== Includes =========================================

#import "S4PdfLink.h"
#import "S4CommonDefines.h"


// ================================== Defines ==========================================



// ================================== Typedefs ==========================================



// ================================== Globals ==========================================



// ============================= Forward Declarations ==================================



// ================================== Inlines ==========================================



// ============================= Begin Class S4PdfLink () ==============================

@interface S4PdfLink ()

@property (nonatomic, assign, readwrite) CGRect									linkRect;
@property (nonatomic, retain, readwrite) NSObject								*linkTarget;

@end



// ======================= Begin Class S4PdfLink (privateImpl) =========================

@interface S4PdfLink (PrivateImpl) 

- (CGPDFArrayRef)findDestinationByName: (const char *)destinationName inDestsTree: (CGPDFDictionaryRef)node;

@end


@implementation S4PdfLink (PrivateImpl)

//============================================================================
//	S4PdfLink (privateImpl) :: findDestinationByName:
//============================================================================
- (CGPDFArrayRef)findDestinationByName: (const char *)destinationName inDestsTree: (CGPDFDictionaryRef)node
{
	CGPDFArrayRef					limitsArray;
	CGPDFStringRef					lowerLimit;
	CGPDFStringRef					upperLimit;
	const unsigned char				*llCharPtr;
	const unsigned char				*ulCharPtr;
	CGPDFArrayRef					namesArray;
	size_t							namesCount;
	size_t							idx;
	CGPDFStringRef					destName;
	const unsigned char				*destNameCharPtr;
	CGPDFDictionaryRef				destDictionary;
	CGPDFArrayRef					kidsArray;
	size_t							kidsCount;
	CGPDFDictionaryRef				kidNode;
    CGPDFArrayRef					destinationArray = NULL;

    if (true == CGPDFDictionaryGetArray(node, "Limits", &limitsArray))
	{
        if (true == CGPDFArrayGetString(limitsArray, 0, &lowerLimit))
		{
            if (true == CGPDFArrayGetString(limitsArray, 1, &upperLimit))
			{
                llCharPtr = CGPDFStringGetBytePtr(lowerLimit);
                ulCharPtr = CGPDFStringGetBytePtr(upperLimit);
                if ((strcmp(destinationName, (const char*)llCharPtr) < 0) ||
                    (strcmp(destinationName, (const char*)ulCharPtr) > 0))
				{
                    return NULL;
                }
            }
        }
    }

    if (true == CGPDFDictionaryGetArray(node, "Names", &namesArray))
	{
        namesCount = CGPDFArrayGetCount(namesArray);
        for (idx = 0; idx < namesCount; idx = idx + 2)
		{
            if (true == CGPDFArrayGetString(namesArray, idx, &destName))
			{
                destNameCharPtr = CGPDFStringGetBytePtr(destName);
                if (strcmp((const char *)destNameCharPtr, destinationName) == 0)
				{
                    if (true == CGPDFArrayGetDictionary(namesArray, idx + 1, &destDictionary))
					{
                        if (true == CGPDFDictionaryGetArray(destDictionary, "D", &destinationArray))
						{
                            return destinationArray;
                        }
                    }
                }
            }
        }
    }

    if (true == CGPDFDictionaryGetArray(node, "Kids", &kidsArray))
	{
        kidsCount = CGPDFArrayGetCount(kidsArray);
        for (idx = 0; idx < kidsCount; idx++)
		{
            if (true == CGPDFArrayGetDictionary(kidsArray, idx, &kidNode))
			{
                destinationArray = [self findDestinationByName: destinationName inDestsTree: kidNode];
                if (NULL != destinationArray)
				{
                    return destinationArray;
                }
            }
        }
    }
    return NULL;
}

@end



// ============================= Begin Class S4PdfLink =================================

@implementation S4PdfLink


//============================================================================
//	S4PdfLink synthesize properties
//============================================================================
@synthesize linkRect = m_pdfRectangle;
@synthesize linkTarget = m_linkTarget;


//============================================================================
//	S4PdfLink :: init
//============================================================================
- (id)init
{
    self = [super init];
    if (nil != self)
	{
		m_pdfRectangle = CGRectZero;
		m_linkTarget = nil;
    }
    return self;
}


//============================================================================
//	S4PdfLink :: dealloc
//============================================================================
- (void)dealloc
{
	self.linkTarget = nil;

    [super dealloc];
}


//============================================================================
//	S4PdfLink :: setLinkForDocument:
//============================================================================
- (BOOL)setLinkForDocument: (CGPDFDocumentRef)pdfDocumentRef withAnnotDictionary: (CGPDFDictionaryRef)annotDictionaryRef
{
	CGPDFArrayRef						rectArray;
	CGPDFReal							llx;
	CGPDFReal							lly;
	CGPDFReal							urx;
	CGPDFReal							ury;
	CGPDFReal							tmpReal1;
	CGPDFReal							tmpReal2;
	CGPDFArrayRef						destArray;
	CGPDFDictionaryRef					actionDictionary;
	const char							*actionType;
	CGPDFStringRef						uriRef;
	const char							*uri;
	CGPDFStringRef						destName;
	CGPDFDictionaryRef					catalogDictionary;
	CGPDFDictionaryRef					namesDictionary;
	CGPDFDictionaryRef					destsDictionary;
	const char							*destinationName;
	size_t								pageCount;
	NSUInteger							targetPageIndex;
	size_t								idx;
	CGPDFPageRef						pdfPageRef;
	CGPDFDictionaryRef					pageDictionaryFromDestArray;
	CGPDFDictionaryRef					pageDictionaryFromPage;
	CGPDFInteger						pagePdfNumber;
	BOOL								bResult = NO;

	// clear existing instance vars - PDFLinks can be reused
	m_pdfRectangle = CGRectZero;
	self.linkTarget = nil;

	// *************** first find the rect containing the link ***************
	if (true == CGPDFDictionaryGetArray(annotDictionaryRef, "Rect", &rectArray))
	{
		if ((true == CGPDFArrayGetNumber(rectArray, 0, &llx)) &&
			(true == CGPDFArrayGetNumber(rectArray, 1, &lly)) &&
			(true == CGPDFArrayGetNumber(rectArray, 2, &urx)) &&
			(true == CGPDFArrayGetNumber(rectArray, 3, &ury)))
		{
			if (llx > urx)
			{
				tmpReal1 = llx;
				llx = urx;
				urx = tmpReal1;
			}

			if (lly > ury)
			{
				tmpReal2 = lly;
				lly = ury;
				ury = tmpReal2;
			}
			m_pdfRectangle = CGRectMake(llx, lly, urx - llx, ury - lly);

			// *************** now find the object the link points to (either a page number or a URL) ***************
			// Link target can be stored either in A entry or in Dest entry in annotation dictionary.
			// Dest entry is the destination array that we're looking for. It can be a direct array definition
			// or a name. If it is a name, we need to search recursively for the corresponding destination array 
			// in document's Dests tree.
			// A entry is an action dictionary. There are many action types, we're looking for GoTo and URI actions.
			// GoTo actions are used for links within the same document. The GoTo action has a D entry which is the
			// destination array, same format like Dest entry in annotation dictionary.
			// URI actions are used for links to web resources. The URI action has a 
			// URI entry which is the destination URL.
			// If both entries are present, A entry takes precedence.

			// set to NULL and test later to see if this is a page number
			destArray = NULL;

			if (true == CGPDFDictionaryGetDictionary(annotDictionaryRef, "A", &actionDictionary))
			{
				if (true == CGPDFDictionaryGetName(actionDictionary, "S", &actionType))
				{
					if (strcmp(actionType, "GoTo") == 0)
					{
						CGPDFDictionaryGetArray(actionDictionary, "D", &destArray);
					}

					if (strcmp(actionType, "URI") == 0)
					{
						if (true == CGPDFDictionaryGetString(actionDictionary, "URI", &uriRef))
						{
							uri = (const char *)CGPDFStringGetBytePtr(uriRef);
							self.linkTarget = [[NSString alloc] initWithCString: uri encoding: NSASCIIStringEncoding];
							bResult = YES;
						}
					}
				}
			}
			else
			{
				// Dest entry can be either a string object or an array object.
				if (false == CGPDFDictionaryGetArray(annotDictionaryRef, "Dest", &destArray))
				{
					if (true == CGPDFDictionaryGetString(annotDictionaryRef, "Dest", &destName))
					{
						// Traverse the Dests tree to locate the destination array.
						catalogDictionary = CGPDFDocumentGetCatalog(pdfDocumentRef);
						if (NULL != catalogDictionary)
						{
							if (true == CGPDFDictionaryGetDictionary(catalogDictionary, "Names", &namesDictionary))
							{
								if (true == CGPDFDictionaryGetDictionary(namesDictionary, "Dests", &destsDictionary))
								{
									destinationName = (const char *)CGPDFStringGetBytePtr(destName);
									destArray = [self findDestinationByName: destinationName inDestsTree: destsDictionary];
								}
							}
						}
					}
				}
			}

			if (NULL != destArray)
			{
				targetPageIndex = 0;
				if (true == CGPDFArrayGetDictionary(destArray, 0, &pageDictionaryFromDestArray))
				{
					// First entry in the array is the number of the page in the link
					pageCount = CGPDFDocumentGetNumberOfPages(pdfDocumentRef);
					for (idx = 1; idx <= pageCount; idx++)
					{
						pdfPageRef = CGPDFDocumentGetPage(pdfDocumentRef, idx);
						pageDictionaryFromPage = CGPDFPageGetDictionary(pdfPageRef);
						if (pageDictionaryFromPage == pageDictionaryFromDestArray)
						{
							targetPageIndex = idx;
							break;
						}
					}
				}
				else
				{
					// Some PDF generators use incorrectly the page number as the first element of the array 
					// instead of a reference to the actual page.
					if (true == CGPDFArrayGetInteger(destArray, 0, &pagePdfNumber))
					{
						targetPageIndex = pagePdfNumber + 1;
					}
				}

				if (targetPageIndex > 0)
				{
					self.linkTarget = [[NSNumber alloc] initWithUnsignedInteger: targetPageIndex];
					bResult = YES;
				}
			}
		}
	}
	return (bResult);
}


//============================================================================
//	S4PdfLink :: linkTargetForTappedPdfPoint:
//============================================================================
- (NSObject *)linkTargetForTappedPdfPoint: (CGPoint)tapCGPoint
{
    if ((m_pdfRectangle.origin.x <= tapCGPoint.x) &&
        (m_pdfRectangle.origin.y <= tapCGPoint.y) &&
        (tapCGPoint.x <= m_pdfRectangle.origin.x + m_pdfRectangle.size.width) &&
        (tapCGPoint.y <= m_pdfRectangle.origin.y + m_pdfRectangle.size.height))
	{
        return (self.linkTarget);
    }
	return (nil);
}


//============================================================================
//	S4PdfLink :: getLinkRect
//============================================================================
- (CGRect)getLinkRect
{
	return (m_pdfRectangle);
}

@end
