/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the S4 iOS Libraries.
 *
 * The Initial Developer of the Original Code is
 * Michael Papp dba SeaStones Software Company.
 * All software created by the Initial Developer are Copyright (C) 2008-2011
 * the Initial Developer. All Rights Reserved.
 *
 * Original Author:
 *			Michael Papp, San Francisco, CA, USA
 *
 * ***** END LICENSE BLOCK ***** */

/* ***** FILE BLOCK ******
 *
 * Name:		S4PdfPage.m
 * Module:		UI
 * Library:		S4 iOS Libraries
 *
 * ***** FILE BLOCK *****/


// ================================== Includes =========================================

#import "S4PdfPage.h"
#import "S4PdfDocument.h"
#import "S4PdfLink.h"
#import "S4CommonDefines.h"


// ================================== Defines ==========================================



// ================================== Typedefs ==========================================



// ================================== Globals ==========================================



// ============================= Forward Declarations ==================================



// ================================== Inlines ==========================================



// ============================ Begin Class S4PdfPage () ===============================

@interface S4PdfPage ()

@property (nonatomic, strong, readwrite) NSMutableArray							*pageLinksArray;
@property (nonatomic, assign, readwrite) CGRect									cropBoxForPage;
@property (nonatomic, assign, readwrite) int									rotationAngleForPage;

@end



// ============================= Begin Class S4PdfPage =================================

@implementation S4PdfPage


//============================================================================
//	S4PdfPage synthesize properties
//============================================================================
@synthesize pageLinksArray = m_pageLinksArray;
@synthesize cropBoxForPage = m_cropBoxForPage;
@synthesize rotationAngleForPage = m_rotationAngleForPage;


//============================================================================
//	S4PdfPage :: init
//============================================================================
- (id)init
{
	self = [super init];
	if (nil != self)
	{
		m_pageLinksArray = [[NSMutableArray alloc] init];
		m_cropBoxForPage = CGRectZero;
		m_rotationAngleForPage = 0;
	}
	return (self);
}


//============================================================================
//	S4PdfPage :: dealloc
//============================================================================
- (void)dealloc
{
	NS_SAFE_RELEASE(m_pageLinksArray);

	[super dealloc];
}


//============================================================================
//	S4PdfPage :: setImage:
//============================================================================
- (BOOL)setImage: (CGImageRef)cgImage forPage: (NSUInteger)newPageIndex
{
	S4PdfDocument					*pdfDocument;
	CGPDFPageRef					pdfPageRef;
	CGPDFDictionaryRef				pageDictionary;
	CGPDFArrayRef					annotsArray;
	size_t							annotsCount;
	size_t							idx;
	CGPDFDictionaryRef				annotationDictionary;
	const char						*annotationType;
	CGPDFDocumentRef				pdfDocumentRef;
	S4PdfLink						*pdfLink;
	BOOL							bResult = NO;

	pdfDocument = (S4PdfDocument *)self.dataSource;
	if ((0 < newPageIndex) && (IS_NOT_NULL(pdfDocument)) && (pdfDocument.pageCount >= newPageIndex))
	{
		if (YES == [super setImage: cgImage forPage: newPageIndex])
		{
			pdfPageRef = [pdfDocument getPageRefForPageIndex: newPageIndex];
			if (NULL != pdfPageRef)
			{
				CGPDFPageRetain(pdfPageRef);

				// clear the links array
				[self.pageLinksArray removeAllObjects];

				// get the page's CropBox for hit testing (later on)
				self.cropBoxForPage = CGPDFPageGetBoxRect(pdfPageRef, kCGPDFCropBox);

				// and get the page's rotation angle
				self.rotationAngleForPage = CGPDFPageGetRotationAngle(pdfPageRef);;

				// now find any internal links
				pageDictionary = CGPDFPageGetDictionary(pdfPageRef);
				if (NULL != pageDictionary)
				{
					// PDF links are link annotations stored in the page's Annots array
					if (true == CGPDFDictionaryGetArray(pageDictionary, "Annots", &annotsArray))
					{
						pdfDocumentRef = CGPDFPageGetDocument(pdfPageRef);
						annotsCount = CGPDFArrayGetCount(annotsArray);
						for (idx = 0; idx < annotsCount; idx++)
						{
							if (true == CGPDFArrayGetDictionary(annotsArray, idx, &annotationDictionary))
							{
								if (true == CGPDFDictionaryGetName(annotationDictionary, "Subtype", &annotationType))
								{
									// Link annotations are identified by Link name stored in Subtype key in annotation dictionary.
									if (strcmp(annotationType, "Link") == 0)
									{
										pdfLink = [[S4PdfLink alloc] init];
										if (YES == [pdfLink setLinkForDocument: pdfDocumentRef withAnnotDictionary: annotationDictionary])
										{
											[self.pageLinksArray addObject: pdfLink];
										}
										[pdfLink release];
									}
								}
							}
						}
					}
				}
				bResult = YES;
				CGPDFPageRelease(pdfPageRef);
			}
		}
	}
	return (bResult);
}


//============================================================================
//	S4PdfPage :: findTargetLinkForViewPoint:
//		Find if there is a link annotation at the viewPoint parameter in this
//		page.  The z-order for the links is defined by the link position in
//		the pageLinksArray array.  Last link in the array is top most.
//============================================================================
- (NSObject *)findLinkForViewPoint: (CGPoint)viewPoint inViewRect: (CGRect)pageRenderRect linkTypePtr: (LinkTargetType *)linkType
{
	CGPoint							pdfPoint;
	__block NSObject				*linkTarget;
	LinkTargetType					targetType = LinkIsInvalid;
	NSObject						*targetResult = nil;

	if (0 < [self.pageLinksArray count])
	{
		// first convert the viewPoint parameter from UIKit to PDF page geometry
		pdfPoint = [self pdfPointFromViewPoint: viewPoint inViewRect: pageRenderRect];

		// and now query the link array for this page
		if (NSNotFound != [self.pageLinksArray indexOfObjectPassingTest: ^BOOL(id obj, NSUInteger idx, BOOL *stop)
						   {
							   if (nil != (linkTarget = [obj linkTargetForTappedPdfPoint: pdfPoint]))
							   {
								   *stop = YES;
							   }
							   else
							   {
								   *stop = NO;
							   }
							   return (*stop);
						   }])
		{
			if ([linkTarget isKindOfClass: [NSNumber class]])
			{
				targetType = LinkIsPageNumber;
				targetResult = linkTarget;
			}
			else if ([linkTarget isKindOfClass: [NSString class]])
			{
				targetType = LinkIsURL;
				targetResult = [NSURL URLWithString: (NSString *)linkTarget];
			}
			else
			{
				targetType = LinkTypeUnknown;
				targetResult = linkTarget;
			}
		}
	}

	if (NULL != linkType)
	{
		*linkType = targetType;
	}
	return (targetResult);
}


//============================================================================
//	S4PdfPage :: pdfPointFromViewPoint:
//============================================================================
- (CGPoint)pdfPointFromViewPoint: (CGPoint)viewPoint inViewRect: (CGRect)pageRenderRect
{
	CGPoint						pdfPoint;

	pdfPoint = CGPointMake(0, 0);
	pdfPoint.x = self.cropBoxForPage.size.width * (viewPoint.x - pageRenderRect.origin.x) / pageRenderRect.size.width;
	pdfPoint.y = self.cropBoxForPage.size.height * (pageRenderRect.size.height - (viewPoint.y - pageRenderRect.origin.y)) / pageRenderRect.size.height;
	pdfPoint.x = pdfPoint.x + self.cropBoxForPage.origin.x;
	pdfPoint.y = pdfPoint.y+ self.cropBoxForPage.origin.y;
	return (pdfPoint);
}


//============================================================================
//	S4PdfPage :: viewPointFromPDFPoint:
//============================================================================
- (CGPoint)viewPointFromPDFPoint: (CGPoint)pdfPoint inViewRect: (CGRect)pageRenderRect
{
	CGPoint					viewPoint;
	
    viewPoint = CGPointMake(0, 0);
	viewPoint.x = pageRenderRect.size.width * (pdfPoint.x - self.cropBoxForPage.origin.x) / self.cropBoxForPage.size.width;
	viewPoint.y = pageRenderRect.size.height * (self.cropBoxForPage.size.height - pdfPoint.y) / self.cropBoxForPage.size.height;
	viewPoint.x = viewPoint.x + pageRenderRect.origin.x;
	viewPoint.y = viewPoint.y + pageRenderRect.origin.y;
	return (viewPoint);
}

@end
