/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the S4 iOS Libraries.
 *
 * The Initial Developer of the Original Code is
 * Michael Papp dba SeaStones Software Company.
 * All software created by the Initial Developer are Copyright (C) 2008-2011
 * the Initial Developer. All Rights Reserved.
 *
 * Original Author:
 *			Michael Papp, San Francisco, CA, USA
 *
 * ***** END LICENSE BLOCK ***** */

/* ***** FILE BLOCK ******
 *
 * Name:		S4PageCurlModel.m
 * Module:		UI
 * Library:		S4 iOS_UI Libraries
 *
 * ***** FILE BLOCK *****/


// ================================== Includes =========================================

#import "S4PageCurlModel.h"
#import "S4BaseDataViewController.h"
#import "S4CommonDefines.h"


// ================================== Defines ==========================================



// ================================== Typedefs ==========================================



// ================================== Globals ==========================================

static const NSUInteger						kMaxPagesInCache = 8;


// ============================= Forward Declarations ==================================



// ================================== Inlines ==========================================



// ========================= Begin Class S4PageCurlModel () ============================

@interface S4PageCurlModel ()

@property (nonatomic, strong, readwrite) NSMutableArray				*pageRingBuffer;
@property (nonatomic, assign, readwrite) NSUInteger					cacheIndex;

@end



// ========================== Begin Class S4PageCurlModel ==============================

@implementation S4PageCurlModel


//============================================================================
//	S4PageCurlModel synthesize properties
//============================================================================
@dynamic maxCacheCapacity;
@synthesize pageRingBuffer = m_pageRingBuffer;
@synthesize cacheIndex = m_cacheIndex;
@synthesize dataSource = m_dataSource;


//============================================================================
//	S4PageCurlModel :: init
//============================================================================
- (id)init
{
	NSUInteger					idx;
	NSNull						*nullObject;

    self = [super init];
    if (nil != self)
	{
		m_maxCacheCapacity = kMaxPagesInCache;

		// set up the cache ring buffer
		m_pageRingBuffer = [[NSMutableArray alloc] initWithCapacity: kMaxPagesInCache];
		nullObject = [NSNull null];
		for (idx = 0; idx < m_maxCacheCapacity; idx++)
		{
			[m_pageRingBuffer addObject: nullObject];
		}
		m_cacheIndex = 0;
		m_dataSource = nil;
    }
    return self;
}


//============================================================================
//	S4PageCurlModel :: dealloc
//============================================================================
- (void)dealloc
{
	NS_SAFE_RELEASE(m_pageRingBuffer);
	NS_SAFE_RELEASE(m_dataSource);

	[super dealloc];
}


//============================================================================
//	S4PageCurlModel :: maxCacheCapacity
//============================================================================
- (NSUInteger)maxCacheCapacity
{
	return (m_maxCacheCapacity);
}


//============================================================================
//	S4PageCurlModel :: setMaxCacheCapacity:
//============================================================================
- (void)setMaxCacheCapacity: (NSUInteger)newCapacity
{
	NSUInteger				delta;
	NSNull					*nullObject;
	
	if (newCapacity > m_maxCacheCapacity)
	{
		nullObject = [NSNull null];
		delta = newCapacity - m_maxCacheCapacity;
		while (0 < delta)
		{
			[m_pageRingBuffer addObject: nullObject];
			delta--;
		}
	}
	m_maxCacheCapacity = newCapacity;
}


//============================================================================
//	S4PageCurlModel :: cachedViewControllerForIndex:
//============================================================================
- (UIViewController *)cachedViewControllerForIndex: (NSUInteger)vcIndex
{
	__block S4BaseDataViewController			*curViewController;

	if (NSNotFound != [self.pageRingBuffer indexOfObjectPassingTest: ^BOOL(id obj, NSUInteger idx, BOOL *stop)
					   {
						   curViewController = (S4BaseDataViewController *)obj;
						   if ((IS_NOT_NULL(curViewController)) && ([curViewController indexOfViewController] == vcIndex))
						   {
							   *stop = YES;
						   }
						   else
						   {
							   *stop = NO;
						   }
						   return (*stop);
					   }])
	{
		return (curViewController);
	}
	return (nil);
}


//============================================================================
//	S4PageCurlModel :: cacheViewController:
//============================================================================
- (void)cacheViewController: (S4BaseDataViewController *)viewController
{
	@synchronized(self)
	{		
		// add it to the ring buffer and advance the cacheIndex (or roll it over)
		[m_pageRingBuffer replaceObjectAtIndex: self.cacheIndex withObject: viewController];
		if ((self.cacheIndex + 1) == m_maxCacheCapacity)
		{
			self.cacheIndex = 0;
		}
		else
		{
			self.cacheIndex = self.cacheIndex + 1;
		}
	}
}


//============================================================================
//	S4PageCurlModel :: flushCache
//============================================================================
- (void)flushCache
{
	NSUInteger					idx;
	NSNull						*nullObject;

	// reset the ring buffer cache
	nullObject = [NSNull null];
	for (idx = 0; idx < m_maxCacheCapacity; idx++)
	{
		[m_pageRingBuffer replaceObjectAtIndex: idx withObject: nullObject];
	}
	m_cacheIndex = 0;
}


//============================================================================
//	S4PageCurlModel :: viewControllerAtIndex:
//============================================================================
- (UIViewController *)viewControllerAtIndex: (NSUInteger)index
{   
    return (nil);
}


//============================================================================
//	S4PageCurlModel :: indexOfViewController:
//============================================================================
- (NSUInteger)indexOfViewController: (UIViewController *)viewController
{   
    return (0);
}


//============================================================================
//	S4PageCurlModel :: pageCount
//============================================================================
- (NSUInteger)pageCount
{
	return (0);
}



#pragma mark - Page View Controller Data Source

//============================================================================
//	S4PageCurlModel :: viewControllerBeforeViewController:
//============================================================================
- (UIViewController *)pageViewController: (UIPageViewController *)pageViewController
	  viewControllerBeforeViewController: (UIViewController *)viewController
{
	NSUInteger						index;

	index = [self indexOfViewController: viewController];
    if ((0 < index) && (NSNotFound != index))
	{
        return ([self viewControllerAtIndex: (index - 1)]);
    }
    return (nil);
}


//============================================================================
//	S4PageCurlModel :: viewControllerAfterViewController:
//============================================================================
- (UIViewController *)pageViewController: (UIPageViewController *)pageViewController
	   viewControllerAfterViewController: (UIViewController *)viewController
{
	NSUInteger						index;

    index = [self indexOfViewController: viewController];
    if ((NSNotFound != index) && ([self pageCount] > (index + 1)))
	{
		return ([self viewControllerAtIndex: (index + 1)]);
    }
    return (nil);
}

@end
