/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the S4 iOS Libraries.
 *
 * The Initial Developer of the Original Code is
 * Michael Papp dba SeaStones Software Company.
 * All software created by the Initial Developer are Copyright (C) 2008-2011
 * the Initial Developer. All Rights Reserved.
 *
 * Original Author:
 *			Michael Papp, San Francisco, CA, USA
 *
 * ***** END LICENSE BLOCK ***** */

/* ***** FILE BLOCK ******
 *
 * Name:		S4CircularArray.m
 * Module:		Cache
 * Library:		S4 iOS Libraries
 *
 * ***** FILE BLOCK *****/


// ================================== Includes =========================================

#import "S4CircularArray.h"
#import "S4CommonDefines.h"


// =================================== Defines =========================================

#define MIN_CAPACITY						20


// ================================== Typedefs =========================================



// =================================== Globals =========================================



// ============================= Forward Declarations ==================================



// ================================== Inlines ==========================================




// ========================= Begin Class S4CircularArray ===============================

@implementation S4CircularArray


//============================================================================
//	S4CircularArray :: circularArrayWithCapacity:
//============================================================================
+ (id)circularArrayWithCapacity: (NSUInteger)numItems
{
	return [[[S4CircularArray alloc] initWithCapacity: numItems] autorelease];
}


//============================================================================
//	S4CircularArray :: initWithCapacity:
//============================================================================
- (id)initWithCapacity: (NSUInteger)numItems
{
	NSNull					*nullObject;
	NSUInteger				count;

	self = [super init];
	if (nil != self)
	{
		if (0 >= numItems)
		{
			count = MIN_CAPACITY;
		}
		else
		{
			count = numItems;
		}
		buffer = [[[NSMutableArray alloc] initWithCapacity: count] retain];
		nextIndex = 0;
		nullObject = [NSNull null];
		while (0 < count)
		{
			[buffer addObject: nullObject];
			count--;
		}
	}
	return (self);
}


//============================================================================
//	S4CircularArray :: dealloc
//============================================================================
- (void)dealloc
{
	[buffer release];

	[super dealloc];
}


//============================================================================
//	S4CircularArray :: count
//============================================================================
- (NSUInteger)count
{
	return ([buffer count]);
}


//============================================================================
//	S4CircularArray :: addObject:
//============================================================================
- (id)addObject: (id)newObject
{
	id						replacedObject;

	replacedObject = [self swapObject: newObject atIndex: nextIndex];
	if ((nextIndex + 1) < [buffer count])
	{
		nextIndex++;
	}
	else
	{ 
		nextIndex = 0;
	}
	return (replacedObject);
}


//============================================================================
//	S4CircularArray :: swapObject:
//============================================================================
- (id)swapObject: (id)anObject atIndex: (NSUInteger)idx
{
	id						swappedObject;
	id						idResult = nil;

	swappedObject = [buffer objectAtIndex: idx];
	if (swappedObject != [NSNull null])
	{
		// because when swappedObject is replaced in the buffer, it will be released and might dealloc
		idResult = [[swappedObject retain] autorelease];
	}
	[buffer replaceObjectAtIndex: nextIndex withObject: anObject];
	return (idResult);
}


-(int)findIndexOfObject:(id)key {
	NSNull* nullObj = [NSNull null];
	id anObject;
	int i = nextIndex-1; //start looking at the last object
	if (i<0) { i=0; }
	for (int n = [buffer count]-1; n>=0; n--)
	{
		if (i<0)
		{ 
			i=[buffer count]-1;  //i has to wrap around to the end of the buffer array
		}
		anObject = [buffer objectAtIndex:i];
		if (anObject!=nullObj) {
			BOOL result = [anObject isEqual:key];
			if (result) {
				return i;
			}
		}
		i--; 
	}
	return -1;
}

-(id)findObject:(id)key {
	int i = [self findIndexOfObject:key];
	if (i<0) { 
		return nil;
	}
	return [self objectAtIndex:i];
}

-(id)objectAtIndex:(int)i {
	id anObject = [buffer objectAtIndex:i];
	if (anObject==[NSNull null]) { 
		return nil; 
	}
	return anObject;
}

-(void)removeObjectAtIndex:(int)i {
	[buffer replaceObjectAtIndex:i withObject:[NSNull null]];
}	

-(void)removeObject:(id)key {
	int i = [self findIndexOfObject:key];
	if (i<0) {
		return; 
	}
	//NSLog(@"remove at index %i",i);
	[self removeObjectAtIndex:i];
}

-(NSArray*)allObjects {
	NSMutableArray* all = [NSMutableArray arrayWithCapacity:[buffer count]];
	for (id anObject in buffer) {
		if (![anObject isKindOfClass:[NSNull class]]) {
			[all addObject:anObject];
		}
	}
	return all;
}

@end
