/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the S4 iOS Libraries.
 *
 * The Initial Developer of the Original Code is
 * Michael Papp dba SeaStones Software Company.
 * All software created by the Initial Developer are Copyright (C) 2008-2012
 * the Initial Developer. All Rights Reserved.
 *
 * Original Author:
 *			Gabriel Handford on 7/23/09
 *
 * ***** END LICENSE BLOCK ***** */

/* ***** FILE BLOCK ******
 *
 * Name:		NSObject+S4Yajl.h
 * Module:		Data
 * Library:		S4 iOS Libraries
 *
 * ***** FILE BLOCK *****/


// ================================== Includes =========================================

#import <Foundation/Foundation.h>
#import "S4YajlSerializer.h"
#import "S4YajlParser.h"


// =================================== Defines =========================================



// ================================== Typedefs =========================================



// =================================== Globals =========================================



// ============================= Forward Declarations ==================================



// ================================== Protocols ========================================



// ============================= Class NSObject (S4Yajl) ===============================

/*!
 Generate JSON string from NSArray, NSDictionary or custom object or parse JSON from NSString or custom object.
 
 Parse JSON:
 @code
 NSData *JSONData = [NSData dataWithContentsOfFile:@"example.json"];
 NSArray *arrayFromData = [JSONData yajl_JSON];
 
 NSString *JSONString = @"[\"Test\"]";
 NSArray *arrayFromString = [JSONString yajl_JSON];
 
 // With options and out error
 NSError *error = nil;
 NSArray *arrayFromString = [JSONString yajl_JSONWithOptions:S4JSONParserOptionsAllowComments error:&error];
 @endcode
 
 Generate JSON:
 @code
 NSDictionary *dict = [NSDictionary dictionaryWithObject:@"value" forKey:@"key"];
 NSString *JSONString = [dict yajl_JSONString];
 
 // Beautified with custon indent string
 NSArray *array = [NSArray arrayWithObjects:@"value1", @"value2", nil];
 NSString *JSONString = [dict yajl_JSONStringWithOptions:S4JSONSerializerOptionsBeautify indentString:@"    "];
 @endcode
 */
@interface NSObject (S4Yajl)

/*!
 Create JSON string from object.
 Supported objects include: NSArray, NSDictionary, NSNumber, NSString, NSNull
 To override JSON value to encode (or support custom objects), implement (id)JSON; See YAJLCoding in S4JSONSerializer.h
 @throws S4JSONSerializerInvalidObjectException If object is invalid
 @result JSON String
 */
- (NSString *)yajl_JSONString;

/*!
 Create JSON string from object.
 Supported objects include: NSArray, NSDictionary, NSNumber, NSString, NSNull
 To override JSON value to encode (or support custom objects), implement (id)JSON; See YAJLCoding in S4JSONSerializer.h
 @throws S4JSONSerializerInvalidObjectException If object is invalid
 @param options
  - S4JSONSerializerOptionsNone: No options
  - S4JSONSerializerOptionsBeautify: Beautifiy JSON output
  - S4JSONSerializerOptionsIgnoreUnknownTypes: Ignore unknown types (will use null value)
  - S4JSONSerializerOptionsIncludeUnsupportedTypes: Handle non-JSON types (including NSDate, NSData, NSURL) 
 
 @param indentString
 @result JSON String
 */
- (NSString *)yajl_JSONStringWithOptions: (S4JSONSerializerOptions)options indentString: (NSString *)indentString;

/*!
 Parse JSON (NSString or NSData or dataUsingEncoding:).
 @result JSON object
 @throws S4JSONParserException If a parse error occured
 @throws YAJLParsingUnsupportedException If not NSData or doesn't respond to dataUsingEncoding:
 
 @code
 NSString *JSONString = @"{'foo':['bar', true]}";
 id JSONValue = [JSONString yajl_JSON];
 
 NSData *JSONData = ...;
 id JSONValue = [JSONData yajl_JSON];
 @endcode
 */
- (id)yajl_JSON;

/*!
 Parse JSON (NSString or NSData or dataUsingEncoding:) with out error.
 
 If an error occurs, the returned object will be the current state of the object when
 the error occurred.

 @param error Error to set if we failed to parse
 @result JSON object
 @throws S4JSONParserException If a parse error occured
 @throws YAJLParsingUnsupportedException If not NSData or doesn't respond to dataUsingEncoding:
 
 @code
 NSString *JSONString = @"{'foo':['bar', true]}";
 NSError *error = nil;
 [JSONString yajl_JSON:error];
 if (error) ...;
 @endcode
 */
- (id)yajl_JSON: (NSError **)error;

/*!
 Parse JSON (NSString or NSData or dataUsingEncoding:) with options and out error.
 
 If an error occurs, the returned object will be the current state of the object when
 the error occurred.
 
 @param options Parse options
  - S4JSONParserOptionsNone: No options
  - S4JSONParserOptionsAllowComments: Javascript style comments will be allowed in the input (both /&asterisk; &asterisk;/ and //)
  - S4JSONParserOptionsCheckUTF8: Invalid UTF8 strings will cause a parse error
  - S4JSONParserOptionsStrictPrecision: If YES will force strict precision and return integer overflow error
 
 @param error Error to set if we failed to parse
 @result JSON object
 @throws S4JSONParserException If a parse error occured
 @throws YAJLParsingUnsupportedException If not NSData or doesn't respond to dataUsingEncoding:
 
 @code
 NSString *JSONString = @"{'foo':['bar', true]} // comment";
 NSError *error = nil;
 [JSONString yajl_JSONWithOptions:S4JSONParserOptionsAllowComments error:error];
 if (error) ...;
 @endcode
 */
- (id)yajl_JSONWithOptions: (S4JSONParserOptions)options error: (NSError **)error;

@end

