/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the S4 iOS Libraries.
 *
 * The Initial Developer of the Original Code is
 * Michael Papp dba SeaStones Software Company.
 * All software created by the Initial Developer are Copyright (C) 2008-2011
 * the Initial Developer. All Rights Reserved.
 *
 * Original Author:
 *			Michael Papp, San Francisco, CA, USA
 *
 * ***** END LICENSE BLOCK ***** */

/* ***** FILE BLOCK ******
 *
 * Name:		S4OperationsHandler.m
 * Module:		Core
 * Library:		S4 iOS Libraries
 *
 * ***** FILE BLOCK *****/


// ================================== Includes =========================================

#import "S4OperationsHandler.h"
#import "S4CommonDefines.h"


// =================================== Defines =========================================

#define MAX_CONCURRENT_OPERATIONS						(NSInteger)NSOperationQueueDefaultMaxConcurrentOperationCount

// ALL S4 LIBS SHOULD DEFINE THIS:
#define LIB_DOMAIN_NAME_STR								@"S4OperationsHandler"


// ================================== Typedefs =========================================



// =================================== Globals =========================================



// ============================= Forward Declarations ==================================



// ================================== Inlines ==========================================



// ======================== Begin Class S4OperationsHandler () =========================

@interface S4OperationsHandler ()
{
	NSOperationQueue									*m_operationQueue;
	BOOL												m_isCanceled;
}

@property (readwrite, assign, atomic) BOOL								isCanceled;

@end




// ===================== Begin Class S4OperationsHandler =====================

@implementation S4OperationsHandler

//============================================================================
//	S4OperationsHandler :: properties
//============================================================================

@synthesize isCanceled = m_isCanceled;
@dynamic suspended;
@dynamic queueSize;


//============================================================================
//	S4OperationsHandler :: init
//============================================================================
- (id)init
{
	return ([self initWithOperationQueue: nil]);
}


//============================================================================
//	S4OperationsHandler :: initWithOperationQueue:
//============================================================================
- (id)initWithOperationQueue: (NSOperationQueue *)queue
{
	self = [super init];
	if (nil != self)
	{
		if IS_NOT_NULL(queue)
		{
			m_operationQueue = [queue retain];
		}
		else
		{
			m_operationQueue = [[NSOperationQueue alloc] init];
			[m_operationQueue setMaxConcurrentOperationCount: MAX_CONCURRENT_OPERATIONS];
		}
		m_isCanceled = NO;
	}
	return (self);
}


//============================================================================
//	S4OperationsHandler :: dealloc
//============================================================================
- (void)dealloc
{
	S4_NS_SAFE_RELEASE(m_operationQueue);

	[super dealloc];
}


//============================================================================
//	S4OperationsHandler () :: addBlock:
//============================================================================
- (BOOL)addBlock: (VoidNoArgsBlockType)block
  withMoreBlocks: (NSArray *)arrayOfBlocks
   queuePriority: (NSOperationQueuePriority)priority
 completionBlock: (VoidNoArgsBlockType)finishedBlock
{
	NSBlockOperation					*blockOperation;
	BOOL								bResult = NO;

	if (NULL != block)
	{
		blockOperation = [NSBlockOperation blockOperationWithBlock: block];
		if (IS_NOT_NULL(blockOperation))
		{
			if ((IS_NOT_NULL(arrayOfBlocks)) && ([arrayOfBlocks count] > 0))
			{
				for (VoidNoArgsBlockType anotherBlock in arrayOfBlocks)
				{
					[blockOperation addExecutionBlock: anotherBlock];
				}
			}

			// finish setting up the NSBlockOperation
			[blockOperation setQueuePriority: priority];
			if (NULL != finishedBlock)
			{
				[blockOperation setCompletionBlock: finishedBlock];
			}
			[m_operationQueue addOperation: blockOperation];
			bResult = YES;
			self.isCanceled = NO;
		}
	}
	return (bResult);
}


//============================================================================
//	S4OperationsHandler () :: addSelector:
//============================================================================
- (BOOL)addSelector: (SEL)selector
		   onTarget: (id)target
	  withArguments: (NSArray *)argArray
	  queuePriority: (NSOperationQueuePriority)priority
	completionBlock: (VoidNoArgsBlockType)finishedBlock
{
	NSMethodSignature					*methodSignature;
	NSUInteger							argCount;
	NSInvocation						*invocation;
	NSUInteger							argIndex;
	NSInvocationOperation				*invokeOperation;
	BOOL								bResult = NO;

	if (IS_NOT_NULL(target) && (NULL != selector))
	{
		if (YES == [target respondsToSelector: selector])
		{
			// create an NSInvocation for the target/selector tuple
			argCount = 2;
			if (IS_NOT_NULL(argArray))
			{
				argCount += [argArray count];
			}
			methodSignature = [target methodSignatureForSelector: selector];
			if ([methodSignature numberOfArguments] == argCount)
			{
				invocation = [NSInvocation invocationWithMethodSignature: methodSignature];
				if (IS_NOT_NULL(invocation))
				{
					// make sure everything is retained by the Invocation object
					[invocation retainArguments];

					// the target and selector MUST be set
					[invocation setSelector: selector];
					[invocation setTarget: target];

					// now add the arguments from the NSArray, if any
					if (IS_NOT_NULL(argArray))
					{
						// Indices 0 and 1 indicate the hidden arguments self and _cmd, respectively;
						//  you should set these values directly with the setTarget: and setSelector: methods.
						//  Use indices 2 and greater for the arguments normally passed in a message.			
						argIndex = 2;

						for (id localObject in argArray)
						{
							// When the argument value is an object, pass a pointer to the variable
							//  (or memory) from which the object should be copied:
							[invocation setArgument: &localObject atIndex: argIndex];
							argIndex++;
						}
					}

					// now wrap the NSInvocation in an NSOperation subclass
					invokeOperation = [[NSInvocationOperation alloc] initWithInvocation: invocation];
					if (IS_NOT_NULL(invokeOperation))
					{
						[invokeOperation setQueuePriority: priority];
						if (NULL != finishedBlock)
						{
							[invokeOperation setCompletionBlock: finishedBlock];
						}
						[m_operationQueue addOperation: invokeOperation];
						[invokeOperation release];
						bResult = YES;
						self.isCanceled = NO;
					}
				}
			}
		}
	}
	return (bResult);
}


//============================================================================
//	S4OperationsHandler () :: addOperation:
//============================================================================
- (BOOL)addOperation: (NSOperation *)operation
{
	BOOL								bResult = NO;

	if ((IS_NOT_NULL(operation)) && ([operation isKindOfClass: [NSOperation class]] == YES))
	{
		[m_operationQueue addOperation: operation];
		bResult = YES;
		self.isCanceled = NO;
	}
	return (bResult);
}


//============================================================================
//	S4OperationsHandler :: allOperations
//============================================================================
- (NSArray *)allOperations
{
	return ([m_operationQueue operations]);
}


//============================================================================
//	S4OperationsHandler :: cancelAllOperations
//============================================================================
- (void)cancelAllOperations
{
	self.isCanceled = YES;
	[m_operationQueue cancelAllOperations];
}


//============================================================================
//	S4OperationsHandler :: waitUntilQueueEmpties
//============================================================================
- (void)waitUntilQueueEmpties
{
	[m_operationQueue waitUntilAllOperationsAreFinished];
}


/**********************************  Property Accessor Methods **********************************/

//============================================================================
//	S4OperationsHandler :: isSuspended
//============================================================================
- (BOOL)suspended
{
	return ([m_operationQueue isSuspended]);
}


//============================================================================
//	S4OperationsHandler :: setSuspended:
//============================================================================
- (void)setSuspended: (BOOL)suspend
{
	[m_operationQueue setSuspended: suspend];
}


//============================================================================
//	S4OperationsHandler :: queueSize
//============================================================================
- (NSInteger)queueSize
{
	return ([m_operationQueue operationCount]);
}


//============================================================================
//	S4OperationsHandler :: setNetworkQueueSize:
//============================================================================
- (void)setQueueSize: (NSInteger)count
{
	[m_operationQueue setMaxConcurrentOperationCount: count];
}

@end
