/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the S4 iOS Libraries.
 *
 * The Initial Developer of the Original Code is
 * Michael Papp dba SeaStones Software Company.
 * All software created by the Initial Developer are Copyright (C) 2008-2012
 * the Initial Developer. All Rights Reserved.
 *
 * Original Author:
 *			Michael Papp, San Francisco, CA, USA
 *
 * ***** END LICENSE BLOCK ***** */

/* ***** FILE BLOCK ******
 *
 * Name:		S4OperationsManager.m
 * Module:		Core
 * Library:		S4 iOS Libraries
 *
 * ***** FILE BLOCK *****/


// ================================== Includes =========================================

#import "S4OperationsManager.h"
#import "S4SingletonClass.h"
#import "S4CommonDefines.h"


// =================================== Defines =========================================

#define MAX_CONCURRENT_OPERATIONS						(NSInteger)NSOperationQueueDefaultMaxConcurrentOperationCount

// ALL S4 LIBS SHOULD DEFINE THIS:
#define LIB_DOMAIN_NAME_STR								@"S4OperationsManager"


// ================================== Typedefs =========================================



// =================================== Globals =========================================



// ============================= Forward Declarations ==================================



// ================================== Inlines ==========================================



// ======================== Begin Class S4OperationsManager () =========================

@interface S4OperationsManager ()
{
	S4OperationsHandler									*m_networkOpsHandler;
	S4OperationsHandler									*m_applicationOpsHandler;
}

- (void)oneTimeInit;

@end




// ===================== Begin Class S4OperationsManager =====================

@implementation S4OperationsManager


//============================================================================
//	S4OperationsManager :: properties
//============================================================================
@dynamic networkQueueSuspended;
@dynamic applicationQueueSuspended;
@dynamic networkQueueSize;
@dynamic applicationQueueSize;


///////////////////////////////////// START SINGLETON METHODS /////////////////////////////////////


SYNTHESIZE_SINGLETON_CLASS_FOR_SHARED_ACCESSOR(S4OperationsManager, defaultManager)


//////////////////////////////////////// INSTANCE METHODS /////////////////////////////////////////


//============================================================================
//	S4OperationsManager () :: oneTimeInit
//============================================================================
- (void)oneTimeInit
{
	m_networkOpsHandler = [[S4OperationsHandler alloc] initWithOperationQueue: nil];
	m_applicationOpsHandler = [[S4OperationsHandler alloc] initWithOperationQueue: nil];
}


//============================================================================
//	S4OperationsManager :: addBlockToNetworkQueue:
//============================================================================
- (BOOL)addBlockToNetworkQueue: (VoidNoArgsBlockType)block
				withMoreBlocks: (NSArray *)arrayOfBlocks
				 queuePriority: (NSOperationQueuePriority)priority
			   completionBlock: (VoidNoArgsBlockType)finishedBlock
{
	return ([m_networkOpsHandler addBlock: block
						   withMoreBlocks: arrayOfBlocks
							queuePriority: priority
						  completionBlock: finishedBlock]);
}


//============================================================================
//	S4OperationsManager :: addBlockToApplicationQueue:
//============================================================================
- (BOOL)addBlockToApplicationQueue: (VoidNoArgsBlockType)block
					withMoreBlocks: (NSArray *)arrayOfBlocks
					 queuePriority: (NSOperationQueuePriority)priority
				   completionBlock: (VoidNoArgsBlockType)finishedBlock
{
	return ([m_applicationOpsHandler addBlock: block
							   withMoreBlocks: arrayOfBlocks
								queuePriority: priority
							  completionBlock: finishedBlock]);
}


//============================================================================
//	S4OperationsManager :: addSelectorToNetworkQueue:
//============================================================================
- (BOOL)addSelectorToNetworkQueue: (SEL)selector
						 onTarget: (id)target
					withArguments: (NSArray *)argArray
					queuePriority: (NSOperationQueuePriority)priority
				  completionBlock: (VoidNoArgsBlockType)finishedBlock
{
	return ([m_networkOpsHandler addSelector: selector
									onTarget: target
							   withArguments: argArray
							   queuePriority: priority
							 completionBlock: finishedBlock]);
}


//============================================================================
//	S4OperationsManager :: addSelectorToApplicationQueue:
//============================================================================
- (BOOL)addSelectorToApplicationQueue: (SEL)selector
							 onTarget: (id)target
						withArguments: (NSArray *)argArray
						queuePriority: (NSOperationQueuePriority)priority
					  completionBlock: (VoidNoArgsBlockType)finishedBlock
{
	return ([m_applicationOpsHandler addSelector: selector
										onTarget: target
								   withArguments: argArray
								   queuePriority: priority
								 completionBlock: finishedBlock]);
}


//============================================================================
//	S4OperationsManager :: addOperationToNetworkQueue:
//============================================================================
- (BOOL)addOperationToNetworkQueue: (NSOperation *)operation
{
	return ([m_networkOpsHandler addOperation: operation]);
}


//============================================================================
//	S4OperationsManager :: addOperationToApplicationQueue:
//============================================================================
- (BOOL)addOperationToApplicationQueue: (NSOperation *)operation
{
	return ([m_applicationOpsHandler addOperation: operation]);
}


//============================================================================
//	S4OperationsManager :: allNetworkOperations
//============================================================================
- (NSArray *)allNetworkOperations
{
	return ([m_networkOpsHandler allOperations]);
}


//============================================================================
//	S4OperationsManager :: cancelAllNetworkOperations
//============================================================================
- (void)cancelAllNetworkOperations
{
	[m_networkOpsHandler cancelAllOperations];
}


//============================================================================
//	S4OperationsManager :: waitUntilNetworkQueueEmpties
//============================================================================
- (void)waitUntilNetworkQueueEmpties
{
	[m_networkOpsHandler waitUntilQueueEmpties];
}


//============================================================================
//	S4OperationsManager :: allApplicationOperations
//============================================================================
- (NSArray *)allApplicationOperations
{
	return ([m_applicationOpsHandler allOperations]);
}


//============================================================================
//	S4OperationsManager :: cancelAllApplicationOperations
//============================================================================
- (void)cancelAllApplicationOperations
{
	[m_applicationOpsHandler cancelAllOperations];
}


//============================================================================
//	S4OperationsManager :: waitUntilApplicationQueueEmpties
//============================================================================
- (void)waitUntilApplicationQueueEmpties
{
	[m_applicationOpsHandler waitUntilQueueEmpties];
}


/**********************************  Property Accessor Methods **********************************/

//============================================================================
//	S4OperationsManager :: networkQueueSuspended
//============================================================================
- (BOOL)networkQueueSuspended
{
	return (m_networkOpsHandler.isSuspended);
}


//============================================================================
//	S4OperationsManager :: setNetworkQueueSuspended:
//============================================================================
- (void)setNetworkQueueSuspended: (BOOL)suspend
{
	m_networkOpsHandler.suspended = suspend;
}


//============================================================================
//	S4OperationsManager :: networkQueueSize
//============================================================================
- (NSInteger)networkQueueSize
{
	return (m_networkOpsHandler.queueSize);
}


//============================================================================
//	S4OperationsManager :: setNetworkQueueSize:
//============================================================================
- (void)setNetworkQueueSize: (NSInteger)count
{
	m_networkOpsHandler.queueSize = count;
}


//============================================================================
//	S4OperationsManager :: applicationQueueSuspended
//============================================================================
- (BOOL)applicationQueueSuspended
{
	return (m_applicationOpsHandler.isSuspended);
}


//============================================================================
//	S4OperationsManager :: setApplicationQueueSuspended:
//============================================================================
- (void)setApplicationQueueSuspended: (BOOL)suspend
{
	m_applicationOpsHandler.suspended = suspend;
}

//============================================================================
//	S4OperationsManager :: applicationQueueSize
//============================================================================
- (NSInteger)applicationQueueSize
{
	return (m_applicationOpsHandler.queueSize);
}

//============================================================================
//	S4OperationsManager :: setApplicationQueueSize:
//============================================================================
- (void)setApplicationQueueSize: (NSInteger)count
{
	m_applicationOpsHandler.queueSize = count;
}

@end
