/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the S4 iOS Libraries.
 *
 * The Initial Developer of the Original Code is
 * Michael Papp dba SeaStones Software Company.
 * All software created by the Initial Developer are Copyright (C) 2008-2012
 * the Initial Developer. All Rights Reserved.
 *
 * Original Author:
 *			Michael Papp, San Francisco, CA, USA
 *
 * ***** END LICENSE BLOCK ***** */

/* ***** FILE BLOCK ******
 *
 * Name:		S4JSONCommon.m
 * Module:		Data
 * Library:		S4 iOS Libraries
 *
 * ***** FILE BLOCK *****/


// ================================== Includes =========================================

#import "S4JSONCommon.h"


// =================================== Defines =========================================



// ================================== Typedefs =========================================



// =================================== Globals =========================================

// ALL S4 LIBS SHOULD DEFINE THIS:
S4_INTERN_CONSTANT_NSSTR						S4JSONCommonErrorDomain = @"S4JSONCommonErrorDomain";


// ============================= Forward Declarations ==================================



// ================================== Inlines ==========================================



// ========================== Begin Class S4JSONCommon () ==============================

@interface S4JSONCommon ()

@end



// ============================= Begin Class S4JSONCommon ==============================

@implementation S4JSONCommon


//============================================================================
//	S4JSONCommon :: properties
//============================================================================



//============================================================================
//	S4JSONCommon :: init
//============================================================================
- (id)init
{
	self = [super init];
	if (nil != self)
	{
		DECLARE_ABSTRACT_CLASS(S4JSONCommon)
	}
	return (self);
}


//============================================================================
//	S4JSONCommon :: errorforCode:
//============================================================================
+ (NSError *)errorforCode: (S4JSONParserError)code
			  description: (NSString *)descStr
				   reason: (NSString *)failStr
{
	NSMutableDictionary						*userDict = nil;
	NSString								*localizedDescription;
	NSString								*localizedFailureReason;
	NSError									*error;

	if (S4JSONParserNoError == code)				// if no error, set values to nil
	{
		error = nil;
	}
	else											// there was an error
	{
		// create an NSDictionary to hold values for the NSError
		userDict = [NSMutableDictionary dictionaryWithCapacity: 2];
		if (IS_NOT_NULL(userDict))
		{
			if (STR_NOT_EMPTY(descStr))
			{
				localizedDescription = descStr;
			}
			else
			{
				if (S4JSONParserParsingError == code)
				{
					localizedDescription = @"JSON parsing error";
				}
				else if (S4JSONParserInvalidDataError == code)
				{
					localizedDescription = @"JSON data error";
				}
				else if (S4JSONParserAllocationError == code)
				{
					localizedDescription = @"Allocation error";
				}
				else if (S4JSONParserDoubleOverflowError == code)
				{
					localizedDescription = @"Double overflow error";
				}
				else if (S4JSONParserIntegerOverflowError == code)
				{
					localizedDescription = @"Integer overflow error";
				}
				else if (S4JSONParserCanceledError == code)
				{
					localizedDescription = @"Operation canceled error";
				}
				else
				{
					localizedDescription = @"Unknown error";
				}
			}

			if (STR_NOT_EMPTY(failStr))
			{
				localizedFailureReason = failStr;
			}
			else
			{
				if (S4JSONParserParsingError == code)
				{
					localizedFailureReason = @"The JSON parser could not parse the data";
				}
				else if (S4JSONParserInvalidDataError == code)
				{
					localizedFailureReason = @"The JSON could not be found or was malformed";
				}
				else if (S4JSONParserAllocationError == code)
				{
					localizedFailureReason = @"The application is out of memory";
				}
				else if (S4JSONParserDoubleOverflowError == code)
				{
					localizedFailureReason = @"A double being evaluated overflowed";
				}
				else if (S4JSONParserIntegerOverflowError == code)
				{
					localizedFailureReason = @"An integer being evaluated overflowed";
				}
				else if (S4JSONParserCanceledError == code)
				{
					localizedFailureReason = @"Parsing operation canceled due to previous failures";
				}
				else
				{
					localizedFailureReason = @"An unknown error was encountered by the parser";
				}
			}

			// set the keys in the error dictionary
			[userDict setObject: localizedDescription forKey: NSLocalizedDescriptionKey];
			[userDict setObject: localizedFailureReason forKey: NSLocalizedFailureReasonErrorKey];
		}

		// create the NSError
		error = [NSError errorWithDomain: S4JSONCommonErrorDomain code: (NSInteger)code userInfo: userDict];
	}
	return (error);
}


//============================================================================
//	S4JSONCommon :: classTypeForJSONObject:
//============================================================================
+ (S4JSONClassType)classTypeForJSONObject: (id)jsonObject
{
	S4JSONClassType						jsonClassTypeResult;
	
	if (IS_NOT_NULL(jsonObject))
	{
		if ([jsonObject isKindOfClass: [NSDictionary class]])
		{
			jsonClassTypeResult = kNSDictionaryJSONClass;
		}
		else if ([jsonObject isKindOfClass: [NSArray class]])
		{
			jsonClassTypeResult = kNSArrayJSONClass;
		}
		else if ([jsonObject isKindOfClass: [NSString class]])
		{
			jsonClassTypeResult = kNSStringJSONClass;
		}
		else if ([jsonObject isKindOfClass: [NSNumber class]])
		{
			jsonClassTypeResult = kNSNumberJSONClass;
		}
		else
		{
			jsonClassTypeResult = kUnknownJSONClass;
		}			
	}
	else
	{
		jsonClassTypeResult = kInvalidJSONClass;
	}
	return (jsonClassTypeResult);
}


//============================================================================
//	S4JSONCommon :: dateFromJsonString:
//============================================================================
+ (NSDate *)dateFromJsonString: (NSString *)dateString
{
	NSString				*dateStrCopy;
	NSLocale				*enUSPOSIXLocale;
	NSDateFormatter			*dateFormatter;
	NSDate					*date = nil;
	
	if (STR_NOT_EMPTY(dateString))
	{
		dateStrCopy = [dateString copy];
		enUSPOSIXLocale = [[NSLocale alloc] initWithLocaleIdentifier: @"en_US_POSIX"];
		if (IS_NOT_NULL(enUSPOSIXLocale))
		{
			dateFormatter = [[NSDateFormatter alloc] init];
			if (IS_NOT_NULL(dateFormatter))
			{
				[dateFormatter setLocale: enUSPOSIXLocale];
				[dateFormatter setDateFormat: @"dd-MMM-yyyy HH:mm:ss zzz"];
				[dateFormatter setTimeZone: [NSTimeZone timeZoneForSecondsFromGMT: 0]];
				[dateFormatter setLenient: NO];
				date = [dateFormatter dateFromString: dateStrCopy];
				[dateFormatter release];
			}
			[enUSPOSIXLocale release];
		}
		[dateStrCopy release];
	}
	return (date);
}

@end
