/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the S4 iOS Libraries.
 *
 * The Initial Developer of the Original Code is
 * Michael Papp dba SeaStones Software Company.
 * All software created by the Initial Developer are Copyright (C) 2008-2011
 * the Initial Developer. All Rights Reserved.
 *
 * Original Author:
 *			Michael Papp, San Francisco, CA, USA
 *
 * ***** END LICENSE BLOCK ***** */

/* ***** FILE BLOCK ******
 *
 * Name:		S4UserDefaultsManager.m
 * Module:		Data
 * Library:		S4 iOS Libraries
 *
 * ***** FILE BLOCK *****/


// ================================== Includes =========================================

#import "S4UserDefaultsManager.h"
#import "SFHFKeychainUtils.h"
#import "S4SingletonClass.h"
#import "S4CommonDefines.h"


// =================================== Defines =========================================

#define DFLT_DICT_SIZE					(NSUInteger)8


// ================================== Typedefs =========================================



// =================================== Globals =========================================

S4_INTERN_CONSTANT_NSSTR				kS4KeyChainServicePrefix = @"S4KeyChainService";


// ============================= Forward Declarations ==================================



// ================================== Inlines ==========================================



// ======================= Begin Class S4UserDefaultsManager () ========================

@interface S4UserDefaultsManager ()
{
	NSMutableDictionary				*m_userDefaultsDictionary;
}

- (void)oneTimeInit;

@end



// =================== Begin Class S4UserDefaultsManager =====================

@implementation S4UserDefaultsManager


///////////////////////////////////// START SINGLETON METHODS /////////////////////////////////////


SYNTHESIZE_SINGLETON_CLASS_FOR_SHARED_ACCESSOR(S4UserDefaultsManager, defaultManager)


/////////////////////////////////////// CLASS METHODS //////////////////////////////////////

//============================================================================
//	S4UserDefaultsManager () :: registerDefaultsFromFileAtPath:
//============================================================================
+ (BOOL)registerDefaultsFromFileAtPath: (NSString *)userDefaultsPath
{
	NSDictionary					*userDefaultsDictionary;

	if (STR_NOT_EMPTY(userDefaultsPath))
	{
		userDefaultsDictionary = [NSDictionary dictionaryWithContentsOfFile: userDefaultsPath];
		if ((IS_NOT_NULL(userDefaultsDictionary)) && (0 < [userDefaultsDictionary count]))
		{
			[[NSUserDefaults standardUserDefaults] registerDefaults: userDefaultsDictionary];
			return ([[NSUserDefaults standardUserDefaults] synchronize]);
		}
	}
	return (NO);
}


//============================================================================
//	S4UserDefaultsManager () :: isEnabledForiCloud
//============================================================================
+ (BOOL)isEnabledForiCloud
{
	if (nil != NSClassFromString(@"NSUbiquitousKeyValueStore"))		// is iOS 5
	{
        if (nil != [NSUbiquitousKeyValueStore defaultStore])		// is iCloud enabled
		{
			return (YES);
		}
	}
	return (NO);
}



///////////////////////////////////// INSTANCE METHODS /////////////////////////////////////

//============================================================================
//	S4UserDefaultsManager () :: oneTimeInit
//============================================================================
- (void)oneTimeInit
{
	m_userDefaultsDictionary = [[NSMutableDictionary alloc] initWithCapacity: DFLT_DICT_SIZE];
}


//============================================================================
//	S4UserDefaultsManager :: registerBooleanDefault:
//============================================================================
- (void)registerBooleanDefault: (BOOL)bValue forKey: (NSString *)key
{
	[m_userDefaultsDictionary setObject: [NSNumber numberWithBool: bValue] forKey: key];
}


//============================================================================
//	S4UserDefaultsManager :: registerDoubleDefault:
//============================================================================
- (void)registerDoubleDefault: (double)dValue forKey: (NSString *)key
{
	[m_userDefaultsDictionary setObject: [NSNumber numberWithDouble: dValue] forKey: key];
}


//============================================================================
//	S4UserDefaultsManager :: registerFloatDefault:
//============================================================================
- (void)registerFloatDefault: (float)fValue forKey: (NSString *)key
{
	[m_userDefaultsDictionary setObject: [NSNumber numberWithFloat: fValue] forKey: key];
}


//============================================================================
//	S4UserDefaultsManager :: registerIntegerDefault:
//============================================================================
- (void)registerIntegerDefault: (NSInteger)iValue forKey: (NSString *)key
{
	[m_userDefaultsDictionary setObject: [NSNumber numberWithInteger: iValue] forKey: key];
}


//============================================================================
//	S4UserDefaultsManager :: registerObjectDefault:
//============================================================================
- (void)registerObjectDefault: (id)object forKey: (NSString *)key
{
	if (IS_NOT_NULL(object))
	{
		[m_userDefaultsDictionary setObject: object forKey: key];
	}
}


//============================================================================
//	S4UserDefaultsManager :: registerNSCodingObjectDefault:
//============================================================================
- (void)registerNSCodingObjectDefault: (id)object forKey: (NSString *)key
{
	if ((IS_NOT_NULL(object)) && ([object conformsToProtocol: @protocol(NSCoding)]))
	{
		[m_userDefaultsDictionary setObject: [NSKeyedArchiver archivedDataWithRootObject: object]
									 forKey: key];
	}
}


//============================================================================
//	S4UserDefaultsManager :: registerDefaultsAndResetAllValues:
//============================================================================
- (BOOL)registerDefaultsAndResetAllValues: (BOOL)reset
{
	NSUserDefaults				*stdUserDefaults;
	NSString					*curKey;
	id							curValue;

	if (0 < [m_userDefaultsDictionary count])
	{
		// reset == NO means preserve the existing (user modified) settings
		stdUserDefaults = [NSUserDefaults standardUserDefaults];
		for (curKey in m_userDefaultsDictionary)
		{
			curValue = [stdUserDefaults objectForKey: curKey];
			if ((YES == reset) || (nil == curValue))
			{
				[self setObject: [m_userDefaultsDictionary objectForKey: curKey] forKey: curKey];
			}
		}
		[m_userDefaultsDictionary removeAllObjects];
		return (YES);
	}
	return (NO);
}


//============================================================================
//	S4UserDefaultsManager :: arrayForKey:
//============================================================================
- (NSMutableArray *)arrayForKey: (NSString *)key
{
	return ((NSMutableArray *)[[[self objectForKey: key] mutableCopy] autorelease]);
}


//============================================================================
//	S4UserDefaultsManager :: setArray:
//============================================================================
- (BOOL)setArray: (NSArray *)value forKey: (NSString *)key
{
	return ([self setObject: value forKey: key]);
}


//============================================================================
//	S4UserDefaultsManager :: boolForKey:
//============================================================================
- (BOOL)boolForKey: (NSString *)key
{
	return ([(NSNumber *)[self objectForKey: key] boolValue]);
}


//============================================================================
//	S4UserDefaultsManager :: setBool:
//============================================================================
- (BOOL)setBool: (BOOL)bValue forKey: (NSString *)key
{
	return ([self setObject: [NSNumber numberWithBool: bValue] forKey: key]);
}


//============================================================================
//	S4UserDefaultsManager :: dataForKey:
//============================================================================
- (NSMutableData *)dataForKey: (NSString *)key
{
	return ((NSMutableData *)[[[self objectForKey: key] mutableCopy] autorelease]);
}


//============================================================================
//	S4UserDefaultsManager :: setData:
//============================================================================
- (BOOL)setData: (NSData *)value forKey: (NSString *)key
{
	return ([self setObject: value forKey: key]);
}


//============================================================================
//	S4UserDefaultsManager :: dateForKey:
//============================================================================
- (NSDate *)dateForKey: (NSString *)key
{
	return ((NSDate *)[self objectForKey: key]);
}


//============================================================================
//	S4UserDefaultsManager :: setDate:
//============================================================================
- (BOOL)setDate: (NSDate *)value forKey: (NSString *)key
{
	return ([self setObject: value forKey: key]);
}


//============================================================================
//	S4UserDefaultsManager :: dictionaryForKey:
//============================================================================
- (NSMutableDictionary *)dictionaryForKey: (NSString *)key
{
	return ((NSMutableDictionary *)[[[self objectForKey: key] mutableCopy] autorelease]);
}


//============================================================================
//	S4UserDefaultsManager :: setDictionary:
//============================================================================
- (BOOL)setDictionary: (NSDictionary *)value forKey: (NSString *)key
{
	return ([self setObject: value forKey: key]);
}


//============================================================================
//	S4UserDefaultsManager :: doubleForKey:
//============================================================================
- (double)doubleForKey: (NSString *)key
{
	return ([(NSNumber *)[self objectForKey: key] doubleValue]);
}


//============================================================================
//	S4UserDefaultsManager :: setDouble:
//============================================================================
- (BOOL)setDouble: (double)dValue forKey: (NSString *)key
{
	return ([self setObject: [NSNumber numberWithDouble: dValue] forKey: key]);
}


//============================================================================
//	S4UserDefaultsManager :: floatForKey:
//============================================================================
- (float)floatForKey: (NSString *)key
{
	return ([(NSNumber *)[self objectForKey: key] floatValue]);
}


//============================================================================
//	S4UserDefaultsManager :: setFloat:
//============================================================================
- (BOOL)setFloat: (float)fValue forKey: (NSString *)key
{
	return ([self setObject: [NSNumber numberWithFloat: fValue] forKey: key]);
}


//============================================================================
//	S4UserDefaultsManager :: integerForKey:
//============================================================================
- (NSInteger)integerForKey: (NSString *)key
{
	return ([(NSNumber *)[self objectForKey: key] integerValue]);
}


//============================================================================
//	S4UserDefaultsManager :: setInteger:
//============================================================================
- (BOOL)setInteger: (NSInteger)iValue forKey: (NSString *)key
{
	return ([self setObject: [NSNumber numberWithInteger: iValue] forKey: key]);
}


//============================================================================
//	S4UserDefaultsManager :: stringForKey:
//============================================================================
- (NSMutableString *)stringForKey: (NSString *)key
{
	return ((NSMutableString *)[[[self objectForKey: key] mutableCopy] autorelease]);
}


//============================================================================
//	S4UserDefaultsManager :: setString:
//============================================================================
- (BOOL)setString: (NSString *)value forKey: (NSString *)key
{
	return ([self setObject: value forKey: key]);
}


//============================================================================
//	S4UserDefaultsManager :: objectForKey:
//============================================================================
- (id)objectForKey: (NSString *)key
{
	if (STR_NOT_EMPTY(key))
	{
		return ([[NSUserDefaults standardUserDefaults] objectForKey: key]);
	}
	return (nil);
}


//============================================================================
//	S4UserDefaultsManager :: setObject:
//============================================================================
- (BOOL)setObject: (id)object forKey: (NSString *)key
{
	if ((STR_NOT_EMPTY(key)) && (IS_NOT_NULL(object)))
	{
		[[NSUserDefaults standardUserDefaults] setObject: object forKey: key];
		return ([[NSUserDefaults standardUserDefaults] synchronize]);
	}
	return (NO);
}


//============================================================================
//	S4UserDefaultsManager :: getNSCodingObjectForKey:
//============================================================================
- (id)getNSCodingObjectForKey: (NSString *)key
{
	NSData					*data;
	
	if (STR_NOT_EMPTY(key))
	{
		data = [[NSUserDefaults standardUserDefaults] objectForKey: key];
		if ((IS_NOT_NULL(data)) && (YES == [data isKindOfClass: [NSData class]]))
		{
			return ([NSKeyedUnarchiver unarchiveObjectWithData: data]);
		}
	}
	return (nil);
}


//============================================================================
//	S4UserDefaultsManager (PrivateImpl) :: setNSCodingObject:
//============================================================================
- (BOOL)setNSCodingObject: (id)object forKey: (NSString *)key
{
	NSData				*data;
	
	if ((IS_NOT_NULL(object)) && ([object conformsToProtocol: @protocol(NSCoding)]))
	{
		data = [NSKeyedArchiver archivedDataWithRootObject: object];
		return ([self setObject: data forKey: key]);
	}
	return (NO);
}


//============================================================================
//	S4UserDefaultsManager :: colorForKey:
//============================================================================
- (UIColor *)colorForKey: (NSString *)key
{
	return ((UIColor *)[self getNSCodingObjectForKey: key]);	
}


//============================================================================
//	S4UserDefaultsManager :: setColor:
//============================================================================
- (BOOL)setColor: (UIColor *)cValue forKey: (NSString *)key
{
	return ([self setNSCodingObject: cValue forKey: key]);
}


//============================================================================
//	S4UserDefaultsManager :: imageForKey:
//============================================================================
- (UIImage *)imageForKey: (NSString *)key
{
	NSData						*data;

	data = [self dataForKey: key];
    if (IS_NOT_NULL(data))
	{
		return ([UIImage imageWithData: data]);
	}
	return (nil);
}


//============================================================================
//	S4UserDefaultsManager :: numberForKey:
//============================================================================
- (NSNumber *)numberForKey: (NSString *)key
{
	return ((NSNumber *)[self objectForKey: key]);
}


//============================================================================
//	S4UserDefaultsManager :: setNumber:
//============================================================================
- (BOOL)setNumber: (NSNumber *)value forKey: (NSString *)key
{
	return ([self setObject: value forKey: key]);
}


//============================================================================
//	S4UserDefaultsManager :: removeKey:
//============================================================================
- (BOOL)removeKey: (NSString *)key
{
	if (STR_NOT_EMPTY(key))
	{
		[[NSUserDefaults standardUserDefaults] removeObjectForKey: key];
		return ([[NSUserDefaults standardUserDefaults] synchronize]);
	}
	return (NO);
}


//============================================================================
//	S4UserDefaultsManager :: getKeychainValueForKey:
//============================================================================
- (NSString *)getKeychainValueForKey: (NSString *)key
						   forSharer: (NSString *)sharerId
						 accessGroup: (NSString *)accessGroupOrNil
							   error: (NSError **)errorOrNil
{
	if ((STR_NOT_EMPTY(key)) && (STR_NOT_EMPTY(sharerId)))
	{
		return [SFHFKeychainUtils getPasswordForUsername: key
										  andServiceName: [NSString stringWithFormat: @"%@%@", kS4KeyChainServicePrefix, sharerId]
											 accessGroup: accessGroupOrNil
												   error: errorOrNil];
	}
	return (nil);
}


//============================================================================
//	S4UserDefaultsManager :: setKeychainValue:
//============================================================================
- (void)setKeychainValue: (NSString *)value
				  forKey: (NSString *)key 
			   forSharer: (NSString *)sharerId 
			 accessGroup: (NSString *)accessGroupOrNil 
				   error: (NSError **)errorOrNil
{
	if ((STR_NOT_EMPTY(value)) && (STR_NOT_EMPTY(key)) && (STR_NOT_EMPTY(sharerId)))
	{
		[SFHFKeychainUtils storeUsername: key
							 andPassword: value
						  forServiceName: [NSString stringWithFormat: @"%@%@", kS4KeyChainServicePrefix, sharerId]
							 accessGroup: accessGroupOrNil
						  updateExisting: YES
								   error: errorOrNil];
	}
}


//============================================================================
//	S4UserDefaultsManager :: removeKeychainValueForKey:
//============================================================================
- (void)removeKeychainValueForKey: (NSString *)key 
						forSharer: (NSString *)sharerId 
					  accessGroup: (NSString *)accessGroupOrNil 
							error: (NSError **)errorOrNil
{
	if ((STR_NOT_EMPTY(key)) && (STR_NOT_EMPTY(sharerId)))
	{
		[SFHFKeychainUtils deleteItemForUsername: key
								  andServiceName: [NSString stringWithFormat: @"%@%@", kS4KeyChainServicePrefix, sharerId]
									 accessGroup: accessGroupOrNil
										   error: errorOrNil];
	}
}


//============================================================================
//	S4UserDefaultsManager :: updateToiCloud
//============================================================================
- (BOOL)updateToiCloud
{
	NSDictionary							*localUserDfltsDict;
	NSUbiquitousKeyValueStore				*iCloudStore;

    if ([S4UserDefaultsManager isEnabledForiCloud])
	{
		localUserDfltsDict = [[NSUserDefaults standardUserDefaults] dictionaryRepresentation];
		if (IS_NOT_NULL(localUserDfltsDict))
		{
			iCloudStore = [NSUbiquitousKeyValueStore defaultStore];
			[localUserDfltsDict enumerateKeysAndObjectsUsingBlock: ^(id key, id obj, BOOL *stop)
			 {
				 [iCloudStore setObject: obj forKey: key];
			 }];
			return ([iCloudStore synchronize]);
		}
	}
	return (NO);
}


//============================================================================
//	S4UserDefaultsManager :: updateFromiCloud:
//============================================================================
- (BOOL)updateFromiCloud
{
	NSDictionary							*iCloudDfltsDict;
	NSUserDefaults							*userDflts;
	BOOL									bResult = NO;

    if ([S4UserDefaultsManager isEnabledForiCloud])
	{
		iCloudDfltsDict = [[NSUbiquitousKeyValueStore defaultStore] dictionaryRepresentation];
		if (IS_NOT_NULL(iCloudDfltsDict))
		{
			// prevent NSUserDefaultsDidChangeNotification from being posted while we update from iCloud
			[[NSNotificationCenter defaultCenter] removeObserver: self 
															name: NSUserDefaultsDidChangeNotification 
														  object: nil];

			userDflts = [NSUserDefaults standardUserDefaults];
			[iCloudDfltsDict enumerateKeysAndObjectsUsingBlock: ^(id key, id obj, BOOL *stop)
			 {
				 [userDflts setObject:obj forKey:key];
			 }];
			bResult = [userDflts synchronize];

			// enable NSUserDefaultsDidChangeNotification notifications again
			[[NSNotificationCenter defaultCenter] addObserver: self 
													 selector: @selector(updateToiCloud:) 
														 name: NSUserDefaultsDidChangeNotification                                                    
													   object: nil];
		}
	}
	return (bResult);
}

@end
