/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the S4 iOS Libraries.
 *
 * The Initial Developer of the Original Code is
 * Michael Papp dba SeaStones Software Company.
 * All software created by the Initial Developer are Copyright (C) 2008-2012
 * the Initial Developer. All Rights Reserved.
 *
 * Original Author:
 *			Michael Papp, San Francisco, CA, USA
 *
 * ***** END LICENSE BLOCK ***** */

/* ***** FILE BLOCK ******
 *
 * Name:		S4YajlJSONFile.m
 * Module:		Data
 * Library:		S4 iOS Libraries
 *
 * ***** FILE BLOCK *****/


// ================================== Includes =========================================

#import "S4YajlJSONFile.h"
#import "S4NetUtilities.h"
#import "S4FileUtilities.h"
#import "S4OperationsManager.h"


// =================================== Defines =========================================

#define MAX_READ_BYTES							(NSUInteger)32768
#define MAX_READ_BUFFER_SZ						sizeof(uint8_t) * MAX_READ_BYTES					


// ================================== Typedefs =========================================



// =================================== Globals =========================================

NSInteger										S4JSONFileStackCapacity = 20;


// ============================= Forward Declarations ==================================



// ================================== Inlines ==========================================



// ========================= Begin Class S4YajlJSONFile () =============================

@interface S4YajlJSONFile ()
{
	S4YajlParser										*m_parser;
	__weak NSMutableDictionary							*m_dict;
	__weak NSMutableArray								*m_array;
	__weak NSString										*m_key;
	NSMutableArray										*m_stack;
	NSMutableArray										*m_keyStack;
	S4JSONClassType										m_curClassType;
	BOOL												m_bDoneParsing;
	S4JSONParserOptions									m_parserOptions;
	NSTimeInterval										m_requestTimeout;
	BOOL												m_bUseCache;
	NSOperationQueuePriority							m_queuePriority;
	BOOL												m_bAsyncResponse;
}

- (BOOL)startParsingPath: (NSString *)pathStr forSelector: (SEL)parsingSelector;
- (BOOL)setupParse;
- (void)cleanupParse;
- (void)downloadAndParseForUrl: (NSString *)urlStr;
- (void)readAndParseAtFilePath: (NSString *)filePathStr;
- (BOOL)parseDataPrivate: (NSData *)data;
- (void)callDelegateForParseStarted;
- (void)callDelegateWithParserError: (NSError *)error;
- (void)callDelegateForParseEnded;
- (void)popJsonStack;
- (void)popJsonKeyStack;

@end




// ============================ Begin Class S4YajlJSONFile =============================

@implementation S4YajlJSONFile


//============================================================================
//	S4YajlJSONFile :: properties
//============================================================================



//============================================================================
//	S4YajlJSONFile :: init
//============================================================================
- (id)init
{
	return ([self initWithParserOptions: S4JSONParserOptionsAllowComments]);
}


//============================================================================
//	S4YajlJSONFile :: initWithParserOptions:
//============================================================================
- (id)initWithParserOptions: (S4JSONParserOptions)parserOptions
{
	self = [super initWithParserOptions: parserOptions];
	if (nil != self)
	{
		m_parser = nil;
		m_dict = nil;
		m_array = nil;
		m_key = nil;
		m_stack = nil;
		m_keyStack = nil;  
		m_curClassType = kInvalidJSONClass;
		m_bDoneParsing = YES;
		m_parserOptions = parserOptions;
		m_requestTimeout = 0.0;
		m_bUseCache = NO;
		m_queuePriority = NSOperationQueuePriorityNormal;
		m_bAsyncResponse = NO;
	}
	return (self);
}


//============================================================================
//	S4YajlJSONFile :: dealloc
//============================================================================
- (void)dealloc
{
	S4_NS_SAFE_RELEASE(m_stack);
	S4_NS_SAFE_RELEASE(m_keyStack);
	if (IS_NOT_NULL(m_parser))
	{
		m_parser.delegate = nil;
		S4_NS_SAFE_RELEASE(m_parser);
	}

	[super dealloc];
}


//============================================================================
//	S4YajlJSONFile :: parseData:
//============================================================================
- (void)parseData: (NSData *)data error: (NSError **)error
{
	S4JSONParserError					parseError = S4JSONParserNoError;

	// set the contents of the error param to nil
	if (nil != error)
	{
		*error = nil;
	}

	if (IS_NOT_NULL(data))
	{
		if ([self setupParse] == YES)
		{
			parseError = [m_parser parse: data];
			[self cleanupParse];
		}
		else
		{
			parseError = S4JSONParserAllocationError;
		}
	}
	else
	{
		parseError = S4JSONParserInvalidDataError;
	}

	// set the contents of the error param - if there was an error
	if ((nil != error) && (S4JSONParserNoError != parseError))
	{
		*error = [S4JSONCommon errorforCode: parseError description: nil reason: nil];
	}
}


//============================================================================
//	S4YajlJSONFile :: requestJSONfromURLStr:
//============================================================================
- (BOOL)requestJSONfromURLStr: (NSString *)urlStr
				  forDelegate: (NSObject <S4JSONFileDelegate> *)delegate
					  timeout: (NSTimeInterval)requestTimeout
				  useWebCache: (BOOL)bUseCache
				queuePriority: (NSOperationQueuePriority)priority
			useAsyncCallbacks: (BOOL)useAsync
			 useDispatchQueue: (dispatch_queue_t)dispatchQueue
{
	if ((STR_NOT_EMPTY(urlStr)) && (IS_NOT_NULL(delegate)) &&
		([delegate conformsToProtocol: @protocol(S4JSONFileDelegate)]))
	{
		// set the instance vars
		m_delegate = [delegate retain];
		m_requestTimeout = requestTimeout;
		m_bUseCache = bUseCache;
		m_queuePriority = priority;
		m_bAsyncResponse = useAsync;

		// start the parse
		return ([self startParsingPath: urlStr
						   forSelector: @selector(downloadAndParseForUrl:)]);
	}
	return (NO);
}


//============================================================================
//	S4YajlJSONFile :: parseJSONfromFilePath:
//============================================================================
- (BOOL)parseJSONfromFilePath: (NSString *)pathStr
				  forDelegate: (NSObject <S4JSONFileDelegate> *)delegate
				queuePriority: (NSOperationQueuePriority)priority
			useAsyncCallbacks: (BOOL)useAsync
			 useDispatchQueue: (dispatch_queue_t)dispatchQueue
{
	if ((STR_NOT_EMPTY(pathStr)) && ([S4FileUtilities fileExists: pathStr]) &&
		(IS_NOT_NULL(delegate)) && ([delegate conformsToProtocol: @protocol(S4JSONFileDelegate)]))
	{
		// set the instance vars
		m_delegate = [delegate retain];
		m_queuePriority = priority;
		m_bAsyncResponse = useAsync;

		// start the parse
		return ([self startParsingPath: pathStr
						   forSelector: @selector(readAndParseAtFilePath:)]);
	}
	return (NO);
}


//============================================================================
//	S4JSONFile :: cancel
//============================================================================
- (void)cancel
{
	// Sets to end the run loop or while condition during parse
	m_bDoneParsing = YES;
}



/********************************** Private Methods **********************************/

//============================================================================
//	S4YajlJSONFile () :: startParsingPath:
//============================================================================
- (BOOL)startParsingPath: (NSString *)pathStr forSelector: (SEL)parsingSelector
{
	NSMutableArray						*argArray;
	BOOL								bResult = NO;

	argArray = [NSMutableArray arrayWithCapacity: (NSUInteger)2];
	if IS_NOT_NULL(argArray)
	{
		// put the pathStr argument on the called methods parameter list
		[argArray addObject: pathStr];

		// create an NSInvocation for the parsing operation on put it on the global NSOperationQueue
		bResult = [[S4OperationsManager defaultManager] addSelectorToNetworkQueue: parsingSelector
																		 onTarget: self
																	withArguments: argArray
																	queuePriority: m_queuePriority
																  completionBlock: NULL];
		
	}
	return (bResult);
}


//============================================================================
//	S4YajlJSONFile () :: setupParse:
//============================================================================
- (BOOL)setupParse
{
	BOOL				bResult = NO;

	m_parser = [[S4YajlParser alloc] initWithParserOptions: m_parserOptions];
	if (IS_NOT_NULL(m_parser))
	{
		m_stack = [[NSMutableArray alloc] initWithCapacity: S4JSONFileStackCapacity];
		if (IS_NOT_NULL(m_stack))
		{
			m_keyStack = [[NSMutableArray alloc] initWithCapacity: S4JSONFileStackCapacity];
			if (IS_NOT_NULL(m_keyStack))
			{
				m_parser.delegate = self;
				m_dict = nil;
				m_array = nil;
				m_key = nil;
				m_curClassType = kInvalidJSONClass;
				bResult = YES;
			}
		}
	}
	return (bResult);
}


//============================================================================
//	S4YajlJSONFile () :: cleanupParse
//============================================================================
- (void)cleanupParse
{
	S4_NS_SAFE_RELEASE(m_stack);
	S4_NS_SAFE_RELEASE(m_keyStack);
	if (IS_NOT_NULL(m_parser))
	{
		m_parser.delegate = nil;
		S4_NS_SAFE_RELEASE(m_parser);
	}
	m_dict = nil;
	m_array = nil;
	m_key = nil;
}


//============================================================================
//	S4YajlJSONFile () :: downloadAndParseForUrl:
//============================================================================
- (void)downloadAndParseForUrl: (NSString *)urlStr
{
	NSURL								*url;
	NSMutableURLRequest					*request;
	S4HttpConnection					*s4HttpConnection;
	
	@autoreleasepool
	{
		if ([self setupParse] == YES)
		{
			// set the parser executing flag
			m_bDoneParsing = NO;

			// create a properly escaped NSURL from the string params
			url = [S4NetUtilities createNSUrlForPathStr: urlStr baseStr: nil];
			if (IS_NOT_NULL(url))
			{
				// create the request
				request = [S4NetUtilities createRequestForURL: url
													 useCache: m_bUseCache
											  timeoutInterval: m_requestTimeout
													 postData: nil
												   dataIsForm: NO
												handleCookies: YES];
				if (IS_NOT_NULL(request))
				{
					[request setValue: @"application/json; charset=utf-8" forHTTPHeaderField: @"Content-Type"];
					[request setValue: @"application/json" forHTTPHeaderField: @"Accept"];
					[request setValue: @"utf-8" forHTTPHeaderField: @"Accept-Charset"];
					[request setHTTPMethod: @"GET"];

					s4HttpConnection = [[S4HttpConnection alloc] init];
					if (IS_NOT_NULL(s4HttpConnection))
					{
						// open the HTTP connection
						if ([s4HttpConnection openNonCachingConnectionForRequest: request delegate: self])
						{
							// tell the client that the operation has begun (put up a network spinner, etc.)
							if (YES == m_bAsyncResponse)
							{
								[self performSelector: @selector(callDelegateForParseStarted)
											 onThread: [NSThread currentThread]
										   withObject: nil
										waitUntilDone: YES];
							}
							else
							{
								[self performSelectorOnMainThread: @selector(callDelegateForParseStarted)
													   withObject: nil
													waitUntilDone: NO];
							}

							// loop on the runLoop until the m_bdoneParsing flag is set
							do
							{
								[[NSRunLoop currentRunLoop] runMode: NSDefaultRunLoopMode beforeDate: [NSDate distantFuture]];
							}
							while (NO == m_bDoneParsing);
						}

						// dump the connection
						[s4HttpConnection cancelConnection];
						[s4HttpConnection autorelease];
						s4HttpConnection = nil;
					}
				}
			}
			[self cleanupParse];
		}
	}
}


//============================================================================
//	S4YajlJSONFile () :: readAndParseAtFilePath:
//============================================================================
- (void)readAndParseAtFilePath: (NSString *)filePathStr
{
	NSInputStream						*inputStream;
	uint8_t								*readBufferPtr;
	NSInteger							bytesRead;
	NSData								*jsonData;
	NSError								*error;
	
	@autoreleasepool
	{
		if ([self setupParse] == YES)
		{
			// set the parser executing flag
			m_bDoneParsing = NO;

			// read the JSON data from the file
			inputStream = [NSInputStream inputStreamWithFileAtPath: filePathStr];
			if (IS_NOT_NULL(inputStream))
			{
				readBufferPtr = malloc(MAX_READ_BUFFER_SZ);
				if (NULL != readBufferPtr)
				{
					// tell the client that the operation has begun (put up a network spinner, etc.)
					if (YES == m_bAsyncResponse)
					{
						[self performSelector: @selector(callDelegateForParseStarted)
									 onThread: [NSThread currentThread]
								   withObject: nil
								waitUntilDone: YES];
					}
					else
					{
						[self performSelectorOnMainThread: @selector(callDelegateForParseStarted)
											   withObject: nil
											waitUntilDone: NO];
					}

					// read from stream
					[inputStream open];
					while ((YES == [inputStream hasBytesAvailable]) && (NO == m_bDoneParsing))
					{
						bytesRead = [inputStream read: readBufferPtr maxLength: MAX_READ_BYTES];
						if (bytesRead > 0)
						{
							jsonData = [[NSData alloc] initWithBytesNoCopy: (void *)readBufferPtr
																	length: (NSUInteger)bytesRead
															  freeWhenDone: NO];
							if (IS_NOT_NULL(jsonData))
							{
								// Process the block of data
								[self parseDataPrivate: jsonData];
								[jsonData autorelease];
							}
							else
							{
								error = [S4JSONCommon errorforCode: S4JSONParserParsingError description: nil reason: nil];
								if (YES == m_bAsyncResponse)
								{
									[self performSelector: @selector(callDelegateWithParserError:)
												 onThread: [NSThread currentThread]
											   withObject: error
											waitUntilDone: YES];
								}
								else
								{
									[self performSelectorOnMainThread: @selector(callDelegateWithParserError:)
														   withObject: error
														waitUntilDone: NO];
								}
							}
						}
					}

					// close the stream
					[inputStream close];

					// pass the parsed JSON back to the delegate
					if (YES == m_bAsyncResponse)
					{
						[self performSelector: @selector(callDelegateForParseEnded)
									 onThread: [NSThread currentThread]
								   withObject: nil
								waitUntilDone: YES];
					}
					else
					{
						[self performSelectorOnMainThread: @selector(callDelegateForParseEnded)
											   withObject: nil
											waitUntilDone: NO];
					}

					// free the buffer
					free(readBufferPtr);
				}
			}
			[self cleanupParse];
		}
	}
}


//============================================================================
//	S4YajlJSONFile () :: parseDataPrivate:
//============================================================================
- (BOOL)parseDataPrivate: (NSData *)data
{
	S4JSONParserError					parseError;
	NSError								*localError;
	BOOL								bResult = NO;
	
	parseError = [m_parser parse: data];
	if (S4JSONParserNoError == parseError)
	{
		bResult = YES;
	}
	else
	{
		localError = [S4JSONCommon errorforCode: parseError description: nil reason: nil];
		if (IS_NOT_NULL(localError))
		{
			if (YES == m_bAsyncResponse)
			{
				[self performSelector: @selector(callDelegateWithParserError:)
							 onThread: [NSThread currentThread]
						   withObject: localError
						waitUntilDone: YES];
			}
			else
			{
				[self performSelectorOnMainThread: @selector(callDelegateWithParserError:)
									   withObject: localError
									waitUntilDone: NO];
			}
		}
	}
	return (bResult);
}


//============================================================================
//	S4YajlJSONFile () :: popJsonStack
//============================================================================
- (void)popJsonStack
{
	[m_stack removeLastObject];
	m_array = nil;
	m_dict = nil;
	m_curClassType = kInvalidJSONClass;
	
	id value = nil;
	if ([m_stack count] > 0)
	{
		value = [m_stack objectAtIndex: ([m_stack count] - 1)];
	}
	
	if ([value isKindOfClass: [NSArray class]])
	{    
		m_array = (NSMutableArray *)value;
		m_curClassType = kNSArrayJSONClass;
	}
	else if ([value isKindOfClass: [NSDictionary class]])
	{    
		m_dict = (NSMutableDictionary *)value;
		m_curClassType = kNSDictionaryJSONClass;
	}
}


//============================================================================
//	S4YajlJSONFile () :: popJsonKeyStack
//============================================================================
- (void)popJsonKeyStack
{
	m_key = nil;
	[m_keyStack removeLastObject];
	if ([m_keyStack count] > 0) 
	{
		m_key = [m_keyStack objectAtIndex: ([m_keyStack count] - 1)];
	}
}


//============================================================================
//	S4YajlJSONFile () :: callDelegateForParseStarted
//============================================================================
- (void)callDelegateForParseStarted
{
	[m_delegate jsonFileDidBeginParsingData: self];
}


//============================================================================
//	S4YajlJSONFile () :: callDelegateWithParserError:
//============================================================================
- (void)callDelegateWithParserError: (NSError *)error
{
	[m_delegate jsonFile: self didFailWithError: error];
	m_bDoneParsing = YES;
}


//============================================================================
//	S4YajlJSONFile () :: callDelegateForParseEnded
//		This is called explicitly by the read file parser code
//		and implicitly by the S4HttpConnection's completion
//		delegate method
//============================================================================
- (void)callDelegateForParseEnded
{
	S4JSONParserError						parseError;
	S4JSONClassType							jsonClassType;
	NSError									*error;
	
	parseError = [m_parser parseCompleted];
	if ((S4JSONParserNoError == parseError) && (IS_NOT_NULL(m_rootJSONObject)))
	{
		jsonClassType = [S4JSONCommon classTypeForJSONObject: m_rootJSONObject];

		// call the delegate with the JSON from the server parsed into an NSObject class
		[m_delegate jsonFile: self didEndParsingJSON: m_rootJSONObject ofType: jsonClassType];
	}
	else	// could not create a JSON parser
	{
		error = [S4JSONCommon errorforCode: S4JSONParserParsingError description: nil reason: nil];
		[self callDelegateWithParserError: error];
	}
	m_bDoneParsing = YES;
}



// ====================================== S4YajlParser delegate methods ==========================================

//============================================================================
//	S4YajlJSONFile :: parser:didAdd:
//============================================================================
- (void)parser: (S4YajlParser *)parser didAdd: (id)value
{
	switch(m_curClassType)
	{
		case kUnknownJSONClass:
		case kInvalidJSONClass:
		{
			break;
		}

		case kNSArrayJSONClass:
		{
			[m_array addObject: value];
			break;
		}

		case kNSDictionaryJSONClass:
		{
			if (STR_NOT_EMPTY(m_key))
			{
				[m_dict setObject: value forKey: m_key];
			}
			[self popJsonKeyStack];
			break;
		} 

		default:
			break;
	} 
}


//============================================================================
//	S4YajlJSONFile :: parser:didMapKey:
//============================================================================
- (void)parser: (S4YajlParser *)parser didMapKey: (NSString *)key
{
	m_key = key;
	[m_keyStack addObject: m_key];
}


//============================================================================
//	S4YajlJSONFile :: parserDidStartDictionary:
//============================================================================
- (void)parserDidStartDictionary: (S4YajlParser *)parser
{
	NSMutableDictionary				*dict;

	dict = [[NSMutableDictionary alloc] initWithCapacity: S4JSONFileStackCapacity];
	if (IS_NULL(m_rootJSONObject))
	{
		m_rootJSONObject = [dict retain];
	}
	[m_stack addObject: dict];
	[dict release];
	m_dict = dict;
	m_curClassType = kNSDictionaryJSONClass;  
}


//============================================================================
//	S4YajlJSONFile :: parserDidEndDictionary:
//============================================================================
- (void)parserDidEndDictionary: (S4YajlParser *)parser
{
	id value = [[m_stack objectAtIndex: ([m_stack count] - 1)] retain];
	[self popJsonStack];
	[self parser: parser didAdd: value];
	[value release];
}


//============================================================================
//	S4YajlJSONFile :: parserDidStartArray:
//============================================================================
- (void)parserDidStartArray: (S4YajlParser *)parser
{
	NSMutableArray *array = [[NSMutableArray alloc] initWithCapacity: S4JSONFileStackCapacity];
	if (IS_NULL(m_rootJSONObject))
	{
		m_rootJSONObject = [array retain];
	}
	[m_stack addObject: array];
	[array release];
	m_array = array;
	m_curClassType = kNSArrayJSONClass;
}


//============================================================================
//	S4YajlJSONFile :: parserDidEndArray:
//============================================================================
- (void)parserDidEndArray: (S4YajlParser *)parser
{
	id value = [[m_stack objectAtIndex: ([m_stack count] - 1)] retain];
	[self popJsonStack];  
	[self parser: parser didAdd: value];
	[value release];
}



// ================================== S4HttpConnection delegate methods ==========================================

//============================================================================
//	S4YajlJSONFile :: httpConnection:receivedData:
//============================================================================
- (BOOL)httpConnection: (S4HttpConnection *)connection receivedData: (NSData *)data
{
	return ([self parseDataPrivate: data]);
}


//============================================================================
//	S4YajlJSONFile :: httpConnection:failedWithError:
//============================================================================
- (void)httpConnection: (S4HttpConnection *)connection failedWithError: (NSError *)error
{
	if (YES == m_bAsyncResponse)
	{
		[self performSelector: @selector(callDelegateWithParserError:)
					 onThread: [NSThread currentThread]
				   withObject: error
				waitUntilDone: YES];
	}
	else
	{
		[self performSelectorOnMainThread: @selector(callDelegateWithParserError:)
							   withObject: error
							waitUntilDone: NO];
	}
}


//============================================================================
//	S4YajlJSONFile :: httpConnection:completedWithData:
//============================================================================
- (void)httpConnection: (S4HttpConnection *)connection completedWithData: (NSMutableData *)data
{
	if (YES == m_bAsyncResponse)
	{
		[self performSelector: @selector(callDelegateForParseEnded)
					 onThread: [NSThread currentThread]
				   withObject: nil
				waitUntilDone: YES];
	}
	else
	{
		[self performSelectorOnMainThread: @selector(callDelegateForParseEnded)
							   withObject: nil
							waitUntilDone: NO];
	}
}

@end
