/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the S4 iOS Libraries.
 *
 * The Initial Developer of the Original Code is
 * Michael Papp dba SeaStones Software Company.
 * All software created by the Initial Developer are Copyright (C) 2008-2011
 * the Initial Developer. All Rights Reserved.
 *
 * Original Author:
 *			Michael Papp, San Francisco, CA, USA
 *
 * ***** END LICENSE BLOCK ***** */

/* ***** FILE BLOCK ******
 *
 * Name:		S4SingletonClass.h
 * Module:		Headers
 * Library:		S4 iOS Libraries
 *
 * ***** FILE BLOCK *****/


// ================================== Includes =========================================

#import <Foundation/Foundation.h>



// =================================== Defines =========================================

/*	S4SingletonClass defines a MACRO that turns your class into an "Apple-compliant"
	singleton class.  To use, just:
 
	1. Include this header file in your .m class implementation file.

	2. In your "classname".h file, add the following lines and once again replace 
		"classname" with the name of your class:

			a. If you use the "SYNTHESIZE_SINGLETON_CLASS" macro:
				+ (classname *)singleton;

			b. If you use the "SYNTHESIZE_SINGLETON_CLASS_WITH_METHOD" macro:
				+ (classname *)<yourMethodNameForGettingTheSharedInstance>;

	3. Place the SYNTHESIZE_SINGLETON_CLASS(classname) within the implementation
		section of your class replacing "classname" with the name of your class
		and add the following method (probably in private category)

			- (void)oneTimeInit;

	4. Implement the (void)OneTimeInit method and perform your
		instance-specific initialization there (if needed).
*/


#define SYNTHESIZE_SINGLETON_CLASS_IMPL(classname, methodname)					\
																				\
static classname __strong				*g_clsInstance##classname = nil;		\
static id __strong						g_clsAlloc##classname = nil;			\
static dispatch_once_t					onceTokenInstance = 0;					\
static dispatch_once_t					onceTokenClass = 0;						\
static dispatch_once_t					onceTokenInit = 0;						\
																				\
																				\
+ (classname *)methodname														\
{																				\
	__block classname				*localInstance;								\
																				\
	dispatch_once(&onceTokenInstance, ^{										\
		localInstance = [[classname alloc] init];								\
	});																			\
	return (g_clsInstance##classname);											\
}																				\
																				\
																				\
+ (id)allocWithZone: (NSZone *)zone												\
{																				\
	dispatch_once(&onceTokenClass, ^{											\
		g_clsAlloc##classname = [super allocWithZone: zone];					\
	});																			\
	return (g_clsAlloc##classname);												\
}																				\
																				\
																				\
- (id)init																		\
{																				\
	dispatch_once(&onceTokenInit, ^{											\
		g_clsInstance##classname = [super init];								\
		if (nil != g_clsInstance##classname)									\
		{																		\
			[g_clsInstance##classname oneTimeInit];								\
		}																		\
	});																			\
	self = g_clsInstance##classname;											\
	return (g_clsInstance##classname);											\
}																				\
																				\
																				\
- (id)copyWithZone: (NSZone *)zone												\
{																				\
	return (self);																\
}



#define NON_ARC_ADDITIONAL_SINGLETON_METHODS									\
																				\
- (id)retain																	\
{																				\
	return (self);																\
}																				\
																				\
																				\
- (unsigned)retainCount															\
{																				\
	return (UINT_MAX);															\
}																				\
																				\
																				\
- (oneway void)release															\
{																				\
}																				\
																				\
																				\
- (id)autorelease																\
{																				\
	return (self);																\
}



#if (!__has_feature(objc_arc))						/* if singleton is not using ARC */

#define SYNTHESIZE_SINGLETON_CLASS_FOR_SHARED_ACCESSOR(classname, methodname)	\
			SYNTHESIZE_SINGLETON_CLASS_IMPL(classname, methodname)				\
			NON_ARC_ADDITIONAL_SINGLETON_METHODS

#else												/* if singleton is using ARC */

#define SYNTHESIZE_SINGLETON_CLASS_FOR_SHARED_ACCESSOR(classname, methodname)	\
			SYNTHESIZE_SINGLETON_CLASS_IMPL(classname, methodname)

#endif


#define SYNTHESIZE_SINGLETON_CLASS(classname)									\
			SYNTHESIZE_SINGLETON_CLASS_FOR_SHARED_ACCESSOR(classname, singleton)


// ================================== Typedefs =========================================



// =================================== Globals =========================================



// ============================= Forward Declarations ==================================


