/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the S4 iOS Libraries.
 *
 * The Initial Developer of the Original Code is
 * Michael Papp dba SeaStones Software Company.
 * All software created by the Initial Developer are Copyright (C) 2008-2012
 * the Initial Developer. All Rights Reserved.
 *
 * Original Author:
 *			Michael Papp, San Francisco, CA, USA
 *
 * ***** END LICENSE BLOCK ***** */

/* ***** FILE BLOCK ******
 *
 * Name:		S4DownloadInvokeOperation.m
 * Module:		Network
 * Library:		S4 iOS Libraries
 *
 * ***** FILE BLOCK *****/


// ================================== Includes =========================================

#import "S4DownloadInvokeOperation.h"
#import "S4NetUtilities.h"
#import "S4FileUtilities.h"
#import "S4CommonDefines.h"


// ================================== Defines ==========================================



// ================================== Typedefs ==========================================



// ================================== Globals ==========================================



// ============================= Forward Declarations ==================================



// ================================== Inlines ==========================================



// ====================== Begin Class S4DownloadInvokeOperation () =====================

@interface S4DownloadInvokeOperation ()
{
	NSURLRequest								*m_urlRequest;
	S4HttpConnection							*m_s4HttpConnection;
	NSInvocation								*m_progressInvocation;
	NSInvocation								*m_errInvocation;
	NSInvocation								*m_completionInvocation;
	NSString									*m_filePath;
	NSString									*m_fileName;
	NSString									*m_localFileNamePath;
	long long									m_bytesDownloaded;
	long long									m_fileSize;
	BOOL										m_bCancelled;
	BOOL										m_bExecuting;
	BOOL										m_bFinished;
	BOOL										m_bPrepCalled;
	BOOL										m_bUseTmpFile;
}

- (void)setExitFlags;
- (void)invokeProgressCallback: (NSNumber *)percent;
- (void)invokeErrorCallback: (NSError *)returnError;
- (void)invokeCompletionCallback: (NSNumber *)bytesRead;
- (void)deleteUnfinishedFile;

@end



// ==================== Begin Class S4DownloadInvokeOperation ================

@implementation S4DownloadInvokeOperation

//============================================================================
//	S4DownloadInvokeOperation :: properties
//============================================================================
@synthesize filePath = m_filePath;
@synthesize fileName = m_fileName;
@synthesize	localFileNamePath = m_localFileNamePath;
@synthesize bytesDownloaded = m_bytesDownloaded;
@synthesize fileSize = m_fileSize;


//============================================================================
//	S4DownloadInvokeOperation :: downloadInvokeOperationForRequest:
//============================================================================
+ (S4DownloadInvokeOperation *)downloadInvokeOperationForRequest: (NSURLRequest *)request
											  progressInvocation: (NSInvocation *)progressInvoke
												   errInvocation: (NSInvocation *)errInvoke
											completionInvocation: (NSInvocation *)doneInvoke
														filePath: (NSString *)path
														fileName: (NSString *)name
													  useTmpFile: (BOOL)bUseTmpFile
{
	S4DownloadInvokeOperation				*tmpOp;
	S4DownloadInvokeOperation				*opResult = nil;

	tmpOp = [[S4DownloadInvokeOperation alloc] init];
	if (IS_NOT_NULL(tmpOp))
	{
		if ([tmpOp prepareForRequest: request
				  progressInvocation: progressInvoke
					   errInvocation: errInvoke
				completionInvocation: doneInvoke
							filePath: path
							fileName: name
						  useTmpFile: bUseTmpFile] == YES)
		{
			opResult = [tmpOp autorelease];
		}
		else
		{
			[tmpOp release];
		}
	}
	return (opResult);
}


//============================================================================
//	S4DownloadInvokeOperation :: init
//============================================================================
- (id)init
{
	self = [super init];
	if (nil != self)
	{
		m_urlRequest = nil;
		m_s4HttpConnection = nil;
		m_progressInvocation = nil;
		m_errInvocation = nil;
		m_completionInvocation = nil;
		m_filePath = nil;
		m_fileName = nil;
		m_localFileNamePath = nil;
		m_bytesDownloaded = 0;
		m_fileSize = 0;
		m_bCancelled = NO;
		m_bExecuting = NO;
		m_bFinished = NO;
		m_bPrepCalled = NO;
		m_bUseTmpFile = NO;
	}
	return (self);
}


//============================================================================
//	S4DownloadInvokeOperation :: dealloc
//============================================================================
- (void)dealloc
{
	S4_NS_SAFE_RELEASE(m_s4HttpConnection);
	S4_NS_SAFE_RELEASE(m_urlRequest);
	S4_NS_SAFE_RELEASE(m_progressInvocation);
	S4_NS_SAFE_RELEASE(m_errInvocation);
	S4_NS_SAFE_RELEASE(m_completionInvocation);
	S4_NS_SAFE_RELEASE(m_filePath);
	S4_NS_SAFE_RELEASE(m_fileName);
	S4_NS_SAFE_RELEASE(m_localFileNamePath);
	
	[super dealloc];
}


//============================================================================
//	S4DownloadInvokeOperation :: prepareForRequest:
//============================================================================
- (BOOL)prepareForRequest: (NSURLRequest *)request
	   progressInvocation: (NSInvocation *)progressInvoke
			errInvocation: (NSInvocation *)errInvoke
	 completionInvocation: (NSInvocation *)doneInvoke
				 filePath: (NSString *)path
				 fileName: (NSString *)name
			   useTmpFile: (BOOL)bUseTmpFile
{
	BOOL				bResult = NO;

	if ((NO == m_bExecuting) && (IS_NOT_NULL(request)) && (IS_NOT_NULL(doneInvoke)) &&
		(STR_NOT_EMPTY(path)) && (STR_NOT_EMPTY(name)))
	{
		m_filePath = [path copy];
		m_fileName = [name copy];
		m_bUseTmpFile = bUseTmpFile;
		m_localFileNamePath = [[m_filePath stringByAppendingPathComponent: m_fileName] copy];
		if ([S4FileUtilities fileExists: m_localFileNamePath] == NO)
		{
			m_urlRequest = [request retain];
			m_completionInvocation = [doneInvoke retain];
			[m_completionInvocation retainArguments];

			if (IS_NOT_NULL(progressInvoke))
			{
				m_progressInvocation = [progressInvoke retain];
				[m_progressInvocation retainArguments];
			}

			if (IS_NOT_NULL(errInvoke))
			{
				m_errInvocation = [errInvoke retain];
				[m_errInvocation retainArguments];
			}

			// we are ready, set the flags
			m_bPrepCalled = YES;
			bResult = YES;

			// keep ourselves alive while our request is running
			[self retain];
		}
		else
		{
			S4_NS_SAFE_RELEASE(m_filePath);
			S4_NS_SAFE_RELEASE(m_fileName);
			S4_NS_SAFE_RELEASE(m_localFileNamePath);
		}
	}
	return (bResult);
}




// ================================== NSOperation subclassed methods ==========================================

//============================================================================
//	S4DownloadInvokeOperation :: start
//============================================================================
- (void)start
{
	if ([NSThread isMainThread] == YES)
	{
		if (([self isCancelled] == NO) && (NO == m_bExecuting) && (YES == m_bPrepCalled))
		{
			m_s4HttpConnection = [[S4HttpConnection alloc] init];
			if (IS_NOT_NULL(m_s4HttpConnection))
			{
				if ([m_s4HttpConnection openConnectionForRequest: m_urlRequest
														delegate: self
											   writeToFileAtPath: m_filePath
														withName: m_fileName
													  useTmpFile: m_bUseTmpFile] == YES)
				{
					// KVO compliance
					[self willChangeValueForKey: @"isExecuting"];
					m_bExecuting = YES;
					[self didChangeValueForKey: @"isExecuting"];
				}

				// release the request now
				S4_NS_SAFE_RELEASE(m_urlRequest);
			}
		}
	}
	else	// iOS 4 bug fix
	{
		[self performSelectorOnMainThread: @selector(start) withObject: nil waitUntilDone: NO];
	}
}


//============================================================================
//	S4DownloadInvokeOperation :: main
//============================================================================
- (void)main
{
}


//============================================================================
//	S4DownloadInvokeOperation :: cancel
//============================================================================
- (void)cancel
{
	[super cancel];
	[self willChangeValueForKey: @"isCancelled"];
	m_bCancelled = YES;
	[self didChangeValueForKey: @"isCancelled"];
}


//============================================================================
//	S4DownloadInvokeOperation :: isCancelled
//============================================================================
- (BOOL)isCancelled
{
	return (m_bCancelled);
}


//============================================================================
//	S4DownloadInvokeOperation :: isConcurrent
//============================================================================
- (BOOL)isConcurrent
{
	return (YES);
}


//============================================================================
//	S4DownloadInvokeOperation :: isExecuting
//============================================================================
- (BOOL)isExecuting
{
	return (m_bExecuting);
}


//============================================================================
//	S4DownloadInvokeOperation :: isFinshed
//============================================================================
- (BOOL)isFinished
{
	return (m_bFinished);
}


//============================================================================
//	S4DownloadInvokeOperation :: isReady
//============================================================================
- (BOOL)isReady
{
	if ([super isReady])
	{
		return (m_bPrepCalled);
	}
	return (NO);
}



/********************************** Private Methods **********************************/

//============================================================================
//	S4DownloadInvokeOperation () :: setExitFlags
//============================================================================
- (void)setExitFlags
{
	// KVO compliance
	[self willChangeValueForKey: @"isFinished"];
	[self willChangeValueForKey: @"isExecuting"];

	// update internal values
	m_bFinished = YES;
	m_bExecuting = NO;

	// KVO compliance
	[self didChangeValueForKey: @"isExecuting"];
	[self didChangeValueForKey: @"isFinished"];
}


//============================================================================
//	S4DownloadInvokeOperation () :: invokeProgressCallback:
//============================================================================
- (void)invokeProgressCallback: (NSNumber *)percent
{
	NSNumber					*localNumberObject;
	
	localNumberObject = percent;
	[m_progressInvocation setArgument: &localNumberObject atIndex: 2];
	[m_progressInvocation invoke];
}


//============================================================================
//	S4DownloadInvokeOperation () :: invokeErrorCallback:
//============================================================================
- (void)invokeErrorCallback: (NSError *)returnError
{
	NSError					*localErrorObject;
	NSNumber				*localNumberObject;

	// delete the file
	if ([S4FileUtilities fileExists: m_localFileNamePath])
	{
		[S4FileUtilities deleteDiskItemAtPath: m_localFileNamePath];
	}

	localErrorObject = returnError;
	[m_errInvocation setArgument: &localErrorObject atIndex: 2];
	[m_errInvocation invoke];

	// also call completion invocation
	localNumberObject = [NSNumber numberWithLongLong: 0];
	[m_completionInvocation setArgument: &localNumberObject atIndex: 2];
	[m_completionInvocation invoke];

	// it is OK if we are released now
	[self release];
}


//============================================================================
//	S4DownloadInvokeOperation () :: invokeCompletionCallback:
//============================================================================
- (void)invokeCompletionCallback: (NSNumber *)bytesRead
{
	NSNumber					*localNumberObject;
	
	localNumberObject = bytesRead;
	[m_completionInvocation setArgument: &localNumberObject atIndex: 2];
	[m_completionInvocation invoke];

	// it is OK if we are released now
	[self release];
}


//============================================================================
//	S4DownloadInvokeOperation () :: deleteUnfinishedFile
//============================================================================
- (void)deleteUnfinishedFile
{
	if ([S4FileUtilities fileExists: m_localFileNamePath])
	{
		[S4FileUtilities deleteDiskItemAtPath: m_localFileNamePath];
	}

	// it is OK if we are released now
	[self release];
}


// ================================== S4HttpConnection delegate methods ==========================================

//============================================================================
//	S4DownloadInvokeOperation :: httpConnection:receivedData:
//============================================================================
- (BOOL)httpConnection: (S4HttpConnection *)connection receivedData: (NSData *)data
{
	float				pct;
	NSNumber			*percentNumber;
	BOOL				bResult;

	// check to see if we are cancelled
	if ([self isCancelled] == NO)
	{
		if (IS_NOT_NULL(m_progressInvocation))
		{
			m_bytesDownloaded += [data length];
			pct = (float)m_bytesDownloaded/m_fileSize;
			percentNumber = [NSNumber numberWithFloat: pct];
			[self performSelectorOnMainThread: @selector(invokeProgressCallback:) 
								   withObject: percentNumber 
								waitUntilDone: NO];
		}
		bResult = YES;
	}
	else
	{
		// cancelled so delete the file
		[self performSelectorOnMainThread: @selector(deleteUnfinishedFile) withObject: nil waitUntilDone: NO];

		// and set the exit flags
		[self setExitFlags];

		// this tells the S4HttpConnection to cancel its underlying NSURLConnection
		bResult = NO;
	}
	return (bResult);
}


//============================================================================
//	S4DownloadInvokeOperation :: httpConnection:failedWithError:
//============================================================================
- (void)httpConnection: (S4HttpConnection *)connection failedWithError: (NSError *)error
{
	if (IS_NOT_NULL(m_errInvocation))
	{
		[self performSelectorOnMainThread: @selector(invokeErrorCallback:) withObject: error waitUntilDone: NO];
	}
	[self setExitFlags];
}


//============================================================================
//	S4DownloadInvokeOperation :: httpConnection:completedWithData:
//============================================================================
- (void)httpConnection: (S4HttpConnection *)connection completedWithData: (NSMutableData *)data
{
	NSNumber			*bytesRead;

	bytesRead = [NSNumber numberWithLongLong: m_bytesDownloaded];
	[self performSelectorOnMainThread: @selector(invokeCompletionCallback:) withObject: bytesRead waitUntilDone: NO];
	[self setExitFlags];
}


//============================================================================
//	S4DownloadInvokeOperation :: httpConnection:receivedResponse:
//============================================================================
- (BOOL)httpConnection: (S4HttpConnection *)connection receivedResponse: (NSURLResponse *)response
{
	long long				contentLength;
	BOOL					bResult;

	// check to see if we are cancelled
	if ([self isCancelled] == NO)
	{
		if ([response isKindOfClass: [NSHTTPURLResponse class]] == YES)
		{
			contentLength = [(NSHTTPURLResponse *)response expectedContentLength];
			if (NSURLResponseUnknownLength != contentLength)
			{
				m_fileSize = contentLength;
			}
		}
		bResult = YES;
	}
	else
	{
		// cancelled so delete the file
		[self performSelectorOnMainThread: @selector(deleteUnfinishedFile) withObject: nil waitUntilDone: NO];

		// and set the exit flags
		[self setExitFlags];

		// this tells the S4HttpConnection to cancel its underlying NSURLConnection
		bResult = NO;
	}
	return (bResult);
}

@end
