/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the S4 iOS Libraries.
 *
 * The Initial Developer of the Original Code is
 * Michael Papp dba SeaStones Software Company.
 * All software created by the Initial Developer are Copyright (C) 2008-2012
 * the Initial Developer. All Rights Reserved.
 *
 * Original Author:
 *			Michael Papp, San Francisco, CA, USA
 *
 * ***** END LICENSE BLOCK ***** */

/* ***** FILE BLOCK ******
 *
 * Name:		S4FileDownloader.m
 * Module:		UI
 * Library:		S4 iOS Libraries
 *
 * ***** FILE BLOCK *****/


// ================================== Includes =========================================

#import "S4FileDownloader.h"
#import "S4DownloadInvokeOperation.h"
#import "S4NetUtilities.h"
#import "S4OperationsManager.h"


// =================================== Defines =========================================



// ================================== Typedefs =========================================



// =================================== Globals =========================================

// ALL S4 LIBS SHOULD DEFINE THIS:
S4_INTERN_CONSTANT_NSSTR						S4ImageDownloaderErrorDomain = @"S4ImageDownloaderErrorDomain";


// ============================= Forward Declarations ==================================



// ================================== Inlines ==========================================



// ========================= Begin Class S4FileDownloader () ===========================

@interface S4FileDownloader ()
{
	id													m_userObject;
	NSObject <S4FileDownloaderDelegate>					*m_delegate;
	NSString											*m_urlString;
	NSOperationQueue									*m_operationQueue;
}

- (void)progressInvocationMethod: (NSNumber *)percent;
- (void)errorInvocationMethod: (id)error;
- (void)completionInvocationMethod: (NSNumber *)bytesRead;

@end




// =================== Begin Class S4FileDownloader ==========================

@implementation S4FileDownloader


//============================================================================
//	S4FileDownloader :: properties
//============================================================================
@synthesize urlString = m_urlString;
@synthesize operationQueue = m_operationQueue;


//============================================================================
//	S4FileDownloader :: downloaderForFileAtURL
//============================================================================
+ (id)downloaderForFileAtURL: (NSString *)urlStr
					  toPath: (NSString *)path
					fileName: (NSString *)name
				withDelegate: (NSObject <S4FileDownloaderDelegate> *)delegate
				  useTmpFile: (BOOL)bUseTmpFile
					 context: (id)userObject
		 usingOperationQueue: (NSOperationQueue *)queue
{
	S4FileDownloader					*tmpDownloader;
	S4FileDownloader					*downloaderResult = nil;

	tmpDownloader = [[S4FileDownloader alloc] init];
	if (IS_NOT_NULL(tmpDownloader))
	{
		tmpDownloader.operationQueue = queue;
		if ([tmpDownloader downloadFileAtURL: urlStr
									  toPath: path
									fileName: name
								withDelegate: delegate
								  useTmpFile: bUseTmpFile
									 context: userObject] == YES)
		{
			downloaderResult = [tmpDownloader autorelease];
		}
		else
		{
			[tmpDownloader release];
		}
	}
	return (downloaderResult);
}


//============================================================================
//	S4FileDownloader :: init
//============================================================================
- (id)init
{
	self = [super init];
	if (nil != self)
	{
		m_userObject = nil;
		m_delegate = nil;
		m_urlString = nil;
		m_operationQueue = nil;
	}
	return (self);
}


//============================================================================
//	S4FileDownloader :: dealloc
//============================================================================
- (void)dealloc
{
	S4_NS_SAFE_RELEASE(m_userObject);
	S4_NS_SAFE_RELEASE(m_delegate);
	S4_NS_SAFE_RELEASE(m_urlString);

	[super dealloc];
}


//============================================================================
//	S4FileDownloader :: downloadFileAtURL:
//============================================================================
- (BOOL)downloadFileAtURL: (NSString *)urlStr
				   toPath: (NSString *)path
				 fileName: (NSString *)name
			 withDelegate: (NSObject <S4FileDownloaderDelegate> *)delegate
			   useTmpFile: (BOOL)bUseTmpFile
				  context: (id)userObject
{
	NSURLRequest						*urlRequest;
	S4DownloadInvokeOperation			*invokeOperation;
	NSMethodSignature					*progressMethodSignature;
	NSInvocation						*progressInv;
	NSMethodSignature					*errMethodSignature;
	NSInvocation						*errorInv;
	NSMethodSignature					*completionMethodSignature;
	NSInvocation						*completionInv;
	BOOL								bResult = NO;

	if ((STR_NOT_EMPTY(urlStr)) && (IS_NOT_NULL(delegate)) &&
		(YES == [delegate conformsToProtocol: @protocol(S4FileDownloaderDelegate)]))
	{
		// retain the objects we need later
		m_userObject = [userObject retain];
		m_delegate = [delegate retain];
		m_urlString = [urlStr copy];

		// now set up the HTTP download
		urlRequest = [S4NetUtilities createRequestForPath: urlStr
												 useCache: NO
										  timeoutInterval: 0.0
												 postData: nil
											   dataIsForm: NO
											handleCookies: YES];
		// create a properly escaped NSURL from the string params
		if (IS_NOT_NULL(urlRequest))
		{
			invokeOperation = [[S4DownloadInvokeOperation alloc] init];
			if (IS_NOT_NULL(invokeOperation))
			{
				progressMethodSignature = [self methodSignatureForSelector: @selector(progressInvocationMethod:)];
				progressInv = [NSInvocation invocationWithMethodSignature: progressMethodSignature];
				[progressInv setSelector: @selector(progressInvocationMethod:)];
				[progressInv setTarget: self];

				errMethodSignature = [self methodSignatureForSelector: @selector(errorInvocationMethod:)];
				errorInv = [NSInvocation invocationWithMethodSignature: errMethodSignature];
				[errorInv setSelector: @selector(errorInvocationMethod:)];
				[errorInv setTarget: self];

				completionMethodSignature = [self methodSignatureForSelector: @selector(completionInvocationMethod:)];
				completionInv = [NSInvocation invocationWithMethodSignature: completionMethodSignature];
				[completionInv setSelector: @selector(completionInvocationMethod:)];
				[completionInv setTarget: self];

				if ([invokeOperation prepareForRequest: urlRequest
									progressInvocation: progressInv
										 errInvocation: errorInv
								  completionInvocation: completionInv
											  filePath: path
											  fileName: name
											useTmpFile: bUseTmpFile] == YES)
				{
					if (IS_NOT_NULL(m_operationQueue))
					{
						[m_operationQueue addOperation: invokeOperation];
					}
					else
					{
						[[S4OperationsManager defaultManager] addOperationToNetworkQueue: invokeOperation];
					}
					bResult = YES;
				}
				[invokeOperation release];
			}
			[urlRequest release];
		}
	}
	return (bResult);
}


//============================================================================
//	S4FileDownloader (PrivateImpl) :: progressInvocationMethod:
//============================================================================
- (void)progressInvocationMethod: (NSNumber *)percent
{
	if (YES == [m_delegate respondsToSelector: @selector(fileDownloader:percentDownloaded:context:)])
	{
		[m_delegate fileDownloader: self percentDownloaded: percent context: m_userObject];
	}
}


//============================================================================
//	S4FileDownloader (PrivateImpl) :: errorInvocationMethod:
//============================================================================
- (void)errorInvocationMethod: (NSError *)error
{
	if (YES == [m_delegate respondsToSelector: @selector(fileDownloader:didFailWithError:context:)])
	{
		[m_delegate fileDownloader: self didFailWithError: error context: m_userObject];
	}
}


//============================================================================
//	S4FileDownloader (PrivateImpl) :: completionInvocationMethod:
//============================================================================
- (void)completionInvocationMethod: (NSNumber *)bytesRead
{
	[m_delegate fileDownloader: self completedForNumBytes: bytesRead context: m_userObject];
}

@end
