/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the S4 iOS Libraries.
 *
 * The Initial Developer of the Original Code is
 * Michael Papp dba SeaStones Software Company.
 * All software created by the Initial Developer are Copyright (C) 2008-2012
 * the Initial Developer. All Rights Reserved.
 *
 * Original Author:
 *			Michael Papp, San Francisco, CA, USA
 *
 * ***** END LICENSE BLOCK ***** */

/* ***** FILE BLOCK ******
 *
 * Name:		S4FileProgressDownloader.m
 * Module:		Network
 * Library:		S4 iOS Libraries
 *
 * ***** FILE BLOCK *****/


// ================================== Includes =========================================

#import "S4FileProgressDownloader.h"
#import "S4CommonDefines.h"
#import "S4FileUtilities.h"
#import "S4NetUtilities.h"


// ================================== Defines ==========================================



// ================================== Typedefs ==========================================



// ================================== Globals ==========================================



// ============================= Forward Declarations ==================================



// ================================== Inlines ==========================================



// ====================== Begin Class S4FileProgressDownloader () ======================

@interface S4FileProgressDownloader ()
{
	NSObject <S4FileProgressLoaderDelegate>				*m_delegate;
	NSString											*m_fileName;
	NSString											*m_localFilePath;
	NSString											*m_fileNameAndPath;
	unsigned long long									m_bytesDownloaded;
	unsigned long long									m_totalBytes;
	S4HttpConnection									*m_s4HttpConnection;
	UIActivityIndicatorView								*m_activityIndicatorView;
	UIProgressView										*m_progressView;
	S4FileDownloaderUIViewType							m_viewTypeInUse;
}

@end


// ================= Begin Class S4FileProgressDownloader (PrivateImpl) ================

@interface S4FileProgressDownloader (PrivateImpl)

- (void)downloadStarted;
- (void)downloadUpdate: (NSNumber *)percentDownloaded;
- (void)downloadCompleted;
- (void)downloadError: (NSError *)error;
- (void)usingViewOfType;

- (void)callDelegateStarted;
- (void)callDelegateUpdated: (NSNumber *)percentDownloaded;
- (void)callDelegateCompleted;
- (void)callDelegateError: (NSError *)error;

@end




@implementation S4FileProgressDownloader (PrivateImpl)

//============================================================================
//	S4FileProgressDownloader (PrivateImpl) :: downloadStarted
//============================================================================
- (void)downloadStarted
{
	// handle the UI if one or both properties are set
	if ((IS_NOT_NULL(m_activityIndicatorView)) && (ActivityIndicatorViewType == m_viewTypeInUse))
	{
		if (NO == [m_activityIndicatorView isAnimating])
		{
			[m_activityIndicatorView startAnimating];
		}
	}
	else if ((IS_NOT_NULL(m_progressView)) && (ProgressViewType == m_viewTypeInUse))
	{
		m_progressView.progress = (float)0.0;
	}

	// and call the delegate methods always
	[m_delegate downloader: self startedDownloadOfFile: m_fileName];
}


//============================================================================
//	S4FileProgressDownloader (PrivateImpl) :: downloadUpdate:
//============================================================================
- (void)downloadUpdate: (NSNumber *)percentDownloaded
{
	// handle the UI if one or both properties are set
	if ((IS_NOT_NULL(m_activityIndicatorView)) && (ActivityIndicatorViewType == m_viewTypeInUse))
	{
		if (NO == [m_activityIndicatorView isAnimating])
		{
			[m_activityIndicatorView startAnimating];
		}
	}
	else if ((IS_NOT_NULL(m_progressView)) && (ProgressViewType == m_viewTypeInUse))
	{
		m_progressView.progress = [percentDownloaded floatValue];
	}

	// and call the delegate methods always
	[m_delegate downloader: self updatedProgressForFile: m_fileName atPercentage: [percentDownloaded floatValue]];
}


//============================================================================
//	S4FileProgressDownloader (PrivateImpl) :: downloadCompleted
//============================================================================
- (void)downloadCompleted
{
	// handle the UI if one or both properties are set
	if ((IS_NOT_NULL(m_activityIndicatorView)) && (ActivityIndicatorViewType == m_viewTypeInUse))
	{
		if (YES == [m_activityIndicatorView isAnimating])
		{
			[m_activityIndicatorView stopAnimating];
		}
	}
	else if ((IS_NOT_NULL(m_progressView)) && (ProgressViewType == m_viewTypeInUse))
	{
		m_progressView.progress = (float)1.0;
	}

	// and call the delegate methods always
	[m_delegate downloader: self updatedProgressForFile: m_fileName atPercentage: (float)1.0];
	[m_delegate downloader: self completedDownloadOfFile: m_fileName];
}


//============================================================================
//	S4FileProgressDownloader (PrivateImpl) :: downloadError:
//============================================================================
- (void)downloadError: (NSError *)error
{
	// handle the UI if set
	if ((IS_NOT_NULL(m_activityIndicatorView)) && (ActivityIndicatorViewType == m_viewTypeInUse))
	{
		if (YES == [m_activityIndicatorView isAnimating])
		{
			[m_activityIndicatorView stopAnimating];
		}
	}
	
	// and call the delegate methods always
	[m_delegate downloader: self didFailForFile: m_fileName withError: error];
	
	// delete the file
	if ([S4FileUtilities fileExists: m_fileNameAndPath])
	{
		[S4FileUtilities deleteDiskItemAtPath: m_fileNameAndPath];
	}
}


//============================================================================
//	S4FileProgressDownloader (PrivateImpl) :: usingViewOfType
//============================================================================
- (void)usingViewOfType
{
	if ([m_delegate respondsToSelector: @selector(downloader:usingUpdateViewOfType:)])
	{
		[m_delegate downloader: self usingUpdateViewOfType: m_viewTypeInUse];
	}
}


//============================================================================
//	S4FileProgressDownloader (PrivateImpl) :: callDelegateStarted
//============================================================================
- (void)callDelegateStarted
{
	[self performSelectorOnMainThread: @selector(downloadStarted)
						   withObject: nil
						waitUntilDone: [NSThread isMainThread]];
}


//============================================================================
//	S4FileProgressDownloader (PrivateImpl) :: callDelegateUpdated
//============================================================================
- (void)callDelegateUpdated: (NSNumber *)percentDownloaded
{
	[self performSelectorOnMainThread: @selector(downloadUpdate:)
						   withObject: percentDownloaded
						waitUntilDone: [NSThread isMainThread]];
}


//============================================================================
//	S4FileProgressDownloader (PrivateImpl) :: callDelegateCompleted
//============================================================================
- (void)callDelegateCompleted
{
	[self performSelectorOnMainThread: @selector(downloadCompleted)
						   withObject: nil
						waitUntilDone: [NSThread isMainThread]];
}


//============================================================================
//	S4FileProgressDownloader (PrivateImpl) :: callDelegateError:
//============================================================================
- (void)callDelegateError: (NSError *)error
{
	[self performSelectorOnMainThread: @selector(downloadError:)
						   withObject: error
						waitUntilDone: [NSThread isMainThread]];
}


//============================================================================
//	S4FileProgressDownloader (PrivateImpl) :: callDelegateUsingViewOfType
//============================================================================
- (void)callDelegateUsingViewOfType
{
	[self performSelectorOnMainThread: @selector(usingViewOfType)
						   withObject: nil
						waitUntilDone: [NSThread isMainThread]];
}

@end




// =================== Begin Class S4FileProgressDownloader ==================

@implementation S4FileProgressDownloader


//============================================================================
//	S4FileProgressDownloader :: properties
//============================================================================
@synthesize activityIndicatorView = m_activityIndicatorView;
@synthesize progressView = m_progressView;


//============================================================================
//	S4FileProgressDownloader :: init
//============================================================================
- (id)init
{
	id					idResult = nil;

	self = [super init];
	if (nil != self)
	{
		m_delegate = nil;
		m_fileName = nil;
		m_localFilePath = nil;
		m_fileNameAndPath = nil;
		m_bytesDownloaded = 0;
		m_totalBytes = 0;
		m_s4HttpConnection = nil;
		m_activityIndicatorView = nil;
		m_progressView = nil;
		m_viewTypeInUse = UndeterminedViewType;

		idResult = self;
	}
	return (idResult);
}


//============================================================================
//	S4FileProgressDownloader :: dealloc
//============================================================================
- (void)dealloc
{
	S4_NS_SAFE_RELEASE(m_delegate)
	S4_NS_SAFE_RELEASE(m_fileName)
	S4_NS_SAFE_RELEASE(m_localFilePath)
	S4_NS_SAFE_RELEASE(m_fileNameAndPath)
	S4_NS_SAFE_RELEASE(m_s4HttpConnection)
	S4_NS_SAFE_RELEASE(m_activityIndicatorView)
	S4_NS_SAFE_RELEASE(m_progressView)

	[super dealloc];
}


//============================================================================
//	S4FileProgressDownloader :: downloadFile:
//============================================================================
- (BOOL)downloadFile: (NSString *)fileName
			  toPath: (NSString *)dstFilePath
		  fromURLStr: (NSString *)urlStr
		 forDelegate: (NSObject <S4FileProgressLoaderDelegate> *)delegate
		  useTmpFile: (BOOL)bUseTmpFile
{
	NSURLRequest				*request;
	BOOL						bResult = NO;

	if ((STR_NOT_EMPTY(fileName)) && (STR_NOT_EMPTY(dstFilePath)) && (STR_NOT_EMPTY(urlStr)) &&
		(IS_NOT_NULL(delegate)) && (YES == [delegate conformsToProtocol: @protocol(S4FileProgressLoaderDelegate)]))
	{
		S4_SAFE_RETAIN(m_delegate, delegate)
		S4_SAFE_COPY(m_fileName, fileName)
		S4_SAFE_COPY(m_localFilePath, dstFilePath)
		S4_NS_SAFE_RELEASE(m_fileNameAndPath)
		m_fileNameAndPath = [[m_localFilePath stringByAppendingPathComponent: m_fileName] retain];
		m_bytesDownloaded = 0;
		m_totalBytes = 0;
		m_viewTypeInUse = UndeterminedViewType;

		if (NO == [S4FileUtilities fileExists: m_fileNameAndPath])
		{
			// create a request
			request = [S4NetUtilities createRequestForPath: urlStr
												  useCache: YES
										   timeoutInterval: 360.0
												  postData: nil
												dataIsForm: NO
											 handleCookies: YES];
			if (IS_NOT_NULL(request))
			{
				// create an S4HttpConnection for writing to file
				m_s4HttpConnection = [[S4HttpConnection alloc] init];
				if (IS_NOT_NULL(m_s4HttpConnection))
				{
					if ([m_s4HttpConnection openConnectionForRequest: request
															delegate: self
												   writeToFileAtPath: m_localFilePath
															withName: m_fileName
														  useTmpFile: bUseTmpFile])
					{
						bResult = YES;
						[self callDelegateStarted];
					}
				}
			}
		}
		else
		{
			[self callDelegateCompleted];
			bResult = YES;
		}
	}
	return (bResult);
}




// ================================== S4HttpConnection delegate methods ==========================================

//============================================================================
//	S4FileProgressDownloader :: completedWithData:
//============================================================================
- (void)httpConnection: (S4HttpConnection *)connection completedWithData: (NSMutableData *)data
{
	[self callDelegateCompleted];
}


//============================================================================
//	S4FileProgressDownloader :: receivedData:
//============================================================================
- (BOOL)httpConnection: (S4HttpConnection *)connection receivedData: (NSData *)data
{
	float							pctFloat = 0.0f;
	NSNumber						*pctNumber;

	if (ProgressViewType == m_viewTypeInUse)
	{
		m_bytesDownloaded += data.length;
		pctFloat = (m_bytesDownloaded * 1.0f)/(m_totalBytes * 1.0f);
	}
	pctNumber = [NSNumber numberWithFloat: pctFloat];
	[self callDelegateUpdated: pctNumber];
	return (YES);
}


//============================================================================
//	S4FileProgressDownloader :: failedWithError:
//============================================================================
- (void)httpConnection: (S4HttpConnection *)connection failedWithError: (NSError *)error
{
	[self callDelegateError: error];
}


//============================================================================
//	S4FileProgressDownloader :: receivedResponse:
//============================================================================
- (BOOL)httpConnection: (S4HttpConnection *)connection receivedResponse: (NSURLResponse *)response
{
	NSHTTPURLResponse				*httpResponse;
	long long						contentLength;
	
	httpResponse = (NSHTTPURLResponse *)response;
	contentLength = [httpResponse expectedContentLength];
	if (NSURLResponseUnknownLength != contentLength)
	{
		m_totalBytes = contentLength;
		m_viewTypeInUse = ProgressViewType;
	}
	else
	{
		m_totalBytes = 0;
		m_viewTypeInUse = ActivityIndicatorViewType;
	}
	[self callDelegateUsingViewOfType];
	return (YES);
}

@end
