/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the S4 iOS Libraries.
 *
 * The Initial Developer of the Original Code is
 * Michael Papp dba SeaStones Software Company.
 * All software created by the Initial Developer are Copyright (C) 2008-2009
 * the Initial Developer. All Rights Reserved.
 *
 * Original Author:
 *			Michael Papp, San Francisco, CA, USA
 *
 * ***** END LICENSE BLOCK ***** */

/* ***** FILE BLOCK ******
 *
 * Name:		S4ImageFetcher.m
 * Module:		UI
 * Library:		S4 iOS Libraries
 *
 * ***** FILE BLOCK *****/


// ================================== Includes =========================================

#import "S4ImageFetcher.h"
#import "S4HttpInvokeOperation.h"
#import "S4NetUtilities.h"
#import "S4OperationsManager.h"


// =================================== Defines =========================================



// ================================== Typedefs =========================================



// =================================== Globals =========================================

// ALL S4 LIBS SHOULD DEFINE THIS:
S4_INTERN_CONSTANT_NSSTR						S4ImageFetcherErrorDomain = @"S4ImageFetcherErrorDomain";


// ============================= Forward Declarations ==================================



// ================================== Inlines ==========================================



// ========================== Begin Class S4ImageFetcher () ============================

@interface S4ImageFetcher ()
{
	id													m_userObject;
	NSObject <S4ImageFetcherDelegate> *					m_delegate;
	NSString											*m_imageTag;
	NSOperationQueue									*m_operationQueue;
}

- (void)dataInvocationMethod: (NSData *)data;
- (void)errorInvocationMethod: (NSError *)error;

@end




// ========================== Begin Class S4ImageFetcher ===============================

@implementation S4ImageFetcher

//============================================================================
//	S4ImageFetcher :: properties
//============================================================================
@synthesize operationQueue = m_operationQueue;
@synthesize imageTag = m_imageTag;


//============================================================================
//	S4ImageFetcher :: fetcherForImageAtURL
//============================================================================
+ (id)fetcherForImageAtURL: (NSString *)urlStr
			  withDelegate: (NSObject <S4ImageFetcherDelegate> *)delegate
				   context: (id)userObject
	   usingOperationQueue: (NSOperationQueue *)queue
{
	S4ImageFetcher						*tmpFetcher;
	S4ImageFetcher						*fetcherResult = nil;

	tmpFetcher = [[S4ImageFetcher alloc] init];
	if (IS_NOT_NULL(tmpFetcher))
	{
		tmpFetcher.operationQueue = queue;
		if ([tmpFetcher loadImageAtURL: urlStr withDelegate: delegate context: userObject] == YES)
		{
			fetcherResult = [tmpFetcher autorelease];
		}
		else
		{
			[tmpFetcher release];
		}
	}
	return (fetcherResult);
}


//============================================================================
//	S4ImageFetcher :: init
//============================================================================
- (id)init
{
	self = [super init];
	if (nil != self)
	{
		m_userObject = nil;
		m_delegate = nil;
		m_imageTag = nil;
		m_operationQueue = nil;
	}
	return (self);
}


//============================================================================
//	S4ImageFetcher :: dealloc
//============================================================================
- (void)dealloc
{
	S4_NS_SAFE_RELEASE(m_userObject);
	S4_NS_SAFE_RELEASE(m_delegate);
	S4_NS_SAFE_RELEASE(m_imageTag);

	[super dealloc];
}


//============================================================================
//	S4ImageFetcher :: loadImageAtURL:
//============================================================================
- (BOOL)loadImageAtURL: (NSString *)urlStr
		  withDelegate: (NSObject <S4ImageFetcherDelegate> *)delegate
			   context: (id)userObject
{
	NSURLRequest					*urlRequest;
	S4HttpInvokeOperation			*invokeOperation;
	NSMethodSignature				*dataMethodSignature;
	NSInvocation					*dataInv;
	NSMethodSignature				*errMethodSignature;
	NSInvocation					*errInv;
	BOOL							bResult = NO;
	
	if ((STR_NOT_EMPTY(urlStr)) && (IS_NOT_NULL(delegate)) &&
		([delegate conformsToProtocol: @protocol(S4ImageFetcherDelegate)] == YES))
	{
		// retain the objects we need later
		m_userObject = [userObject retain];
		m_delegate = [delegate retain];
		m_imageTag = [urlStr copy];		// imageURL.absoluteString

		// now set up the HTTP download
		urlRequest = [S4NetUtilities createRequestForPath: urlStr
												 useCache: NO
										  timeoutInterval: 0.0
												 postData: nil
											   dataIsForm: NO
											handleCookies: YES];
		// create a properly escaped NSURL from the string params
		if (nil != urlRequest)
		{
			invokeOperation = [[S4HttpInvokeOperation alloc] init];
			if (nil != invokeOperation)
			{
				dataMethodSignature = [self methodSignatureForSelector: @selector(dataInvocationMethod:)];
				dataInv = [NSInvocation invocationWithMethodSignature: dataMethodSignature];
				[dataInv setSelector: @selector(dataInvocationMethod:)];
				[dataInv setTarget: self];

				errMethodSignature = [self methodSignatureForSelector: @selector(errorInvocationMethod:)];
				errInv = [NSInvocation invocationWithMethodSignature: errMethodSignature];
				[errInv setSelector: @selector(errorInvocationMethod:)];
				[errInv setTarget: self];
				
				if ([invokeOperation prepareForRequest: urlRequest dataInvocation: dataInv errInvocation: errInv])
				{
					if (IS_NOT_NULL(m_operationQueue))
					{
						[m_operationQueue addOperation: invokeOperation];
					}
					else
					{
						[[S4OperationsManager defaultManager] addOperationToNetworkQueue: invokeOperation];
					}
					bResult = YES;
				}
				[invokeOperation release];
			}
		}
	}
	return (bResult);
}


//============================================================================
//	S4ImageFetcher () :: dataInvocationMethod:
//============================================================================
- (void)dataInvocationMethod: (NSData *)data
{
	UIImage									*image;
	NSMutableDictionary						*userDict;
	S4ImageFetcherError						errorCode = S4ImageFetcherNoError;
	NSString								*localizedDescription;
	NSString								*localizedFailureReason;
	NSError									*error;
	
	if (IS_NOT_NULL(data))
	{
		image = [UIImage imageWithData: data];
		if (IS_NOT_NULL(image))
		{
			[m_delegate imageFetcher: self loadedImage: image context: m_userObject];
		}
		else
		{
			localizedDescription = @"Invalid image";
			localizedFailureReason = @"The data sent by the server could not be used for an image";
			errorCode = S4ImageFetcherInvalidImageError;
		}
	}
	else
	{
		localizedDescription = @"Invalid data";
		localizedFailureReason = @"The server responded with invalid data";
		errorCode = S4ImageFetcherInvalidDataError;
	}
	
	if (S4ImageFetcherNoError != errorCode)
	{
		// create an NSDictionary to hold values for the NSError
		userDict = [NSMutableDictionary dictionaryWithCapacity: 2];
		if (IS_NOT_NULL(userDict))
		{
			// set the keys in the error dictionary
			[userDict setObject: localizedDescription forKey: NSLocalizedDescriptionKey];
			[userDict setObject: localizedFailureReason forKey: NSLocalizedFailureReasonErrorKey];
		}
		
		// create the NSError
		error = [NSError errorWithDomain: S4ImageFetcherErrorDomain code: (NSInteger)errorCode userInfo: userDict];
		[self errorInvocationMethod: error];
	}
}


//============================================================================
//	S4ImageFetcher () :: errorInvocationMethod:
//============================================================================
- (void)errorInvocationMethod: (NSError *)error
{
	if ([m_delegate respondsToSelector: @selector(imageFetcher:didFailWithError:context:)] == YES)
	{
		[m_delegate imageFetcher: self didFailWithError: error context: m_userObject];
	}
}

@end
