/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the S4 iOS Libraries.
 *
 * The Initial Developer of the Original Code is
 * Michael Papp dba SeaStones Software Company.
 * All software created by the Initial Developer are Copyright (C) 2008-2012
 * the Initial Developer. All Rights Reserved.
 *
 * Original Author:
 *			Michael Papp, San Francisco, CA, USA
 *
 * ***** END LICENSE BLOCK ***** */

/* ***** FILE BLOCK ******
 *
 * Name:		S4OverlayViewController.m
 * Module:		UI
 * Library:		S4 iOS UI Libraries
 *
 * ***** FILE BLOCK *****/


// ================================== Includes =========================================

#import "S4OverlayViewController.h"
#import "S4RoundedRectView.h"
#import "S4SingletonClass.h"
#import <QuartzCore/QuartzCore.h>
#import "S4CommonDefines.h"


// =================================== Defines =========================================

#define TRANSITION_SCALE				1.2f


// ================================== Typedefs =========================================



// =================================== Globals =========================================



// ============================= Forward Declarations ==================================



// ================================== Inlines ==========================================



// ====================== Begin Class S4OverlayViewController () =======================

@interface S4OverlayViewController ()

@property (readwrite, strong, nonatomic) UIWindow								*overlayWindow;
@property (readwrite, strong, nonatomic) S4RoundedRectView						*hudView;
@property (readwrite, strong, nonatomic) UIActivityIndicatorView				*activityIndicator;
@property (readwrite, strong, nonatomic) UILabel								*messageLabel;
@property (readwrite, strong, nonatomic) UIImageView							*imageView;
@property (readwrite, assign, nonatomic) BOOL									visible;

- (void)oneTimeInit;
- (CGImageRef)gradientBackgroundImage;
- (CGImageRef)blackBackgroundImage;
- (CGImageRef)clearBackgroundImage;
- (void)moveToPoint: (CGPoint)newCenter rotateAngle: (CGFloat)angle;
- (void)updateHudView: (NSNotification *)notification;
- (void)registerNotifications;
- (CGFloat)visibleKeyboardHeight;

@end



// ======================== Begin Class S4OverlayViewController ========================

@implementation S4OverlayViewController
{
	UIWindow									*m_overlayWindow;
	S4RoundedRectView							*m_hudView;
	UIActivityIndicatorView						*m_activityIndicator;
	UILabel										*m_messageLabel;
	UIImageView									*m_imageView;
	BOOL										m_bIsVisible;
}


///////////////////////////////////// START SINGLETON METHODS /////////////////////////////////////


SYNTHESIZE_SINGLETON_CLASS(S4OverlayViewController)


///////////////////////////////////// INSTANCE METHODS /////////////////////////////////////

//============================================================================
//	S4OverlayViewController :: properties
//============================================================================
@synthesize overlayWindow = m_overlayWindow;
@synthesize hudView = m_hudView;
@synthesize activityIndicator = m_activityIndicator;
@synthesize messageLabel = m_messageLabel;
@synthesize imageView = m_imageView;
@synthesize visible = m_bIsVisible;


//============================================================================
//	S4OverlayViewController () :: oneTimeInit
//============================================================================
- (void)oneTimeInit
{
	self.overlayWindow = nil;
	self.hudView = nil;
	self.activityIndicator = nil;
	self.messageLabel = nil;
	self.imageView = nil;
	self.visible = NO;
}


//============================================================================
//	S4OverlayViewController :: loadView
//============================================================================
- (void)loadView
{
	self.overlayWindow = [[[UIWindow alloc] initWithFrame: [UIScreen mainScreen].bounds] autorelease];
	self.overlayWindow.autoresizingMask = UIViewAutoresizingFlexibleWidth | UIViewAutoresizingFlexibleHeight;
	self.overlayWindow.backgroundColor = [UIColor clearColor];
	self.overlayWindow.userInteractionEnabled = NO;
	self.overlayWindow.hidden = YES;
	// we want to show above the keyboard but below alerts so split the difference
	self.overlayWindow.windowLevel = (UIWindowLevelNormal + UIWindowLevelAlert) / 2.0f;

	// create the HUD View
	self.hudView = [[[S4RoundedRectView alloc] initWithFrame: CGRectZero] autorelease];
	self.hudView.center = CGPointMake(CGRectGetMidX(self.overlayWindow.bounds), CGRectGetMidY(self.overlayWindow.bounds));
	self.hudView.backgroundColor = [UIColor colorWithWhite: 0.2f alpha: 0.9f];
	self.hudView.autoresizingMask = (UIViewAutoresizingFlexibleBottomMargin | UIViewAutoresizingFlexibleTopMargin |
									 UIViewAutoresizingFlexibleRightMargin | UIViewAutoresizingFlexibleLeftMargin);

	// create the spinner view
	self.activityIndicator = [[[UIActivityIndicatorView alloc] initWithActivityIndicatorStyle: UIActivityIndicatorViewStyleWhiteLarge] autorelease];
	self.activityIndicator.hidesWhenStopped = YES;
	self.activityIndicator.bounds = CGRectMake(0, 0, 37, 37);
	[self.hudView addSubview: self.activityIndicator];

	// create the label that holds client messages
	self.messageLabel = [[[UILabel alloc] initWithFrame: CGRectZero] autorelease];
	self.messageLabel.font = [UIFont systemFontOfSize: [UIFont labelFontSize]];
	self.messageLabel.backgroundColor = [UIColor clearColor];
	self.messageLabel.textColor = [UIColor whiteColor];
	self.messageLabel.adjustsFontSizeToFitWidth = YES;
	self.messageLabel.textAlignment = NSTextAlignmentCenter;
	self.messageLabel.baselineAdjustment = UIBaselineAdjustmentAlignCenters;
	self.messageLabel.shadowColor = [UIColor blackColor];
	self.messageLabel.shadowOffset = CGSizeMake(0, -1);
	self.messageLabel.numberOfLines = 0;
	self.messageLabel.opaque = NO;
	[self.hudView addSubview: self.messageLabel];

	// create the image view
	self.imageView = [[[UIImageView alloc] initWithFrame: CGRectMake(0, 0, 28, 28)] autorelease];
	[self.hudView addSubview: self.imageView];

	// complete the view hierarchy
	[self.overlayWindow addSubview: self.hudView];
	self.view = self.overlayWindow;
}


//============================================================================
//	S4OverlayViewController :: setView:
//============================================================================
- (void)setView: (UIView *)view
{
	[super setView: view];
	if (nil == view)
	{
		self.activityIndicator = nil;
	}
}


//============================================================================
//	S4OverlayViewController :: hide
//============================================================================
- (void)didReceiveMemoryWarning
{
	// check the activity indicator so we don't accidentally load our view just to remove it
	if (self.activityIndicator != nil)
	{
		if (self.view.hidden == YES)
		{
			// it's not showing, we can safely remove it
			self.view = nil;
		}
	}
}


//============================================================================
//	S4OverlayViewController :: showWithMessage:
//============================================================================
- (void)showWithMessage: (NSString *)msgString
		withOverlayMode: (S4OverlayVCStyles)overlayMode
   allowUserInteraction: (BOOL)userInteractionEnabled
{
	BOOL					bPrevAnimationEnabled;

	if (NO == self.visible)
	{
		// if the view is still visible, then it must be in-flight, so don't touch it
		if (self.view.hidden == YES)
		{
			bPrevAnimationEnabled = [UIView areAnimationsEnabled];
			[UIView setAnimationsEnabled: NO];
			self.view.alpha = 0.0f;
			self.view.transform = CGAffineTransformMakeScale(TRANSITION_SCALE, TRANSITION_SCALE);
			self.view.hidden = NO;
			[UIView setAnimationsEnabled: bPrevAnimationEnabled];
		}

		self.imageView.hidden = YES;
		[self setMessage: msgString];
		[self.activityIndicator startAnimating];

		if (kBackgroundClear == overlayMode)
		{
			self.overlayWindow.layer.contents = (id)[self clearBackgroundImage];
		}
		else if (kBackgroundBlack == overlayMode)
		{
			self.overlayWindow.layer.contents = (id)[self blackBackgroundImage];
		}
		else if (kBackgroundGradient == overlayMode)
		{
			self.overlayWindow.layer.contents = (id)[self gradientBackgroundImage];
		}

		[self.overlayWindow makeKeyAndVisible];
		[self updateHudView: nil];
        [self registerNotifications];
		self.hudView.transform = CGAffineTransformScale(self.hudView.transform, 1.3f, 1.3f);
		
		[UIView animateWithDuration: 0.15
							  delay: 0
							options: (UIViewAnimationOptionAllowUserInteraction |
									  UIViewAnimationCurveEaseOut |
									  UIViewAnimationOptionBeginFromCurrentState)
						 animations: ^{	
										self.hudView.transform = CGAffineTransformScale(self.hudView.transform, 1/1.3, 1/1.3);
										self.view.alpha = 1.0f;
										self.view.transform = CGAffineTransformIdentity;
						 }
						 completion: NULL];
		
		[self.view setNeedsDisplay];

		self.visible = YES;
		if (NO == userInteractionEnabled)
		{
			[[UIApplication sharedApplication] beginIgnoringInteractionEvents];
		}
	}
}


//============================================================================
//	S4OverlayViewController :: dismissOverlayWithMessage:
//============================================================================
- (void)dismissOverlayWithMessage: (NSString *)msgString
							image: (UIImage *)imageToDisplay
					   afterDelay: (NSTimeInterval)seconds
{
	if (YES == self.visible)
	{
		if (IS_NOT_NULL(imageToDisplay))
		{
			self.imageView.image = imageToDisplay;
			self.imageView.hidden = NO;
		}

		[self setMessage: msgString];
		[self.activityIndicator stopAnimating];
		[self performSelector: @selector(dismissOverlay) withObject: nil afterDelay: seconds];
	}
}


//============================================================================
//	S4OverlayViewController :: dismissOverlay
//============================================================================
- (void)dismissOverlay
{
	__block NSArray						*appWindowArray;

	if (YES == self.visible)
	{
		[self.activityIndicator stopAnimating];

		[UIView animateWithDuration: 0.15
							  delay: 0
							options: UIViewAnimationCurveEaseIn | UIViewAnimationOptionAllowUserInteraction
						 animations: ^{	
										self.hudView.transform = CGAffineTransformScale(self.hudView.transform, 0.8, 0.8);
										self.view.alpha = 0.0f;
										self.view.transform = CGAffineTransformMakeScale(TRANSITION_SCALE, TRANSITION_SCALE);
									  }
						 completion: ^(BOOL finished) {
										[[NSNotificationCenter defaultCenter] removeObserver: self];

										// find the frontmost window that is an actual UIWindow and make it keyVisible
										appWindowArray = [[UIApplication sharedApplication] windows];
										[appWindowArray enumerateObjectsWithOptions: NSEnumerationReverse
																		 usingBlock: ^(UIWindow *window, NSUInteger idx, BOOL *stop)
																					 {
																						if (([window isKindOfClass: [UIWindow class]]) &&
																							(window.windowLevel == UIWindowLevelNormal))
																						{
																							[window makeKeyWindow];
																							*stop = YES;
																						}
																					 }];
										if (NO == self.visible)
										{
											self.view.hidden = YES;
										}
									}];
		self.visible = NO;
		[[UIApplication sharedApplication] endIgnoringInteractionEvents];
	}
}


//============================================================================
//	S4OverlayViewController :: setMessage:
//============================================================================
- (void)setMessage: (NSString *)msgString
{
    CGFloat						hudWidth;
    CGFloat						hudHeight;
    CGFloat						stringWidth;
    CGFloat						stringHeight;
    CGRect						labelRect;
	CGSize						stringSize;

	hudWidth = 100.0f;
    hudHeight = 100.0f;
    labelRect = CGRectZero;

    if (STR_NOT_EMPTY(msgString))
	{
        stringSize = [msgString sizeWithFont: self.messageLabel.font constrainedToSize: CGSizeMake(200.0f, 300.0f)];
        stringWidth = stringSize.width;
        stringHeight = stringSize.height;
        hudHeight = 80.0f + stringHeight;
        
        if (stringWidth > hudWidth)
		{
            hudWidth = ceil(stringWidth/2) * 2;
        }
		
        if (hudHeight > 100)
		{
            labelRect = CGRectMake(12.0f, 66.0f, hudWidth, stringHeight);
            hudWidth+=24;
        }
		else
		{
            hudWidth+=24;  
            labelRect = CGRectMake(0.0f, 66.0f, hudWidth, stringHeight);   
        }
    }
	
	self.hudView.bounds = CGRectMake(0.0f, 0.0f, hudWidth, hudHeight);
	
    if (STR_NOT_EMPTY(msgString))
	{
        self.imageView.center = CGPointMake(CGRectGetWidth(self.hudView.bounds)/2, 36.0f);
	}
	else
	{
       	self.imageView.center = CGPointMake(CGRectGetWidth(self.hudView.bounds)/2, CGRectGetHeight(self.hudView.bounds)/2);
	}
	
	self.messageLabel.hidden = NO;
	self.messageLabel.text = msgString;
	self.messageLabel.frame = labelRect;
	
	if (STR_NOT_EMPTY(msgString))
	{
		self.activityIndicator.center = CGPointMake(ceil(CGRectGetWidth(self.hudView.bounds)/2) + 0.5f, 40.5f);
	}
	else
	{
		self.activityIndicator.center = CGPointMake(ceil(CGRectGetWidth(self.hudView.bounds)/2) + 0.5f,
													ceil(self.hudView.bounds.size.height/2) + 0.5f);
	}
}



/********************************** Private Methods **********************************/

//============================================================================
//	S4OverlayViewController () :: gradientBackgroundImage
//============================================================================
- (CGImageRef)gradientBackgroundImage
{
	CGContextRef					ctx;
	CGAffineTransform				transform;
	CGPoint							centerPoint;
	CGPoint							newCenter;
	CGFloat							alpha;
	NSArray							*colors;
	CGGradientRef					gradient;
	UIImage							*image;

	UIGraphicsBeginImageContext([UIScreen mainScreen].bounds.size);
	ctx = UIGraphicsGetCurrentContext();
	transform = CGAffineTransformMakeScale(1.0f, 1.1f);
	centerPoint = CGPointMake(CGRectGetMidX([UIScreen mainScreen].bounds), CGRectGetMidY([UIScreen mainScreen].bounds));
	newCenter = CGPointApplyAffineTransform(centerPoint, CGAffineTransformInvert(transform));
	transform = CGAffineTransformTranslate(transform, newCenter.x - centerPoint.x, newCenter.y - centerPoint.y);
	CGContextConcatCTM(ctx, transform);
	alpha = 0.4f;
	
	colors = [NSArray arrayWithObjects: (id)[UIColor colorWithWhite: 0.1f alpha: alpha].CGColor,
			  (id)[UIColor colorWithWhite: 0.7f alpha: alpha].CGColor, nil];
	
	gradient = CGGradientCreateWithColors(CGColorGetColorSpace((CGColorRef)[colors objectAtIndex:0]), (CFArrayRef)colors, NULL);
	
	CGContextDrawRadialGradient(ctx, gradient, centerPoint, 120, centerPoint, 0,
								kCGGradientDrawsBeforeStartLocation | kCGGradientDrawsAfterEndLocation);
	CGGradientRelease(gradient);
	image = UIGraphicsGetImageFromCurrentImageContext();
	UIGraphicsEndImageContext();
	return (image.CGImage);
}


//============================================================================
//	S4OverlayViewController () :: blackBackgroundImage
//============================================================================
- (CGImageRef)blackBackgroundImage
{
	CGContextRef					ctx;
	UIImage							*image;

	UIGraphicsBeginImageContext([UIScreen mainScreen].bounds.size);
	ctx = UIGraphicsGetCurrentContext();
	[[UIColor colorWithWhite: 0.0f alpha: 0.5f] set];
	CGContextFillRect(ctx, [UIScreen mainScreen].bounds);

	image = UIGraphicsGetImageFromCurrentImageContext();
	UIGraphicsEndImageContext();
	return (image.CGImage);
}


//============================================================================
//	S4OverlayViewController () :: clearBackgroundImage
//============================================================================
- (CGImageRef)clearBackgroundImage
{
	CGContextRef					ctx;
	UIImage							*image;
	
	UIGraphicsBeginImageContext([UIScreen mainScreen].bounds.size);
	ctx = UIGraphicsGetCurrentContext();
	[[UIColor colorWithWhite: 0.0f alpha: 0.0f] set];
	CGContextFillRect(ctx, [UIScreen mainScreen].bounds);
	
	image = UIGraphicsGetImageFromCurrentImageContext();
	UIGraphicsEndImageContext();
	return (image.CGImage);
	
}


//============================================================================
//	S4OverlayViewController () :: moveToPoint:
//============================================================================
- (void)moveToPoint: (CGPoint)newCenter rotateAngle: (CGFloat)angle
{
    self.hudView.transform = CGAffineTransformMakeRotation(angle); 
    self.hudView.center = newCenter;
}


//============================================================================
//	S4OverlayViewController () :: updateHudView:
//============================================================================
- (void)updateHudView: (NSNotification *)notification
{
	NSDictionary						*notifyInfoDict;
	CGRect								keyboardFrame;
    double								animationDuration = 0.0;
	CGFloat								keyboardHeight;
    UIInterfaceOrientation				orientation = [[UIApplication sharedApplication] statusBarOrientation];
	CGRect								orientationFrame;
	CGRect								statusBarFrame;
	CGFloat								activeHeight;
	CGFloat								posY;
	CGFloat								posX;
	CGPoint								newCenter;
    CGFloat								rotateAngle;

    if (IS_NOT_NULL(notification))
	{
        notifyInfoDict = [notification userInfo];
        keyboardFrame = [[notifyInfoDict valueForKey: UIKeyboardFrameBeginUserInfoKey] CGRectValue];
        animationDuration = [[notifyInfoDict valueForKey: UIKeyboardAnimationDurationUserInfoKey] doubleValue];
        
        if ((notification.name == UIKeyboardWillShowNotification) ||
			(notification.name == UIKeyboardDidShowNotification))
		{
            if (UIInterfaceOrientationIsPortrait(orientation))
			{
                keyboardHeight = keyboardFrame.size.height;
			}
            else
			{
                keyboardHeight = keyboardFrame.size.width;
			}
        }
		else
		{
            keyboardHeight = 0.0f;
		}
    }
	else
	{
        keyboardHeight = [self visibleKeyboardHeight];
    }
    
    orientationFrame = [UIScreen mainScreen].bounds;
    statusBarFrame = [UIApplication sharedApplication].statusBarFrame;
    
    if (UIInterfaceOrientationIsLandscape(orientation))
	{
        float temp = orientationFrame.size.width;
        orientationFrame.size.width = orientationFrame.size.height;
        orientationFrame.size.height = temp;
        
        temp = statusBarFrame.size.width;
        statusBarFrame.size.width = statusBarFrame.size.height;
        statusBarFrame.size.height = temp;
    }
    
    activeHeight = orientationFrame.size.height;
    
    if (keyboardHeight > 0)
	{
        activeHeight += statusBarFrame.size.height * 2;
	}
    
    activeHeight -= keyboardHeight;
    posY = floor(activeHeight * 0.50);
    posX = orientationFrame.size.width/2;

    switch (orientation)
	{ 
        case UIInterfaceOrientationPortraitUpsideDown:
            rotateAngle = M_PI; 
            newCenter = CGPointMake(posX, orientationFrame.size.height - posY);
            break;

        case UIInterfaceOrientationLandscapeLeft:
            rotateAngle = -M_PI/2.0f;
            newCenter = CGPointMake(posY, posX);
            break;

        case UIInterfaceOrientationLandscapeRight:
            rotateAngle = M_PI/2.0f;
            newCenter = CGPointMake(orientationFrame.size.height - posY, posX);
            break;

		case UIInterfaceOrientationPortrait:
        default:
            rotateAngle = 0.0f;
            newCenter = CGPointMake(posX, posY);
            break;
    } 
    
    if (IS_NOT_NULL(notification))
	{
        [UIView animateWithDuration: animationDuration 
                              delay: 0 
                            options: UIViewAnimationOptionAllowUserInteraction 
                         animations: ^{
										[self moveToPoint: newCenter rotateAngle: rotateAngle];
									  }
						 completion: NULL];
    } 
    
    else
	{
        [self moveToPoint: newCenter rotateAngle: rotateAngle];
    }
    
}


//============================================================================
//	S4OverlayViewController () :: registerNotifications
//============================================================================
- (void)registerNotifications
{
    [[NSNotificationCenter defaultCenter] addObserver: self 
                                             selector: @selector(updateHudView:) 
                                                 name: UIApplicationDidChangeStatusBarOrientationNotification 
                                               object: nil];  

    [[NSNotificationCenter defaultCenter] addObserver: self 
                                             selector: @selector(updateHudView:) 
                                                 name: UIKeyboardWillHideNotification
                                               object: nil];

    [[NSNotificationCenter defaultCenter] addObserver: self 
                                             selector: @selector(updateHudView:) 
                                                 name: UIKeyboardDidHideNotification
                                               object: nil];

    [[NSNotificationCenter defaultCenter] addObserver: self 
                                             selector: @selector(updateHudView:) 
                                                 name: UIKeyboardWillShowNotification
                                               object: nil];

    [[NSNotificationCenter defaultCenter] addObserver: self 
                                             selector: @selector(updateHudView:) 
                                                 name: UIKeyboardDidShowNotification
                                               object: nil];
}


//============================================================================
//	S4OverlayViewController () :: visibleKeyboardHeight
//============================================================================
- (CGFloat)visibleKeyboardHeight
{
	NSArray						*appWindowArray;
	NSArray						*testKbdViewArray;
	UIView						*iOS4KeyboardView;
	UIView						*foundKeyboard = nil;
	
	@autoreleasepool
	{
		appWindowArray = [[UIApplication sharedApplication] windows];
		for (UIWindow *testWindow in appWindowArray)
		{
			if (![[testWindow class] isEqual: [UIWindow class]])
			{
				testKbdViewArray = [testWindow subviews];
				for (UIView *testKeyboardView in testKbdViewArray)
				{
					if ([[testKeyboardView description] hasPrefix: @"<UIKeyboard"])
					{
						foundKeyboard = testKeyboardView;
						break;
					}
					else if ([[testKeyboardView description] hasPrefix: @"<UIPeripheralHostView"])
					{
						// iOS 4 embeds UIKeyboards inside a UIPeripheralHostView
						iOS4KeyboardView = [[testKeyboardView subviews] objectAtIndex: 0];
						if ([[iOS4KeyboardView description] hasPrefix: @"<UIKeyboard"])
						{
							foundKeyboard = testKeyboardView;
							break;
						}
					}                                                                                
				}
			}

			if (IS_NOT_NULL(foundKeyboard))
			{
				break;
			}
		}
	}
	
    if ((IS_NOT_NULL(foundKeyboard)) && (foundKeyboard.bounds.size.height > 100))
	{
        return (foundKeyboard.bounds.size.height);
    }
    return (0.0f);
}

@end
