/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the S4 iOS Libraries.
 *
 * The Initial Developer of the Original Code is
 * Michael Papp dba SeaStones Software Company.
 * All software created by the Initial Developer are Copyright (C) 2008-2011
 * the Initial Developer. All Rights Reserved.
 *
 * Original Author:
 *			Unknown - many sources
 *
 * ***** END LICENSE BLOCK ***** */

/* ***** FILE BLOCK ******
 *
 * Name:		S4PageControl.m
 * Module:		UI
 * Library:		S4 iOS Libraries
 *
 * ***** FILE BLOCK *****/


// ================================== Includes =========================================

#import "S4PageControl.h"
#import "S4CommonDefines.h"


// =================================== Defines =========================================



// ================================== Typedefs =========================================



// =================================== Globals =========================================

static const float kDefaultDotSize = 6.0;
static const float kDefaultDotSpacing = 10.0;


// ============================= Forward Declarations ==================================



// ================================== Inlines ==========================================



// ========================== Begin Class S4PageControl () =============================

@interface S4PageControl ()

- (void)setup;
- (void)updatePageControlNotification: (NSNotification *)orientationNotification;
- (void)privateSetCurrentPage: (NSInteger)page;
- (void)setPageTouched: (UITouch *)touch;

@end




// ============================ Begin Class S4PageControl ==============================

@implementation S4PageControl
{
	NSInteger							m_currentPage;
	NSInteger							m_numberOfPages;
	BOOL								m_defersCurrentPageDisplay;
	BOOL								m_hidesForSinglePage;
	BOOL								m_shouldWrap;
	UIColor								*m_defaultDotColor;
	UIColor								*m_selectedDotColor;
	UIColor								*m_highlightedDotColor;
	CGFloat								m_dotSpacing;
	CGFloat								m_dotSize;
	BOOL								m_shouldEnableHighlight;
	BOOL								m_userIsDragging;
}

//============================================================================
//	S4PageControl :: properties
//============================================================================

@dynamic currentPage;
@dynamic numberOfPages;
@synthesize defersCurrentPageDisplay = m_defersCurrentPageDisplay;
@dynamic hidesForSinglePage;
@synthesize wrap = m_shouldWrap;
@synthesize defaultDotColor = m_defaultDotColor;
@synthesize selectedDotColor = m_selectedDotColor;
@synthesize highlightedDotColor = m_highlightedDotColor;
@synthesize dotSpacing = m_dotSpacing;
@synthesize dotSize = m_dotSize;
@synthesize shouldEnableHighlight = m_shouldEnableHighlight;


//============================================================================
//	S4PageControl :: initWithFrame:
//============================================================================
- (id)initWithFrame: (CGRect)frame
{
	self = [super initWithFrame: frame];
    if (nil != self)
	{
		[self setup];
    }
    return self;
}


//============================================================================
//	S4PageControl :: initWithCoder:
//============================================================================
- (id)initWithCoder: (NSCoder *)aDecoder
{
	self = [super initWithCoder: aDecoder];
	if (nil != self)
	{
		[self setup];
	}
	return self;
}


//============================================================================
//	S4PageControl :: dealloc
//============================================================================
- (void)dealloc
{	
	self.defaultDotColor = nil;
	self.selectedDotColor = nil;
	self.highlightedDotColor = nil;
	[[NSNotificationCenter defaultCenter] removeObserver: self];

    [super dealloc];
}


//============================================================================
//	S4PageControl :: sizeForNumberOfPages:
//============================================================================
- (CGSize)sizeForNumberOfPages: (NSInteger)pageCount
{
	CGFloat					width;
	CGFloat					curDotSpacing;
	
	width = self.bounds.size.width;
	curDotSpacing = width / pageCount;
	if (curDotSpacing > kDefaultDotSpacing)
	{
		curDotSpacing = kDefaultDotSpacing;
	}
	return (CGSizeMake(curDotSpacing * pageCount, curDotSpacing));
}


//============================================================================
//	S4PageControl :: updateCurrentPageDisplay
//============================================================================
- (void)updateCurrentPageDisplay
{
    if (self.defersCurrentPageDisplay)
    {
        [self setNeedsDisplay];
    }
}



/****************************** UIControl Override Methods ******************************/

//============================================================================
//	S4PageControl :: drawRect:
//============================================================================
- (void)drawRect: (CGRect)rect
{
	CGContextRef					context;
	CGRect							selectedPageDotRect;
	CGFloat							width;
	CGFloat							curDotSpacing;
	CGFloat							curDotSize;
	CGRect							curPageDotRect;
	CGFloat							inset;
	CGMutablePathRef				dotsPath;
	CGMutablePathRef				highlightPath;

	[super drawRect: rect];

	if ((self.numberOfPages > 1) || (NO == self.hidesForSinglePage))
	{
		context = UIGraphicsGetCurrentContext();
		CGContextSaveGState(context);

		selectedPageDotRect = CGRectZero;
		width = self.bounds.size.width;
		curDotSpacing = width / self.numberOfPages;
		if (curDotSpacing > kDefaultDotSpacing)
		{
			curDotSpacing = kDefaultDotSpacing;
		}
		curDotSize = kDefaultDotSize;
		
		curPageDotRect.origin.x = (width - curDotSpacing * self.numberOfPages) / 2.0;
		curPageDotRect.origin.y = (self.bounds.size.height - curDotSpacing) / 2.0;
		curPageDotRect.size.height = curDotSpacing;
		curPageDotRect.size.width = curDotSpacing;
		inset = (curDotSpacing - curDotSize) / 2.0;

		for (int dotNum = 0; dotNum < self.numberOfPages; dotNum++)
		{
			dotsPath = CGPathCreateMutable();
			CGPathAddEllipseInRect(dotsPath, nil, CGRectInset(curPageDotRect, inset, inset));
			CGContextAddPath(context, dotsPath);
			
			if (dotNum == self.currentPage) 
			{
				selectedPageDotRect = curPageDotRect;
				[self.selectedDotColor setFill];
				CGContextFillPath(context);
			}
			else
			{
				[self.defaultDotColor setFill];
				CGContextFillPath(context);
			}
			
			curPageDotRect.origin.x += curDotSpacing;
			CGPathRelease(dotsPath);
		}

		// draw highlight
		if (self.shouldEnableHighlight)
		{
			highlightPath = CGPathCreateMutable();
			CGPathAddEllipseInRect(highlightPath, nil, CGRectInset(selectedPageDotRect, inset, inset));
			CGContextAddPath(context, highlightPath);
			CGContextSetLineWidth(context, 1.5f);
			[self.highlightedDotColor setStroke];
			CGContextStrokePath(context);
			CGPathRelease(highlightPath);
		}

		CGContextRestoreGState(context);
	}
}


//============================================================================
//	S4PageControl :: sizeThatFits:
//============================================================================
- (CGSize)sizeThatFits: (CGSize)size
{
    return CGSizeMake(self.superview.bounds.size.width, [self sizeForNumberOfPages: 1].height);
}



/****************************** Property Accessor Methods ******************************/

//============================================================================
//	S4PageControl :: currentPage
//============================================================================
- (NSInteger)currentPage
{
	return (m_currentPage);
}


//============================================================================
//	S4PageControl :: setCurrentPage:
//============================================================================
- (void)setCurrentPage: (NSInteger)page
{
	// do not allow the current page to be set if we are tracking a drag
	if (NO == self.tracking)
	{
		[self privateSetCurrentPage: page];
	}
}


//============================================================================
//	S4PageControl :: numberOfPages
//============================================================================
- (NSInteger)numberOfPages
{
	return (m_numberOfPages);
}


//============================================================================
//	S4PageControl :: setNumberOfPages:
//============================================================================
- (void)setNumberOfPages: (NSInteger)pages
{
	if (m_numberOfPages != pages)
    {
        m_numberOfPages = pages;
        if (self.currentPage >= m_numberOfPages)
        {
            self.currentPage = m_numberOfPages - 1;
        }

		if ((YES == self.hidesForSinglePage) && (m_numberOfPages <= 1))
		{
			self.hidden = YES;
		}
		else
		{
			self.hidden = NO;
		}
        [self setNeedsDisplay];
    }
}


//============================================================================
//	S4PageControl :: hidesForSinglePage
//============================================================================
- (BOOL)hidesForSinglePage
{
	return (m_hidesForSinglePage);
}


//============================================================================
//	S4PageControl :: setHidesForSinglePage:
//============================================================================
- (void)setHidesForSinglePage: (BOOL)shouldHide
{
	if (shouldHide != m_hidesForSinglePage)
	{
		m_hidesForSinglePage = shouldHide;

		if ((YES == m_hidesForSinglePage) && (self.numberOfPages <= 1))
		{
			self.hidden = YES;
		}
		else
		{
			self.hidden = NO;
		}
		[self setNeedsDisplay];
	}
}



/****************************** Touch Delegate Methods ******************************/

//============================================================================
//	S4PageControl :: beginTrackingWithTouch:
//============================================================================
- (BOOL)beginTrackingWithTouch: (UITouch *)touch withEvent: (UIEvent *)event
{
	m_userIsDragging = NO;
	return [super beginTrackingWithTouch: touch withEvent: event];
}


//============================================================================
//	S4PageControl :: continueTrackingWithTouch:
//============================================================================
- (BOOL)continueTrackingWithTouch: (UITouch *)touch withEvent: (UIEvent *)event
{
	// we are now dragging and will no longer respond to endTracking
	m_userIsDragging = YES;

	[self setPageTouched: touch];
	return ([super continueTrackingWithTouch: touch withEvent: event]);
}


//============================================================================
//	S4PageControl :: endTrackingWithTouch:
//============================================================================
- (void)endTrackingWithTouch: (UITouch *)touch withEvent: (UIEvent *)event
{
	if (NO == m_userIsDragging)
	{
		[self setPageTouched: touch];
	}
}



/****************************** Private Methods ******************************/

//============================================================================
//	S4PageControl :: setup
//============================================================================
- (void)setup
{	
    // needs redrawing if bounds change
    // this isn't a private method, but using KVC avoids having to import <QuartzCore>
    [self.layer setValue: [NSNumber numberWithBool: YES] forKey: @"needsDisplayOnBoundsChange"];
	
	if (nil == self.backgroundColor)
	{
		self.backgroundColor = [UIColor clearColor];
	}

	// set defaults
	self.currentPage = 0;
	self.numberOfPages = 0;
	self.defersCurrentPageDisplay = NO;
	self.hidesForSinglePage = NO;
	self.wrap = NO;
	self.defaultDotColor = [UIColor colorWithWhite: 0.0 alpha: 0.25];
    self.selectedDotColor = [UIColor blackColor];
	self.highlightedDotColor = [UIColor grayColor];
	self.dotSpacing = kDefaultDotSpacing;
	self.dotSize = kDefaultDotSize;
	self.shouldEnableHighlight = NO;
	m_userIsDragging = NO;

	// receive notification about device rotations
	[[NSNotificationCenter defaultCenter] addObserver: self 
											 selector: @selector(updatePageControlNotification:) 
												 name: UIApplicationDidChangeStatusBarOrientationNotification 
											   object: nil];
}


//============================================================================
//	S4PageControl :: updatePageControlNotification:
//============================================================================
- (void)updatePageControlNotification: (NSNotification *)orientationNotification
{
	[self setNeedsDisplay];
}


//============================================================================
//	S4PageControl :: privateSetCurrentPage:
//============================================================================
- (void)privateSetCurrentPage: (NSInteger)page
{
	NSInteger				clampedValue;

	if (self.wrap)
    {
        clampedValue = (page + self.numberOfPages) % self.numberOfPages;
    }
    else
    {
        clampedValue = MIN(MAX(0, page), self.numberOfPages - 1);
    }

	if (clampedValue != m_currentPage)
	{
		m_currentPage = clampedValue;
		[self sendActionsForControlEvents: UIControlEventValueChanged];
		[self setNeedsDisplay];
	}
}


//============================================================================
//	S4PageControl :: setPageTouched:
//============================================================================
- (void)setPageTouched: (UITouch *)touch
{
	CGFloat					width;
	CGFloat					curDotSpacing;
	CGFloat					leftMargin;
	CGPoint					localPt;
	NSInteger				pageIndex;

	// figure out the dot positions
	width = self.bounds.size.width;
	curDotSpacing = width / self.numberOfPages;
	if (curDotSpacing > kDefaultDotSpacing)
	{
		curDotSpacing = kDefaultDotSpacing;
	}
	leftMargin = (width - curDotSpacing * self.numberOfPages) / 2.0;

	// set our current page to the one the touch is over
	localPt = [touch locationInView: self];
	pageIndex = (localPt.x - leftMargin) / curDotSpacing;
	[self privateSetCurrentPage: pageIndex];
}

@end
