//
//  CPropertyFactory.m
//
//  Created by Jonathan Wight on 10/3/11.
//  Copyright 2011 Jonathan Wight. All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without modification, are
//  permitted provided that the following conditions are met:
//
//     1. Redistributions of source code must retain the above copyright notice, this list of
//        conditions and the following disclaimer.
//
//     2. Redistributions in binary form must reproduce the above copyright notice, this list
//        of conditions and the following disclaimer in the documentation and/or other materials
//        provided with the distribution.
//
//  THIS SOFTWARE IS PROVIDED BY 2011 TOXICSOFTWARE.COM ``AS IS'' AND ANY EXPRESS OR IMPLIED
//  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
//  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 2011 TOXICSOFTWARE.COM OR
//  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
//  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
//  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
//  ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
//  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
//  ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
//  The views and conclusions contained in the software and documentation are those of the
//  authors and should not be interpreted as representing official policies, either expressed
//  or implied, of 2011 toxicsoftware.com.
//
// See http://developer.apple.com/library/ios/#DOCUMENTATION/Cocoa/Conceptual/ObjCRuntimeGuide/Articles/ocrtPropertyIntrospection.html
//  for all flags and
// http://developer.apple.com/library/mac/#documentation/Cocoa/Conceptual/ObjCRuntimeGuide/Articles/ocrtTypeEncodings.html
//  for all encodings
//


#import "CPropertyFactory.h"
#import <objc/runtime.h>



#define NUM_GETTER(T, T2)	^(id _self) {																\
											return ([inGetter(_self, thePropertyName) T##Value]);		\
										};

#define NUM_SETTER(T, T2)	^(id _self, T value) {																		\
													inSetter(_self, thePropertyName, [NSNumber numberWith##T2:value]);	\
												 };


#define VALUE_GETTER(T)		^(id _self) {															\
											T value;												\
																									\
											[inGetter(_self, thePropertyName) getValue: &value];	\
											return (value);											\
										};

#define VALUE_SETTER(T)		^(id _self, T value) {																			\
													inSetter(_self, thePropertyName, [NSValue valueWithBytes: &value		\
																									objCType: @encode(T)]);	\
												 };



NSString * const								CPropertyFactoryErrorDomain = @"CPropertyFactoryErrorDomain";




@implementation CPropertyFactory

- (BOOL)configure: (Class)inClass
		   getter: (PropertyGetterBlock)inGetter
		   setter: (PropertySetterBlock)inSetter
			error: (NSError **)outError
{
    unsigned int			propertyCount = 0;
	unsigned int			index;
    objc_property_t			*propertiesArray;
	BOOL					bResult = NO;

	if (nil != outError)
	{
		*outError = nil;
	}

	propertiesArray = class_copyPropertyList(inClass, &propertyCount);
	if (NULL != propertiesArray)
	{
		for (index = 0; index != propertyCount; ++index)
		{
			bResult = [self configure: inClass
							 property: propertiesArray[index]
							   getter: inGetter
							   setter: inSetter
								error: outError];
			if (NO == bResult)
			{
				break;
			}
		}
	}

	if (NULL != propertiesArray)
	{
		free(propertiesArray);
	}
	return (bResult);
}


- (BOOL)configure: (Class)inClass
	 propertyName: (NSString *)inProperty
		   getter: (PropertyGetterBlock)inGetter
		   setter: (PropertySetterBlock)inSetter
			error: (NSError **)outError
{
	objc_property_t				curProperty;

	if (nil != outError)
	{
		*outError = nil;
	}

	curProperty = class_getProperty(inClass, [inProperty UTF8String]);
	if (NULL != curProperty)
	{
		return ([self configure: inClass
					   property: curProperty
						 getter: inGetter
						 setter: inSetter
						  error: outError]);
		
	}
	return (NO);
}


- (BOOL)configure: (Class)inClass
		 property: (objc_property_t)inProperty
		   getter: (PropertyGetterBlock)inGetter
		   setter: (PropertySetterBlock)inSetter
			error: (NSError **)outError
{
	PropertyGetterBlock					getterCopy;
	PropertySetterBlock					setterCopy;
	objc_property_attribute_t			*objcAttributes = NULL;
	BOOL								theDynamicFlag = NO;
    BOOL								theReadonlyFlag = NO;
	BOOL								theNonAtomicFlag = NO;
    const char							*theType = NULL;
	unsigned int						attributeCount = 0;
	NSMutableDictionary					*userDict = nil;
	NSString							*localizedDescription = @"CFPropertyFactory configuration failed";
	NSString							*localizedFailureReason = nil;
	CPropertyFactoryError				code = CPropertyFactoryNoError;
    BOOL								theResult = YES;

	if (nil != outError)
	{
		*outError = nil;
	}

    getterCopy = [inGetter copy];
    setterCopy = [inSetter copy];

	objcAttributes = property_copyAttributeList(inProperty, &attributeCount);
	if (NULL != objcAttributes)
	{
		for (unsigned int theAttributeIndex = 0; theAttributeIndex != attributeCount; ++theAttributeIndex)
		{
			if (strlen(objcAttributes[theAttributeIndex].name) == 1)
			{
				switch (objcAttributes[theAttributeIndex].name[0])
				{
					case 'D':
						theDynamicFlag = YES;
						break;
					case 'T':
						theType = objcAttributes[theAttributeIndex].value;
						break;
					case 'R':
						theReadonlyFlag = YES;
						break;
					case 'N':
						theNonAtomicFlag = YES;
						break;
					default:
						break;
				}
			}
		}

		if ((theDynamicFlag == YES) && (theType != NULL))
		{
			NSString		*thePropertyName = [NSString stringWithUTF8String: property_getName(inProperty)];
			NSString		*theGetterName = thePropertyName;
			NSString		*theSetterName = [NSString stringWithFormat: @"set%@%@:",
											  [[thePropertyName substringToIndex:1] uppercaseString],
											  [thePropertyName substringFromIndex:1]];
			NSString		*theGetterType = [NSString stringWithFormat: @"%s@:", theType];
			NSString		*theSetterType = [NSString stringWithFormat: @"v@:%s", theType];
			id				theGetterIMPBlock = NULL;
			id				theSetterIMPBlock = NULL;

			if (theType[0] == '@')
			{
				theGetterIMPBlock = ^(id _self) {
													return(getterCopy(_self, thePropertyName));
												};

				theSetterIMPBlock = ^(id _self, NSString *value) {
																	setterCopy(_self, thePropertyName, value);
																 };
			}
			else if (theType[0] != '{')
			{
				if (strcmp(theType, @encode(BOOL)) == 0)
				{
					theGetterIMPBlock = NUM_GETTER(bool, Bool);
					theSetterIMPBlock = NUM_SETTER(bool, Bool);
				}
				else if (strcmp(theType, @encode(short)) == 0)
				{
					theGetterIMPBlock = NUM_GETTER(short, Short);
					theSetterIMPBlock = NUM_SETTER(short, Short);
				}
				else if (strcmp(theType, @encode(unsigned short)) == 0)
				{
					theGetterIMPBlock = NUM_GETTER(unsignedShort, UnsignedShort);
					theSetterIMPBlock = NUM_SETTER(unsigned short, UnsignedShort);
				}
				else if (strcmp(theType, @encode(int)) == 0)
				{
					theGetterIMPBlock = NUM_GETTER(int, Int);
					theSetterIMPBlock = NUM_SETTER(int, Int);
				}
				else if (strcmp(theType, @encode(unsigned int)) == 0)
				{
					theGetterIMPBlock = NUM_GETTER(unsignedInt, UnsignedInt);
					theSetterIMPBlock = NUM_SETTER(unsigned int, UnsignedInt);
				}
				else if (strcmp(theType, @encode(long)) == 0)
				{
					theGetterIMPBlock = NUM_GETTER(long, Long);
					theSetterIMPBlock = NUM_SETTER(long, Long);
				}
				else if (strcmp(theType, @encode(unsigned long)) == 0)
				{
					theGetterIMPBlock = NUM_GETTER(unsignedLong, UnsignedLong);
					theSetterIMPBlock = NUM_SETTER(unsigned long, UnsignedLong);
				}
				else if (strcmp(theType, @encode(long long)) == 0)
				{
					theGetterIMPBlock = NUM_GETTER(longLong, LongLong);
					theSetterIMPBlock = NUM_SETTER(long long, LongLong);
				}
				else if (strcmp(theType, @encode(unsigned long long)) == 0)
				{
					theGetterIMPBlock = NUM_GETTER(unsignedLongLong, UnsignedLongLong);
					theSetterIMPBlock = NUM_SETTER(unsigned long long, UnsignedLongLong);
				}
				else if (strcmp(theType, @encode(float)) == 0)
				{
					theGetterIMPBlock = NUM_GETTER(float, Float);
					theSetterIMPBlock = NUM_SETTER(float, Float);
				}
				else if (strcmp(theType, @encode(double)) == 0)
				{
					theGetterIMPBlock = NUM_GETTER(double, Double);
					theSetterIMPBlock = NUM_SETTER(double, Double);
				}

				else if (strcmp(theType, @encode(void *)) == 0)
				{
					theGetterIMPBlock = VALUE_GETTER(void *);
					theSetterIMPBlock = VALUE_SETTER(void *);
				}

				else
				{
					localizedFailureReason =  [NSString stringWithFormat: @"Unknown simple type: '%s'", theType];
					code = CPropertyFactoryUnimplementedError;
				}
			}
			else if (theType[0] == '{')
			{
				if (strcmp(theType, @encode(NSRange)) == 0)
				{
					theGetterIMPBlock = VALUE_GETTER(NSRange);
					theSetterIMPBlock = VALUE_SETTER(NSRange);
				}
#if (TARGET_OS_MAC && !(TARGET_OS_EMBEDDED || TARGET_OS_IPHONE))
				else if (strcmp(theType, @encode(NSPoint)) == 0)
				{
					theGetterIMPBlock = VALUE_GETTER(NSPoint);
					theSetterIMPBlock = VALUE_SETTER(NSPoint);
				}
#endif
			}

			if (theGetterIMPBlock == NULL)
			{
				theResult = NO;
				localizedFailureReason = @"No getter implementation block";
				code = CPropertyFactoryNoGetterImplError;
			}
			if (getterCopy == NULL)
			{
				theResult = NO;
				localizedFailureReason = @"No getter block";
				code = CPropertyFactoryNoGetterError;
			}

//			IMP theGetterFunctionPtr = imp_implementationWithBlock((__bridge id)((__bridge void *)theGetterIMPBlock));
			IMP theGetterFunctionPtr = imp_implementationWithBlock(theGetterIMPBlock);
			class_addMethod(inClass, NSSelectorFromString(theGetterName), theGetterFunctionPtr, [theGetterType UTF8String]);

			if (theReadonlyFlag == NO)
			{
				if (theSetterIMPBlock == NULL)
				{
					theResult = NO;
					localizedFailureReason = @"No setter implementation block";
					code = CPropertyFactoryNoSetterImplError;
				}
				if (setterCopy == NULL)
				{
					theResult = NO;
					localizedFailureReason = @"No setter block";
					code = CPropertyFactoryNoSetterError;
				}

//				IMP theSetterFunctionPtr = imp_implementationWithBlock((__bridge id)((__bridge void *)theSetterIMPBlock));
				IMP theSetterFunctionPtr = imp_implementationWithBlock(theSetterIMPBlock);
				class_addMethod(inClass, NSSelectorFromString(theSetterName), theSetterFunctionPtr, [theSetterType UTF8String]);
			}
		}
	}

    if (objcAttributes != NULL)
	{
        free(objcAttributes);
	}

	if ((NO == theResult) && (nil != outError))
	{
		// set the keys in the error dictionary
		userDict = [NSMutableDictionary dictionaryWithCapacity: 2];
		[userDict setObject: localizedDescription forKey: NSLocalizedDescriptionKey];
		[userDict setObject: localizedFailureReason forKey: NSLocalizedFailureReasonErrorKey];
		// create the NSError
		*outError = [NSError errorWithDomain: CPropertyFactoryErrorDomain code: (NSInteger)code userInfo: userDict];
	}

    return (theResult);
}

@end
